/*
 * ARPFormatRule.cc - Formatting rules for ARP and RARP packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ARPFormatRule.h"
#include "LinkFormatRule.h"
#include "NetPacket.h"

/* Some systems don't define these values */
#ifndef REVARP_REQUEST
#define	REVARP_REQUEST	3
#endif
#ifndef	REVARP_REPLY
#define	REVARP_REPLY	4
#endif

ARPFormatRule::ARPFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
}

ARPFormatRule::~ARPFormatRule()
{
}

// Get the ARP header information from an incoming packet.
static int arpExtract(struct ether_arp& arphdr, const u_char *packet,
					  unsigned packetLen)
{
	// Copy the packet into an aligned arp header structure.
	if(packetLen < sizeof(arphdr))
	{
		return 0;
	}
	memcpy(&arphdr, packet, sizeof(arphdr));

	// Convert the header fields into host order.
	arphdr.arp_hrd = ntohs(arphdr.arp_hrd);
	arphdr.arp_pro = ntohs(arphdr.arp_pro);
	arphdr.arp_op = ntohs(arphdr.arp_op);
	return 1;
}

// Ethernet address consisting of all zeros.
static unsigned char etherZeroAddr[6] = {0, 0, 0, 0, 0, 0};

void ARPFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	struct ether_arp arphdr;
	if(!arpExtract(arphdr, packet.subPacket, packet.subPacketLen))
	{
		if(getTag() == PSEUDO_DLT_ARP)
		{
			output.outputShort("truncated arp packet");
		}
		else
		{
			output.outputShort("truncated rarp packet");
		}
	}
	else if((arphdr.arp_pro != ETHERTYPE_IP &&
	         arphdr.arp_pro != ETHERTYPE_TRAIL) ||
			arphdr.arp_hln != sizeof(arphdr.arp_sha) ||
			arphdr.arp_pln != sizeof(arphdr.arp_spa))
	{
		output.outputShort("arp-#%d for proto #%d (%d) hardware #%d (%d)",
						   (int)(arphdr.arp_op), (int)(arphdr.arp_pro),
						   (int)(arphdr.arp_pln), (int)(arphdr.arp_hrd),
						   (int)(arphdr.arp_hln));
	}
	else
	{
		if(arphdr.arp_pro == ETHERTYPE_TRAIL)
		{
			output.outputShort("trailer-");
		}
		switch(arphdr.arp_op)
		{
			case ARPOP_REQUEST:
			{
				output.outputShort("arp who-has %s",
								   output.ipaddrToString(arphdr.arp_tpa));
				if(memcmp(etherZeroAddr, arphdr.arp_tha, 6) != 0)
				{
					output.outputShort(" (%s)",
									   output.etherToString(arphdr.arp_tha));
				}
				output.outputShort(" tell %s",
								   output.ipaddrToString(arphdr.arp_spa));
				if((packet.linkType == DLT_EN10MB ||
				    packet.linkType == DLT_IEEE802) &&
				   packet.linkAddrLen == 14 &&
				   memcmp(packet.linkAddr + 6, arphdr.arp_sha, 6) != 0)
				{
					output.outputShort(" (%s)",
									   output.etherToString(arphdr.arp_sha));
				}
			}
			break;

			case ARPOP_REPLY:
			{
				output.outputShort("arp reply %s",
								   output.ipaddrToString(arphdr.arp_spa));
				if((packet.linkType == DLT_EN10MB ||
				    packet.linkType == DLT_IEEE802) &&
				   packet.linkAddrLen == 14 &&
				   memcmp(packet.linkAddr + 6, arphdr.arp_sha, 6) != 0)
				{
					output.outputShort(" (%s)",
									   output.etherToString(arphdr.arp_sha));
				}
				output.outputShort(" is-at %s",
								   output.etherToString(arphdr.arp_sha));
				if((packet.linkType == DLT_EN10MB ||
				    packet.linkType == DLT_IEEE802) &&
				   packet.linkAddrLen == 14 &&
				   memcmp(packet.linkAddr, arphdr.arp_tha, 6) != 0)
				{
					output.outputShort(" (%s)",
									   output.etherToString(arphdr.arp_tha));
				}
			}
			break;

			case REVARP_REQUEST:
			{
				output.outputShort("rarp who-is %s tell %s",
								   output.etherToString(arphdr.arp_tha),
								   output.etherToString(arphdr.arp_sha));
			}
			break;

			case REVARP_REPLY:
			{
				output.outputShort("rarp reply %s at %s",
								   output.etherToString(arphdr.arp_tha),
								   output.ipaddrToString(arphdr.arp_tpa));
			}
			break;

			default:
			{
				output.outputShort("arp-#%d", (int)(arphdr.arp_op));
			}
			break;
		}
		if(arphdr.arp_hrd != ARPHRD_ETHER)
		{
			output.outputShort(" hardware #%d", (int)(arphdr.arp_hrd));
		}
	}
	if(getTag() == PSEUDO_DLT_ARP)
	{
		output.shortFormColor(FormatColor_ARP);
	}
	else
	{
		output.shortFormColor(FormatColor_RARP);
	}
}

void ARPFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	struct ether_arp arphdr;
	char hexbuf[1024];
	const char *hostname;

	if(getTag() == PSEUDO_DLT_ARP)
	{
		output.outputSection("Address Resolution Protocol (ARP)");
	}
	else
	{
		output.outputSection("Reverse Address Resolution Protocol (RARP)");
	}
	if(!arpExtract(arphdr, packet.subPacket, packet.subPacketLen))
	{
		output.outputHex("Truncated Packet",
						 packet.subPacket, packet.subPacketLen);
	}
	else
	{
		if(arphdr.arp_hrd == ARPHRD_ETHER)
		{
			output.outputLong("Hardware Address Format", "1 (Ethernet)");
		}
		else
		{
			output.outputLong("Hardware Address Format", "%d",
							  (int)(arphdr.arp_hrd));
		}
		if(arphdr.arp_pro == ETHERTYPE_IP)
		{
			output.outputLong("Protocol Address Format", "0x0800 (IP)");
		}
		else if(arphdr.arp_pro == ETHERTYPE_TRAIL)
		{
			output.outputLong("Protocol Address Format", "0x1000 (Trailer)");
		}
		else
		{
			output.outputLong("Protocol Address Format", "0x%04x",
							  (int)(arphdr.arp_pro));
		}
		output.outputLong("Hardware Address Length", "%d",
						  (int)(arphdr.arp_hln));
		output.outputLong("Protocol Address Length", "%d",
						  (int)(arphdr.arp_pln));
		switch(arphdr.arp_op)
		{
			case ARPOP_REQUEST:
			{
				output.outputLong("Opcode", "1 (ARP Request)");
			}
			break;

			case ARPOP_REPLY:
			{
				output.outputLong("Opcode", "2 (ARP Reply)");
			}
			break;

			case REVARP_REQUEST:
			{
				output.outputLong("Opcode", "3 (RARP Request)");
			}
			break;

			case REVARP_REPLY:
			{
				output.outputLong("Opcode", "4 (RARP Reply)");
			}
			break;

			default:
			{
				output.outputLong("Opcode", "%d", (int)(arphdr.arp_op));
			}
			break;
		}
		packet.adjustSubPacket(8);
		if(arphdr.arp_hrd == ARPHRD_ETHER && arphdr.arp_hln == 6)
		{
			output.outputLong("Sender Hardware Address", "%s",
							  output.etherToString(packet.subPacket));
		}
		else
		{
			output.formatHex(hexbuf, packet.subPacket, arphdr.arp_hln, ':');
			output.outputLong("Sender Hardware Address", "%s", hexbuf);
		}
		packet.adjustSubPacket(arphdr.arp_hln);
		if((arphdr.arp_pro == ETHERTYPE_IP ||
		    arphdr.arp_pro == ETHERTYPE_TRAIL) && arphdr.arp_pln == 4)
		{
			hostname = output.ipaddrToDomainName(packet.subPacket);
			if(hostname)
			{
				output.outputLong("Sender Protocol Address", "%s {{%s}}",
						  output.ipaddrToStringNoResolve(packet.subPacket),
						  hostname);
			}
			else
			{
				output.outputLong("Sender Protocol Address", "%s",
						  output.ipaddrToStringNoResolve(packet.subPacket));
			}
		}
		else
		{
			output.formatHex(hexbuf, packet.subPacket, arphdr.arp_pln, ':');
			output.outputLong("Sender Protocol Address", "%s", hexbuf);
		}
		packet.adjustSubPacket(arphdr.arp_pln);
		if(arphdr.arp_hrd == ARPHRD_ETHER && arphdr.arp_hln == 6)
		{
			output.outputLong("Target Hardware Address", "%s",
							  output.etherToString(packet.subPacket));
		}
		else
		{
			output.formatHex(hexbuf, packet.subPacket, arphdr.arp_hln, ':');
			output.outputLong("Target Hardware Address", "%s", hexbuf);
		}
		packet.adjustSubPacket(arphdr.arp_hln);
		if((arphdr.arp_pro == ETHERTYPE_IP ||
		    arphdr.arp_pro == ETHERTYPE_TRAIL) && arphdr.arp_pln == 4)
		{
			hostname = output.ipaddrToDomainName(packet.subPacket);
			if(hostname)
			{
				output.outputLong("Target Protocol Address", "%s {{%s}}",
						  output.ipaddrToStringNoResolve(packet.subPacket),
						  hostname);
			}
			else
			{
				output.outputLong("Target Protocol Address", "%s",
						  output.ipaddrToStringNoResolve(packet.subPacket));
			}
		}
		else
		{
			output.formatHex(hexbuf, packet.subPacket, arphdr.arp_pln, ':');
			output.outputLong("Target Protocol Address", "%s", hexbuf);
		}
	}
}
