/*
 * ColumnFile.cc - Parse a file that contains column-based data.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetIncludes.h"
#include "ColumnFile.h"

ColumnFileParser::ColumnFileParser()
{
	file = NULL;
	linesLeftToSkip = 0;
	separators = NULL;
}

ColumnFileParser::~ColumnFileParser()
{
	if(file != 0)
	{
		fclose(file);
	}
}

void ColumnFileParser::nextLine(char **columns)
{
	// Nothing to do here: will normally be overridden by subclasses.
}

int ColumnFileParser::openFile(const char *filename, int skipLines)
{
	close();
	if((file = fopen(filename, "r")) == NULL)
	{
		return 0;
	}
	else
	{
		linesLeftToSkip = skipLines;
		return 1;
	}
}

int ColumnFileParser::openPipe(const char *command, int skipLines)
{
	close();
	if((file = popen(command, "r")) == NULL)
	{
		return 0;
	}
	else
	{
		linesLeftToSkip = skipLines;
		return 1;
	}
}

void ColumnFileParser::close(void)
{
	if(file)
	{
		fclose(file);
		file = 0;
	}
}

int ColumnFileParser::readColumns(char **& _columns)
{
	int numcolumns = 0;
	int posn;
	if(!file)
	{
		return -1;
	}
	while(fgets(buffer, BUFSIZ, file))
	{
		if(linesLeftToSkip > 0)
		{
			--linesLeftToSkip;
		}
		else if(separators)
		{
			posn = 0;
			while(buffer[posn] != '\0' && (buffer[posn] == ' ' ||
			      buffer[posn] == '\t' || buffer[posn] == '\r' ||
				  buffer[posn] == '\n' ||
				  strchr(separators, buffer[posn]) != 0))
			{
				++posn;
			}
			while(buffer[posn] != '\0')
			{
				columns[numcolumns++] = buffer + posn;
				while(buffer[posn] != '\0' && buffer[posn] != ' ' &&
				      buffer[posn] != '\t' && buffer[posn] != '\r' &&
					  buffer[posn] != '\n' &&
					  strchr(separators, buffer[posn]) == 0)
				{
					++posn;
				}
				if(buffer[posn] != '\0')
				{
					buffer[posn++] = '\0';
				}
				while(buffer[posn] != '\0' && (buffer[posn] == ' ' ||
				      buffer[posn] == '\t' || buffer[posn] == '\r' ||
					  buffer[posn] == '\n' ||
					  strchr(separators, buffer[posn]) != 0))
				{
					++posn;
				}
			}
			_columns = columns;
			return numcolumns;
		}
		else
		{
			posn = 0;
			while(buffer[posn] != '\0' && (buffer[posn] == ' ' ||
			      buffer[posn] == '\t' || buffer[posn] == '\r' ||
				  buffer[posn] == '\n'))
			{
				++posn;
			}
			while(buffer[posn] != '\0')
			{
				columns[numcolumns++] = buffer + posn;
				while(buffer[posn] != '\0' && buffer[posn] != ' ' &&
				      buffer[posn] != '\t' && buffer[posn] != '\r' &&
					  buffer[posn] != '\n')
				{
					++posn;
				}
				if(buffer[posn] != '\0')
				{
					buffer[posn++] = '\0';
				}
				while(buffer[posn] != '\0' && (buffer[posn] == ' ' ||
				      buffer[posn] == '\t' || buffer[posn] == '\r' ||
					  buffer[posn] == '\n'))
				{
					++posn;
				}
			}
			_columns = columns;
			return numcolumns;
		}
	}
	return -1;
}

void ColumnFileParser::process(int columns)
{
	int numcolumns;
	char **cols;
	while((numcolumns = readColumns(cols)) >= 0)
	{
		while(numcolumns < columns)
		{
			cols[numcolumns++] = "";
		}
		nextLine(cols);
	}
}

unsigned long ColumnFileParser::hexToNum(const char *str)
{
	unsigned long num = 0;
	while(*str != '\0')
	{
		if(*str >= '0' && *str <= '9')
			num = num * 16 + (unsigned long)(*str - '0');
		else if(*str >= 'A' && *str <= 'F')
			num = num * 16 + (unsigned long)(*str - 'A' + 10);
		else if(*str >= 'a' && *str <= 'f')
			num = num * 16 + (unsigned long)(*str - 'a' + 10);
		++str;
	}
	return num;
}

char *ColumnFileParser::hexToIP(const char *str)
{
	struct in_addr addr;
	addr.s_addr = hexToNum(str);
	if(addr.s_addr == 0)
	{
		return "*";
	}
	else
	{
		return inet_ntoa(addr);
	}
}
