/*
 * DNSFormatRule.cc - Formatting rules for DNS packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "DNSFormatRule.h"
#include "NetPacket.h"
#include "Resolver.h"
#include <arpa/nameser.h>

/* DNS name lookup tables extracted from the source for tcpdump */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* Compatibility */
#ifndef T_TXT
#define T_TXT		16		/* text strings */
#endif
#ifndef T_RP
#define T_RP		17		/* responsible person */
#endif
#ifndef T_AFSDB
#define T_AFSDB		18		/* AFS cell database */
#endif
#ifndef T_X25
#define T_X25		19		/* X_25 calling address */
#endif
#ifndef T_ISDN
#define T_ISDN		20		/* ISDN calling address */
#endif
#ifndef T_RT
#define T_RT		21		/* router */
#endif
#ifndef T_NSAP
#define T_NSAP		22		/* NSAP address */
#endif
#ifndef T_NSAP_PTR
#define T_NSAP_PTR	23		/* reverse NSAP lookup (deprecated) */
#endif
#ifndef T_SIG
#define T_SIG		24		/* security signature */
#endif
#ifndef T_KEY
#define T_KEY		25		/* security key */
#endif
#ifndef T_PX
#define T_PX		26		/* X.400 mail mapping */
#endif
#ifndef T_GPOS
#define T_GPOS		27		/* geographical position (withdrawn) */
#endif
#ifndef T_AAAA
#define T_AAAA		28		/* IP6 Address */
#endif
#ifndef T_LOC
#define T_LOC		29		/* Location Information */
#endif

#ifndef T_UNSPEC
#define T_UNSPEC	103		/* Unspecified format (binary data) */
#endif
#ifndef T_UNSPECA
#define T_UNSPECA	104		/* "unspecified ascii". Ugly MIT hack */
#endif

#ifndef C_CHAOS
#define C_CHAOS		3		/* for chaos net (MIT) */
#endif
#ifndef C_HS
#define C_HS		4		/* for Hesiod name server (MIT) (XXX) */
#endif

static char *ns_ops[] = {
	"", " inv_q", " stat", " op3", " notify", " op5", " op6", " op7",
	" op8", " updateA", " updateD", " updateDA",
	" updateM", " updateMA", " zoneInit", " zoneRef",
};

static char *ns_ops_long[] = {
	"Standard Query", "Inverse Query",
	"Nameserver Status Query",
	"op3", "Notify Secondary of SOA Change",
	"op5", "op6", "op7", "op8", "Update A",
	"Update D", "Update DA", "Update M",
	"Update MA",
	"Zone Init", "Zone Ref",
};

static char *ns_resp[] = {
	"", " FormErr", " ServFail", " NXDomain",
	" NotImp", " Refused", " Resp6", " Resp7",
	" Resp8", " Resp9", " Resp10", " Resp11",
	" Resp12", " Resp13", " Resp14", " NoChange",
};

static char *ns_resp_long[] = {
	"No Error", "Format Error",
	"Server Failure",
	"Non Existent Domain",
	"Not Implemented",
	"Query Refused",
	"Resp6", "Resp7", "Resp8", "Resp9", "Resp10", "Resp11",
	"Resp12", "Resp13", "Resp14", "No Change",
};

static struct FormatToken type2str[] = {
	{ T_A,		"A" },
	{ T_NS,		"NS" },
	{ T_MD,		"MD" },
	{ T_MF,		"MF" },
	{ T_CNAME,	"CNAME" },
	{ T_SOA,	"SOA" },
	{ T_MB,		"MB" },
	{ T_MG,		"MG" },
	{ T_MR,		"MR" },
	{ T_NULL,	"NULL" },
	{ T_WKS,	"WKS" },
	{ T_PTR,	"PTR" },
	{ T_HINFO,	"HINFO" },
	{ T_MINFO,	"MINFO" },
	{ T_MX,		"MX" },
	{ T_TXT,	"TXT" },
	{ T_RP,		"RP" },
	{ T_AFSDB,	"AFSDB" },
	{ T_X25,	"X25" },
	{ T_ISDN,	"ISDN" },
	{ T_RT,		"RT" },
	{ T_NSAP,	"NSAP" },
	{ T_NSAP_PTR,	"NSAP_PTR" },
	{ T_SIG,	"SIG" },
	{ T_KEY,	"KEY" },
	{ T_PX,		"PX" },
	{ T_GPOS,	"GPOS" },
	{ T_AAAA,	"AAAA" },
	{ T_LOC ,	"LOC " },
#ifndef T_UINFO
#define T_UINFO 100
#endif
	{ T_UINFO,	"UINFO" },
#ifndef T_UID
#define T_UID 101
#endif
	{ T_UID,	"UID" },
#ifndef T_GID
#define T_GID 102
#endif
	{ T_GID,	"GID" },
	{ T_UNSPEC,	"UNSPEC" },
	{ T_UNSPECA,	"UNSPECA" },
	{ T_AXFR,	"AXFR" },
	{ T_MAILB,	"MAILB" },
	{ T_MAILA,	"MAILA" },
	{ T_ANY,	"ANY" },
	{ 0,		NULL }
};

static struct FormatToken class2str[] = {
	{ C_IN,		"IN" },		/* Not used */
	{ C_CHAOS,	"CHAOS)" },
	{ C_HS,		"HS" },
	{ C_ANY,	"ANY" },
	{ 0,		NULL }
};

DNSFormatRule::DNSFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
}

DNSFormatRule::~DNSFormatRule()
{
}

// Skip a domain name within a network packet.
static void nameSkip(NetPacket& packet)
{
	const u_char *buf = packet.subPacket;
	unsigned len = packet.subPacketLen;
	u_char value;
	if(len > 0)
	{
		value = *buf++;
		--len;
		if((value & INDIR_MASK) == INDIR_MASK)
		{
			if(len > 0)
			{
				++buf;
				--len;
			}
		}
		else
		{
			while(value != 0 && len > 0)
			{
				if(len < (unsigned)value)
				{
					buf += len;
					len = 0;
					break;
				}
				buf += (unsigned)value;
				len -= (unsigned)value;
				value = *buf++;
				--len;
				if((value & INDIR_MASK) == INDIR_MASK)
				{
					if(len > 0)
					{
						++buf;
						--len;
					}
					break;
				}
			}
		}
	}
	packet.setSubPacket(buf, len);
}

// Print a domain name to a short-form output stream.
static void namePrint(NetPacket& packet, FormatOutput& output,
					  const u_char *startPacket)
{
	const u_char *buf = packet.subPacket;
	unsigned i;

	i = (unsigned)(*buf++);
	if(i != 0)
	{
		while(i != 0)
		{
			if((i & INDIR_MASK) == INDIR_MASK)
			{
				buf = startPacket + (((i << 8) | (unsigned)(*buf)) & 0x3FFF);
				i = *buf++;
				continue;
			}
			output.outputShortEscaped((char *)buf, i);
			buf += i;
			output.outputShortString(".");
			i = *buf++;
		}
	}
	else
	{
		output.outputShortString(".");
	}
}

// Prime the resolver's cache with a hostname derived from an IP address.
static void primeResolver(NetPacket& packet, const u_char *startQuestion,
					      const u_char *startPacket)
{
	const u_char *buf = packet.subPacket;
	unsigned i;
	char outbuf[2048];
	char *out;
	char *ipaddr;
	struct in_addr addr;
	unsigned long byte;

	// Copy the hostname from the DNS response packet into "outbuf".
	out = outbuf;
	i = (unsigned)(*buf++);
	if(i != 0)
	{
		while(i != 0)
		{
			if((i & INDIR_MASK) == INDIR_MASK)
			{
				buf = startPacket + (((i << 8) | (unsigned)(*buf)) & 0x3FFF);
				i = *buf++;
				continue;
			}
			if(out != outbuf)
			{
				*out++ = '.';
			}
			memcpy(out, buf, i);
			out += i;
			buf += i;
			i = *buf++;
		}
	}
	*out++ = '\0';

	// Copy the IP address lookup string into "outbuf".
	ipaddr = out;
	buf = startQuestion;
	i = (unsigned)(*buf++);
	if(i != 0)
	{
		while(i != 0)
		{
			if((i & INDIR_MASK) == INDIR_MASK)
			{
				buf = startPacket + (((i << 8) | (unsigned)(*buf)) & 0x3FFF);
				i = *buf++;
				continue;
			}
			if(out != ipaddr)
			{
				*out++ = '.';
			}
			memcpy(out, buf, i);
			out += i;
			buf += i;
			i = *buf++;
		}
	}
	*out = '\0';

	// Convert the IP address string into a real IP address.
	addr.s_addr = 0;
	byte = 0;
	while(*ipaddr != '\0' && *ipaddr != 'i' && *ipaddr != 'I')
	{
		if(*ipaddr == '.')
		{
			addr.s_addr = (addr.s_addr >> 8) | (byte << 24);
			byte = 0;
		}
		else if(*ipaddr >= '0' && *ipaddr <= '9')
		{
			byte = byte * 10 + (unsigned long)(*ipaddr - '0');
		}
		++ipaddr;
	}
	addr.s_addr = htonl(addr.s_addr);

	// Prime the resolver's cache with the results that we obtained.
	ResolverAddToCache(addr, outbuf);
}

void DNSFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	struct HEADER header;
	int typ, i, len;
	const u_char *startPacket = packet.subPacket;
	const u_char *startQuestion;
	unsigned nslen = packet.subPacketLen;
	int isReverseResponse;

	// Extract the DNS header information.
	if(packet.subPacketLen < sizeof(header))
	{
		output.outputShortString("truncated dns packet");
		return;
	}
	memcpy(&header, packet.subPacket, sizeof(header));
	header.id = ntohs(header.id);
	header.qdcount = ntohs(header.qdcount);
	header.ancount = ntohs(header.ancount);
	header.nscount = ntohs(header.nscount);
	header.arcount = ntohs(header.arcount);
	packet.adjustSubPacket(sizeof(header));

	// Is this a query or a response?
	if(header.qr)
	{
		// DNS response.
		output.outputShort("%d%s%s%s%s%s",
						   (int)(header.id),
						   ns_ops[header.opcode],
						   ns_resp[header.rcode],
						   (header.aa ? "*" : ""),
						   (header.ra ? "" : "-"),
						   (header.tc ? "|" : ""));
		if(header.qdcount != 1)
		{
			output.outputShort(" [%dq]", (int)(header.qdcount));
		}
		startQuestion = packet.subPacket;
		nameSkip(packet);
		output.outputShort(" %d/%d/%d",
						   (int)(header.ancount),
						   (int)(header.nscount),
						   (int)(header.arcount));
		if(header.ancount > 0 && packet.subPacketLen >= 4)
		{
			isReverseResponse = (packet.getShortNetwork(0) == T_PTR &&
			                     packet.getShortNetwork(2) == C_IN);
			packet.adjustSubPacket(4);
			do
			{
				nameSkip(packet);
				if(packet.subPacketLen < 10)
				{
					break;
				}
				typ = (int)(packet.getShortNetwork(0));
				i = (int)(packet.getShortNetwork(2));
				len = (int)(packet.getShortNetwork(8));
				packet.adjustSubPacket(10);
				if(i != C_IN)
				{
					output.outputShort
						(" %s", output.getToken(class2str, i, "(Class %d)"));
				}
				output.outputShort
						(" %s", output.getToken(type2str, typ, "Type%d"));
				if(len > (int)(packet.subPacketLen))
				{
					break;
				}
				switch(typ)
				{
					case T_A:
					{
						struct in_addr addr;
						memcpy(&addr, packet.subPacket, 4);
						output.outputShort(" %s", inet_ntoa(addr));
					}
					break;

					case T_AAAA:
					{
						output.outputShort(" %s",
							output.ipaddr6ToString(packet.subPacket,
												   IPADDR6_HEX_ONLY));
					}
					break;

					case T_NS:
					case T_CNAME:
					case T_PTR:
					{
						output.outputShortString(" ");
						namePrint(packet, output, startPacket);
						if(isReverseResponse && typ == T_PTR && i == C_IN)
						{
							// Use this reverse DNS reply to prime the
							// resolver's cache with the answer.
							primeResolver(packet, startQuestion, startPacket);
						}
					}
					break;

					case T_MX:
					{
						packet.adjustSubPacket(2);
						output.outputShortString(" ");
						namePrint(packet, output, startPacket);
						packet.unadjustSubPacket(2);
						output.outputShort(" %d",
										   (int)(packet.getShortNetwork(0)));
					}
					break;

					case T_TXT:
					{
						output.outputShortString(" ");
						output.outputShortEscaped
							((char *)(packet.subPacket + 1),
							 (int)(packet.subPacket[0]));
					}
					break;

					case T_UNSPECA:
					{
						output.outputShort(" %.*s", len,
										   (char *)(packet.subPacket));
					}
					break;
				}
				packet.adjustSubPacket(len);
				--(header.ancount);
				if(header.ancount > 0)
				{
					output.outputShortString(",");
				}
			}
			while(header.ancount > 0);
		}
	}
	else
	{
		// DNS query.
		output.outputShort("%d%s%s",
						   (int)(header.id),
						   ns_ops[header.opcode],
						   (header.rd ? "+" : ""));
		if(header.opcode == IQUERY)
		{
			if(header.qdcount > 0)
			{
				output.outputShort(" [%dq]", (int)(header.qdcount));
			}
			if(header.ancount != 1)
			{
				output.outputShort(" [%da]", (int)(header.ancount));
			}
		}
		else
		{
			if(header.ancount > 0)
			{
				output.outputShort(" [%da]", (int)(header.ancount));
			}
			if(header.qdcount != 1)
			{
				output.outputShort(" [%dq]", (int)(header.qdcount));
			}
		}
		if(header.nscount > 0)
		{
			output.outputShort(" [%dn]", (int)(header.nscount));
		}
		if(header.arcount > 0)
		{
			output.outputShort(" [%dau]", (int)(header.arcount));
		}
		const u_char *savepkt = packet.subPacket;
		unsigned savelen = packet.subPacketLen;
		nameSkip(packet);
		if(packet.subPacketLen >= 4)
		{
			i = packet.getShortNetwork(0);
			output.outputShort(" %s", output.getToken(type2str, i, "Type%d"));
			i = packet.getShortNetwork(2);
			if(i != C_IN)
			{
				output.outputShort(" %s",
						output.getToken(class2str, i, "(Class %d)"));
			}
			output.outputShortString("? ");
			packet.setSubPacket(savepkt, savelen);
			namePrint(packet, output, startPacket);
		}
	}
	output.outputShort(" (%u)", nslen);
}

// Print a domain name to a long-form output stream.
static void namePrintLong(NetPacket& packet, FormatOutput& output,
					      const u_char *startPacket, const char *field)
{
	const u_char *buf = packet.subPacket;
	unsigned i;
	char buffer[4096];
	char *outbuf = buffer;
	int ch;

	i = (unsigned)(*buf++);
	if(i != 0)
	{
		while(i != 0)
		{
			if((i & INDIR_MASK) == INDIR_MASK)
			{
				buf = startPacket + (((i << 8) | (unsigned)(*buf)) & 0x3FFF);
				i = *buf++;
				continue;
			}
			while(i-- > 0)
			{
				ch = ((int)(*buf++)) & 0xFF;
				if((ch & 0x80) != 0)
				{
					*outbuf++ = 'M';
					*outbuf++ = '-';
					ch &= 0x7F;
				}
				if(ch < 0x20 || ch >= 0x7F)
				{
					*outbuf++ = '^';
					ch ^= 0x40;
				}
				*outbuf++ = (char)ch;
			}
			*outbuf++ = '.';
			i = *buf++;
		}
	}
	else
	{
		*outbuf++ = '.';
	}
	*outbuf = '\0';

	// If the name ends in ".in-addr.arpa." or ".ip6.int.", then it
	// is a reverse DNS pointer entry.  In that case, put the
	// forward IP address after the backwards one.
	if(((int)(outbuf - buffer)) >= 14 &&
	   (!strcmp(outbuf - 14, ".in-addr.arpa.") ||
	    !strcmp(outbuf - 14, ".IN-ADDR.ARPA.")))
	{
		int posn, posn2;
		posn = (int)((outbuf - 14) - buffer);
		strcpy(outbuf, " {{");
		outbuf += 3;
		while(posn > 0)
		{
			posn2 = posn;
			while(posn > 0 && buffer[posn - 1] != '.')
			{
				--posn;
			}
			strncpy(outbuf, buffer + posn, posn2 - posn);
			outbuf += posn2 - posn;
			--posn;
			if(posn > 0)
			{
				*outbuf++ = '.';
			}
		}
		strcpy(outbuf, "}}");
	}
	else if(((int)(outbuf - buffer)) >= 9 &&
	        (!strcmp(outbuf - 9, ".ip6.int.") ||
	         !strcmp(outbuf - 9, ".IP6.INT.")))
	{
		int posn, count;
		posn = (int)((outbuf - 9) - buffer);
		strcpy(outbuf, " {{");
		outbuf += 3;
		count = 0;
		while(posn > 0)
		{
			--posn;
			if(buffer[posn] != '.')
			{
				*outbuf++ = buffer[posn];
				++count;
				if(count >= 4 && posn > 0)
				{
					*outbuf++ = ':';
					count = 0;
				}
			}
		}
		strcpy(outbuf, "}}");
	}

	// Send the value to the output stream.
	output.outputLong(field, buffer);
}

#define	YESNO(flag)		((flag) ? "Yes" : "No")

void DNSFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	struct HEADER header;
	const u_char *startPacket = packet.subPacket;
	int typ, i, len;
	unsigned long ttl;

	// Print the DNS section header.
	output.outputSection("Domain Name Service (DNS)");

	// Extract the DNS header information.
	if(packet.subPacketLen < sizeof(header))
	{
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}
	memcpy(&header, packet.subPacket, sizeof(header));
	header.id = ntohs(header.id);
	header.qdcount = ntohs(header.qdcount);
	header.ancount = ntohs(header.ancount);
	header.nscount = ntohs(header.nscount);
	header.arcount = ntohs(header.arcount);
	packet.adjustSubPacket(sizeof(header));

	// Print the DNS header in its entirety.
	output.outputLong("Identifier", "%d", (int)(header.id));
	output.outputLong("Response", "%s", YESNO(header.qr));
	output.outputLong("Opcode", "%d {{%s}}", (int)(header.opcode),
					  ns_ops_long[header.opcode]);
	output.outputLong("Authoritative Answer", "%s", YESNO(header.aa));
	output.outputLong("Truncated Message", "%s", YESNO(header.tc));
	output.outputLong("Recursion Desired", "%s", YESNO(header.rd));
	output.outputLong("Recursion Available", "%s", YESNO(header.ra));
	output.outputLong("Authentic Data", "%s", YESNO(header.ad));
	output.outputLong("Checking Disabled", "%s", YESNO(header.cd));
	output.outputLong("Response Code", "%d {{%s}}", (int)(header.rcode),
					  ns_resp_long[header.rcode]);
	output.outputLong("Question Count", "%d", (int)(header.qdcount));
	output.outputLong("Answer Count", "%d", (int)(header.ancount));
	output.outputLong("Authority Count", "%d", (int)(header.nscount));
	output.outputLong("Additional Record Count", "%d", (int)(header.arcount));

	// Print the question blocks.
	const u_char *savepkt;
	unsigned savelen;
	const u_char *savepkt2;
	unsigned savelen2;
	const char *name;
	while(header.qdcount > 0 && packet.subPacketLen >= 4)
	{
		output.outputSection("DNS Question Entry");
		savepkt = packet.subPacket;
		savelen = packet.subPacketLen;
		nameSkip(packet);
		typ = packet.getShortNetwork(0);
		i = packet.getShortNetwork(2);
		name = output.getToken(type2str, typ);
		if(name)
			output.outputLong("Type", "%d {{%s}}", typ, name);
		else
			output.outputLong("Type", "%d", typ);
		name = output.getToken(class2str, i);
		if(name)
			output.outputLong("Class", "%d {{%s}}", i, name);
		else
			output.outputLong("Class", "%d", i);
		packet.adjustSubPacket(4);
		savepkt2 = packet.subPacket;
		savelen2 = packet.subPacketLen;
		packet.setSubPacket(savepkt, savelen);
		namePrintLong(packet, output, startPacket, "Name");
		packet.setSubPacket(savepkt2, savelen2);
		--(header.qdcount);
	}

	// Print the answer blocks.
	while(header.ancount > 0 && packet.subPacketLen >= 10)
	{
		output.outputSection("DNS Answer Entry");
		nameSkip(packet);
		typ = packet.getShortNetwork(0);
		i = packet.getShortNetwork(2);
		ttl = packet.getLongNetwork(4);
		len = packet.getShortNetwork(8);
		packet.adjustSubPacket(10);
		name = output.getToken(type2str, typ);
		if(name)
			output.outputLong("Type", "%d {{%s}}", typ, name);
		else
			output.outputLong("Type", "%d", typ);
		name = output.getToken(class2str, i);
		if(name)
			output.outputLong("Class", "%d {{%s}}", i, name);
		else
			output.outputLong("Class", "%d", i);
		output.outputLong("Time to Live", "%lu", ttl);
		output.outputLong("Length", "%d", len);
		switch(typ)
		{
			case T_A:
			{
				struct in_addr addr;
				memcpy(&addr, packet.subPacket, 4);
				output.outputLong("Address", inet_ntoa(addr));
			}
			break;

			case T_AAAA:
			{
				output.outputLong("Address",
								  output.ipaddr6ToString(packet.subPacket,
								  						 IPADDR6_HEX_ONLY));
			}
			break;

			case T_NS:
			case T_CNAME:
			case T_PTR:
			{
				namePrintLong(packet, output, startPacket, "Address");
			}
			break;

			case T_MX:
			{
				packet.adjustSubPacket(2);
				namePrintLong(packet, output, startPacket, "Address");
				packet.unadjustSubPacket(2);
				output.outputLong("Preference", "%d",
								  (int)(packet.getShortNetwork(0)));
			}
			break;

			case T_UNSPECA:
			{
				output.outputLong("Address", "%.*s", len,
								  (char *)(packet.subPacket));
			}
			break;

			default:
			{
				output.outputHex("Address", packet.subPacket, len);
			}
			break;
		}
		packet.adjustSubPacket(len);
		--(header.ancount);
	}

	// Print the authority blocks.
	while(header.nscount > 0 && packet.subPacketLen >= 10)
	{
		output.outputSection("DNS Authority Entry");
		nameSkip(packet);
		typ = packet.getShortNetwork(0);
		i = packet.getShortNetwork(2);
		ttl = packet.getLongNetwork(4);
		len = packet.getShortNetwork(8);
		packet.adjustSubPacket(10);
		name = output.getToken(type2str, typ);
		if(name)
			output.outputLong("Type", "%d {{%s}}", typ, name);
		else
			output.outputLong("Type", "%d", typ);
		name = output.getToken(class2str, i);
		if(name)
			output.outputLong("Class", "%d {{%s}}", i, name);
		else
			output.outputLong("Class", "%d", i);
		output.outputLong("Time to Live", "%lu", ttl);
		output.outputLong("Length", "%d", len);
		switch(typ)
		{
			case T_A:
			{
				struct in_addr addr;
				memcpy(&addr, packet.subPacket, 4);
				output.outputLong("Address", inet_ntoa(addr));
			}
			break;

			case T_AAAA:
			{
				output.outputLong("Address",
								  output.ipaddr6ToString(packet.subPacket,
								  						 IPADDR6_HEX_ONLY));
			}
			break;

			case T_NS:
			case T_CNAME:
			case T_PTR:
			{
				namePrintLong(packet, output, startPacket, "Address");
			}
			break;

			case T_MX:
			{
				packet.adjustSubPacket(2);
				namePrintLong(packet, output, startPacket, "Address");
				packet.unadjustSubPacket(2);
				output.outputLong("Preference", "%d",
								  (int)(packet.getShortNetwork(0)));
			}
			break;

			case T_UNSPECA:
			{
				output.outputLong("Address", "%.*s", len,
								  (char *)(packet.subPacket));
			}
			break;

			default:
			{
				output.outputHex("Address", packet.subPacket, len);
			}
			break;
		}
		packet.adjustSubPacket(len);
		--(header.nscount);
	}

	// Print the resource blocks.
	while(header.arcount > 0 && packet.subPacketLen >= 10)
	{
		output.outputSection("DNS Additional Record Entry");
		nameSkip(packet);
		typ = packet.getShortNetwork(0);
		i = packet.getShortNetwork(2);
		ttl = packet.getLongNetwork(4);
		len = packet.getShortNetwork(8);
		packet.adjustSubPacket(10);
		name = output.getToken(type2str, typ);
		if(name)
			output.outputLong("Type", "%d {{%s}}", typ, name);
		else
			output.outputLong("Type", "%d", typ);
		name = output.getToken(class2str, i);
		if(name)
			output.outputLong("Class", "%d {{%s}}", i, name);
		else
			output.outputLong("Class", "%d", i);
		output.outputLong("Time to Live", "%lu", ttl);
		output.outputLong("Length", "%d", len);
		switch(typ)
		{
			case T_A:
			{
				struct in_addr addr;
				memcpy(&addr, packet.subPacket, 4);
				output.outputLong("Address", inet_ntoa(addr));
			}
			break;

			case T_AAAA:
			{
				output.outputLong("Address",
								  output.ipaddr6ToString(packet.subPacket,
								  						 IPADDR6_HEX_ONLY));
			}
			break;

			case T_NS:
			case T_CNAME:
			case T_PTR:
			{
				namePrintLong(packet, output, startPacket, "Address");
			}
			break;

			case T_MX:
			{
				packet.adjustSubPacket(2);
				namePrintLong(packet, output, startPacket, "Address");
				packet.unadjustSubPacket(2);
				output.outputLong("Preference", "%d",
								  (int)(packet.getShortNetwork(0)));
			}
			break;

			case T_UNSPECA:
			{
				output.outputLong("Address", "%.*s", len,
								  (char *)(packet.subPacket));
			}
			break;

			default:
			{
				output.outputHex("Address", packet.subPacket, len);
			}
			break;
		}
		packet.adjustSubPacket(len);
		--(header.arcount);
	}
}
