/*
 * DialogBox.cc - Base class for dialog box handling.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gnome.h>
#include "DialogBox.h"

DialogBox::DialogBox(GtkWidget *parent,
			  		 const char *title,
			  		 int stdbuttons,
			  		 int modal)
{
	GtkWidget *frame;
	GtkWidget *outervbox;
	GtkWidget *sep;
	GtkWidget *button;

	// Construct the basic dialog layout.
	defaultbutton = 0;
	dialog = gtk_window_new(GTK_WINDOW_DIALOG);
	frame = gtk_frame_new(NULL);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_OUT);
	gtk_container_add(GTK_CONTAINER(dialog), frame);
	gtk_widget_show(frame);
	outervbox = gtk_vbox_new(FALSE, GNOME_PAD);
	gtk_container_set_border_width(GTK_CONTAINER(outervbox), GNOME_PAD_SMALL);
	gtk_container_add(GTK_CONTAINER(frame), outervbox);
	gtk_widget_show(outervbox);
	gtk_window_set_policy(GTK_WINDOW(dialog), FALSE, FALSE, FALSE);
	vbox = gtk_vbox_new(FALSE, GNOME_PAD);
	gtk_box_pack_start(GTK_BOX(outervbox), vbox, TRUE, TRUE, GNOME_PAD_SMALL);
	gtk_widget_show(vbox);
	buttons = gtk_hbutton_box_new();
	gtk_button_box_set_layout(GTK_BUTTON_BOX(buttons),
							  gnome_preferences_get_button_layout());
	gtk_button_box_set_spacing(GTK_BUTTON_BOX(buttons), GNOME_PAD);
	gtk_box_pack_end(GTK_BOX(vbox), buttons, FALSE, TRUE, 0);
	gtk_widget_show(buttons);
	sep = gtk_hseparator_new();
	gtk_box_pack_end(GTK_BOX(vbox), sep, FALSE, TRUE, GNOME_PAD_SMALL);
	gtk_widget_show(sep);

	// Construct standard buttons.
	if((stdbuttons & DIALOG_BUTTON_OK) != 0)
	{
		button = gnome_stock_or_ordinary_button(GNOME_STOCK_BUTTON_OK);
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(buttons), button, TRUE, TRUE, 0);
		gtk_widget_grab_default(button);
		defaultbutton = button;
		gtk_widget_show(button);
		gtk_signal_connect(GTK_OBJECT(button), "clicked",
						   GTK_SIGNAL_FUNC(cb_dialog_ok), this);
		okbutton = button;
		insert = 1;
	}
	else
	{
		okbutton = 0;
		insert = 0;
	}
	if((stdbuttons & DIALOG_BUTTON_CANCEL) != 0)
	{
		button = gnome_stock_or_ordinary_button(GNOME_STOCK_BUTTON_CANCEL);
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
		gtk_box_pack_end(GTK_BOX(buttons), button, TRUE, TRUE, 0);
		if((stdbuttons & DIALOG_BUTTON_OK) == 0)
		{
			gtk_widget_grab_default(button);
			defaultbutton = button;
		}
		gtk_widget_show(button);
		gtk_signal_connect(GTK_OBJECT(button), "clicked",
						   GTK_SIGNAL_FUNC(cb_dialog_cancel), this);
		cancelbutton = button;
	}
	else
	{
		cancelbutton = 0;
	}
	if((stdbuttons & DIALOG_BUTTON_CLOSE) != 0)
	{
		button = gnome_stock_or_ordinary_button(GNOME_STOCK_BUTTON_CLOSE);
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
		gtk_box_pack_end(GTK_BOX(buttons), button, TRUE, TRUE, 0);
		if((stdbuttons & (DIALOG_BUTTON_OK | DIALOG_BUTTON_CANCEL)) == 0)
		{
			gtk_widget_grab_default(button);
			defaultbutton = button;
		}
		gtk_widget_show(button);
		gtk_signal_connect(GTK_OBJECT(button), "clicked",
						   GTK_SIGNAL_FUNC(cb_dialog_close), this);
		closebutton = button;
	}
	else
	{
		closebutton = 0;
	}

	// Trap the window close button on the title bar.
	gtk_signal_connect(GTK_OBJECT(dialog), "delete_event",
					   GTK_SIGNAL_FUNC(cb_dialog_wmclose), this);

	// Trap key events on the dialog to handle the Escape key.
	gtk_signal_connect(GTK_OBJECT(dialog), "key_press_event",
					   GTK_SIGNAL_FUNC(cb_dialog_key_pressed), this);

	// Set the title if necessary.
	if(title)
	{
		gtk_window_set_title(GTK_WINDOW(dialog), title);
	}

	// Set the dialog's parent if necessary.
	if(parent)
	{
		gtk_window_set_transient_for(GTK_WINDOW(dialog), GTK_WINDOW(parent));
	}

	// Set the modality property if necessary.
	if(modal)
	{
		gtk_window_set_modal(GTK_WINDOW(dialog), TRUE);
	}
}

DialogBox::~DialogBox(void)
{
	if(dialog != 0)
	{
		gtk_widget_destroy(dialog);
	}
}

void DialogBox::cb_dialog_ok(GtkWidget *widget, gpointer data)
{
	((DialogBox *)data)->ok();
}

void DialogBox::cb_dialog_cancel(GtkWidget *widget, gpointer data)
{
	((DialogBox *)data)->cancel();
}

void DialogBox::cb_dialog_close(GtkWidget *widget, gpointer data)
{
	((DialogBox *)data)->close();
}

void DialogBox::cb_dialog_wmclose(GtkWidget *widget, GdkEventAny *event,
								  gpointer data)
{
	((DialogBox *)data)->wmclose();
}

gint DialogBox::cb_dialog_key_pressed(GtkWidget *widget, GdkEventKey *event,
									  gpointer data)
{
	if(event->keyval == GDK_Escape)
	{
		((DialogBox *)data)->wmclose();
		return TRUE;
	}
	return FALSE;
}

void DialogBox::cb_dialog_button(GtkWidget *widget, gpointer data)
{
	((DialogBox *)data)->button(widget);
}

void DialogBox::cb_dialog_activate(GtkWidget *widget, gpointer data)
{
	((DialogBox *)data)->activateDefaultButton();
}

void DialogBox::ok(void)
{
	closeDialog();
}

void DialogBox::cancel(void)
{
	closeDialog();
}

void DialogBox::close(void)
{
	closeDialog();
}

void DialogBox::wmclose(void)
{
	cancel();
}

void DialogBox::button(GtkWidget *buttonWidget)
{
	closeDialog();
}

void DialogBox::activateDefaultButton(void)
{
	if(defaultbutton && GTK_WIDGET_SENSITIVE(defaultbutton))
	{
		if(defaultbutton == okbutton)
		{
			ok();
		}
		else if(defaultbutton == cancelbutton)
		{
			cancel();
		}
		else if(defaultbutton == closebutton)
		{
			close();
		}
		else
		{
			button(defaultbutton);
		}
	}
}

GtkWidget *DialogBox::addButton(const char *name, int isDefault)
{
	GtkWidget *button;
	button = gnome_stock_or_ordinary_button(name);
	GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(buttons), button, TRUE, TRUE, 0);
	if(isDefault)
	{
		gtk_widget_grab_default(button);
		defaultbutton = button;
	}
	gtk_box_reorder_child(GTK_BOX(buttons), button, insert);
	++insert;
	gtk_widget_show(button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
					   GTK_SIGNAL_FUNC(cb_dialog_button), this);
	return button;
}

GtkWidget *DialogBox::addText(int editable)
{
	// Construct the text entry widget and make it editable.
	GtkWidget *widget = gtk_entry_new();
	gtk_entry_set_editable(GTK_ENTRY(widget), editable);

	// Hook the activate signal through to the default button so
	// that hitting the Return/Enter key will cause the default
	// button to be activated.  Why GTK doesn't do this by default
	// I have no idea.
	gtk_signal_connect(GTK_OBJECT(widget), "activate",
					   GTK_SIGNAL_FUNC(cb_dialog_activate), this);

	// Return the text entry widget to the caller.
	return widget;
}

void DialogBox::show(int showcontents)
{
	if(dialog != 0)
	{
		if(showcontents)
			gtk_widget_show_all(vbox);
		gtk_widget_show(dialog);
	}
}

void DialogBox::hide(void)
{
	if(dialog != 0)
	{
		gtk_widget_hide(dialog);
	}
}

void DialogBox::closeDialog(void)
{
	if(dialog != 0)
	{
		gtk_widget_destroy(dialog);
		dialog = 0;
	}
	delete this;
}

int DialogBox::textEmpty(GtkWidget *widget)
{
	gchar *text = gtk_entry_get_text(GTK_ENTRY(widget));
	if(text)
	{
		while(*text != '\0' && (*text == ' ' || *text == '\t'))
		{
			++text;
		}
		if(*text == '\0')
		{
			return 1;
		}
		else
		{
			return 0;
		}
	}
	else
	{
		return 1;
	}
}
