/*
 * FilterCondition.cc - Filter condition handling.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetIncludes.h"
#include "FilterCondition.h"

FilterCondition::FilterCondition()
{
	addrType = FilterAddressType_None;
	addrTo = 0;
	addrFrom = 0;
	addrToOrFrom = 0;
	addrBetween = 0;
	addrAnd = 0;
	protocols = FilterProtocolType_All;
	portEnabled = 0;
	ports = 0;
}

FilterCondition::~FilterCondition()
{
	if(addrTo)
		free(addrTo);
	if(addrFrom)
		free(addrFrom);
	if(addrToOrFrom)
		free(addrToOrFrom);
	if(addrBetween)
		free(addrBetween);
	if(addrAnd)
		free(addrAnd);
	if(ports)
		free(ports);
}

static void setString(char **ptr, const char *value)
{
	if(*ptr)
	{
		free(*ptr);
		*ptr = 0;
	}
	if(value && *value != '\0')
	{
		if((*ptr = (char *)malloc(strlen(value) + 1)) != 0)
		{
			strcpy(*ptr, value);
		}
	}
}

void FilterCondition::setAddrTo(const char *value)
{
	setString(&addrTo, value);
}

void FilterCondition::setAddrFrom(const char *value)
{
	setString(&addrFrom, value);
}

void FilterCondition::setAddrToOrFrom(const char *value)
{
	setString(&addrToOrFrom, value);
}

void FilterCondition::setAddrBetween(const char *value)
{
	setString(&addrBetween, value);
}

void FilterCondition::setAddrAnd(const char *value)
{
	setString(&addrAnd, value);
}

void FilterCondition::setPorts(const char *value)
{
	setString(&ports, value);
}

static void appendString(char **str, const char *value, int len=-1)
{
	int slen = (*str ? strlen(*str) : 0);
	char *newstr;
	if(len < 0)
	{
		len = strlen(value);
	}
	if((newstr = (char *)realloc(*str, slen + len + 1)) != 0)
	{
		*str = newstr;
		if(len > 0)
		{
			memcpy(newstr + slen, value, len);
		}
		newstr[slen + len] = '\0';
	}
}

static void appendHost(char **str, const char *prefix, const char *host)
{
	int len;
	int first = 1;
	appendString(str, "(");
	while(*host != '\0')
	{
		while(*host == ' ' || *host == ',')
			++host;
		len = 0;
		while(host[len] != '\0' && host[len] != ' ' && host[len] != ',')
			++len;
		if(len > 0)
		{
			if(!first)
			{
				appendString(str, " or ");
			}
			else
			{
				first = 0;
			}
			appendString(str, prefix);
			appendString(str, host, len);
		}
		host += len;
	}
	appendString(str, ")");
}

static int hostEmpty(const char *host)
{
	if(!host)
	{
		return 1;
	}
	while(*host != '\0' && (*host == ' ' || *host == ','))
	{
		++host;
	}
	return (*host == '\0');
}

char *FilterCondition::convert(int linkType)
{
	char *str = 0;
	const char *value;
	int len;
	int first;
	int needAnd = 0;
	int needEtherTrail = 0;

	// Check for the easy case of "don't filter at all".
	if(addrType == FilterAddressType_None &&
	   protocols == FilterProtocolType_All &&
	   !portEnabled)
	{
		return 0;
	}

	// Do we need to add Ethernet matching conditions for this link type?
	if(linkType == DLT_EN10MB || linkType == DLT_IEEE802)
	{
		if(!hasProtocol(FilterProtocolType_OtherEthernet))
		{
			// Additive combination of ARP and RARP.
			if(hasProtocol(FilterProtocolType_ARP) &&
			   !hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "arp or ");
			}
			else if(!hasProtocol(FilterProtocolType_ARP) &&
			        hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "rarp or ");
			}
			else if(hasProtocol(FilterProtocolType_ARP) &&
			        hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "arp or rarp or ");
			}
		}
		else
		{
			// Subtractive combination of ARP and RARP.
			if(hasProtocol(FilterProtocolType_ARP) &&
			   !hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "(not rarp and not ip) or ");
			}
			else if(!hasProtocol(FilterProtocolType_ARP) &&
			        hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "(not arp and not ip) or ");
			}
			else if(!hasProtocol(FilterProtocolType_ARP) &&
			        !hasProtocol(FilterProtocolType_RARP))
			{
				appendString(&str, "(not arp and not rarp and not ip) or ");
			}
			else
			{
				appendString(&str, "(not ip) or ");
			}
		}
		appendString(&str, "(ip");
		needAnd = 1;
		needEtherTrail = 1;
	}

	// Make the IP address part of the filter.
	switch(addrType)
	{
		case FilterAddressType_None:		break;

		case FilterAddressType_To:
		{
			if(!hostEmpty(addrTo))
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendHost(&str, "dst host ", addrTo);
				needAnd = 1;
			}
		}
		break;

		case FilterAddressType_From:
		{
			if(!hostEmpty(addrFrom))
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendHost(&str, "src host ", addrFrom);
				needAnd = 1;
			}
		}
		break;

		case FilterAddressType_ToOrFrom:
		{
			if(!hostEmpty(addrToOrFrom))
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendHost(&str, "host ", addrToOrFrom);
				needAnd = 1;
			}
		}
		break;

		case FilterAddressType_Between:
		{
			int firstempty = hostEmpty(addrBetween);
			int secondempty = hostEmpty(addrAnd);
			if(!firstempty && !secondempty)
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendString(&str, "((");
				appendHost(&str, "src host ", addrBetween);
				appendString(&str, " and ");
				appendHost(&str, "dst host ", addrAnd);
				appendString(&str, ") or (");
				appendHost(&str, "dst host ", addrBetween);
				appendString(&str, " and ");
				appendHost(&str, "src host ", addrAnd);
				appendString(&str, "))");
				needAnd = 1;
			}
			else if(firstempty && !secondempty)
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendHost(&str, "host ", addrAnd);
				needAnd = 1;
			}
			else if(secondempty && !firstempty)
			{
				if(needAnd)
				{
					appendString(&str, " and ");
				}
				appendHost(&str, "host ", addrBetween);
				needAnd = 1;
			}
		}
		break;
	}

	// Make the IP protocol part of the filter.  Note: it is possible that
	// all of the protocols could be de-selected.  In that case, we put
	// in a fake test that will always fail.  Thus, all packets will
	// be rejected by the filter.
	if(!hasProtocol(FilterProtocolType_UDP) &&
	   !hasProtocol(FilterProtocolType_TCP) &&
	   !hasProtocol(FilterProtocolType_ICMP) &&
	   !hasProtocol(FilterProtocolType_OtherIP))
	{
		// Reject all IP packets by searching for those that have
		// the protocol field set to 253, which shouldn't occur
		// in practice.
		if(needAnd)
		{
			appendString(&str, " and ip proto 253");
		}
		else
		{
			appendString(&str, "ip proto 253");
			needAnd = 1;
		}
	}
	else if(!hasProtocol(FilterProtocolType_OtherIP))
	{
		// Additive combination of UDP, TCP, and ICMP.
		if(needAnd)
		{
			appendString(&str, " and ");
		}
		first = 1;
		appendString(&str, "(");
		if(hasProtocol(FilterProtocolType_UDP))
		{
			appendString(&str, "udp");
			first = 0;
		}
		if(hasProtocol(FilterProtocolType_TCP))
		{
			if(!first)
			{
				appendString(&str, " or ");
			}
			appendString(&str, "tcp");
			first = 0;
		}
		if(hasProtocol(FilterProtocolType_ICMP))
		{
			if(!first)
			{
				appendString(&str, " or ");
			}
			appendString(&str, "icmp");
		}
		appendString(&str, ")");
		needAnd = 1;
	}
	else if(!hasProtocol(FilterProtocolType_UDP) ||
	        !hasProtocol(FilterProtocolType_TCP) ||
	        !hasProtocol(FilterProtocolType_ICMP))
	{
		// Subtractive combination of UDP, TCP, and ICMP.
		if(needAnd)
		{
			appendString(&str, " and ");
		}
		first = 1;
		appendString(&str, "not (");
		if(!hasProtocol(FilterProtocolType_UDP))
		{
			appendString(&str, "udp");
			first = 0;
		}
		if(!hasProtocol(FilterProtocolType_TCP))
		{
			if(!first)
			{
				appendString(&str, " or ");
			}
			appendString(&str, "tcp");
			first = 0;
		}
		if(!hasProtocol(FilterProtocolType_ICMP))
		{
			if(!first)
			{
				appendString(&str, " or ");
			}
			appendString(&str, "icmp");
		}
		appendString(&str, ")");
		needAnd = 1;
	}

	// Make the IP port part of the filter.
	if(portEnabled && ports != 0)
	{
		first = 1;
		value = ports;
		while(*value != '\0')
		{
			while(*value == ' ' || *value == ',')
				++value;
			len = 0;
			while(value[len] != '\0' && value[len] != ' ' && value[len] != ',')
				++len;
			if(len > 0)
			{
				if(first)
				{
					if(needAnd)
					{
						appendString(&str, " and ");
					}
					appendString(&str, "(");
					first = 0;
				}
				else
				{
					appendString(&str, " or ");
				}
				appendString(&str, "port ");
				appendString(&str, value, len);
				value += len;
			}
		}
		if(!first)
		{
			appendString(&str, ")");
			needAnd = 1;
		}
	}

	// Close off the Ethernet-level matching conditions.
	if(needEtherTrail)
	{
		appendString(&str, ")");
	}

	// Done.
	return str;
}
