/*
 * FormatRule.cc - Base class for packet formating rules.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetPacket.h"
#include "FormatRule.h"
#include "LinkFormatRule.h"
#include "IPFormatRule.h"
#include "ARPFormatRule.h"
#include "Resolver.h"

static FormatToken commonPorts[] = {
	{7,			"Echo"},
	{9,			"Discard"},
	{11,		"System Status"},
	{13,		"Time of Day (ASCII)"},
	{18,		"Message Send Protocol"},
	{19,		"Character Generator"},
	{22,		"Secure Shell Login Protocol (SSH)"},
	{37,		"Time of Day (Binary)"},
	{50,		"Remote Mail Checking Protocol"},
	{53,		"Domain Name Service (DNS)"},
	{67,		"Dynamic Host Configuration Protocol Server (DHCP/BOOTP)"},
	{68,		"Dynamic Host Configuration Protocol Client (DHCP/BOOTP)"},
	{70,		"Gopher"},
	{80,		"Hypertext Transfer Protocol (HTTP)"},
	{88,		"Kerberos"},
	{109,		"Post Office Protocol version 2 (POP2)"},
	{110,		"Post Office Protocol version 3 (POP3)"},
	{111,		"RPC Portmapper"},
	{123,		"Network Time Protocol (NTP)"},
	{137,		"NetBIOS Name Service"},
	{138,		"NetBIOS Datagram Service"},
	{139,		"NetBIOS Session Service"},
	{143,		"Internet Message Access Protocol version 2 (IMAP2)"},
	{177,		"X Display Manager Control Protocol (XDMCP)"},
	{179,		"Border Gateway Protocol (BGP)"},
	{194,		"Internet Relay Chat (IRC)"},
	{210,		"Wide Area Information Service (WAIS)"},
	{220,		"Internet Message Access Protocol version 3 (IMAP3)"},
	{443,		"Secure Hypertext Transfer Protocol (HTTPS)"},
	{444,		"Simple Network Paging Protocol (SNPP)"},
	{487,		"Simple Asynchronous File Transfer (SAFT)"},
	{1080,		"SOCKS Proxy Server"},
	{2049,		"Network File System (NFS)"},
	{6667,		"Internet Relay Chat - Old-Style"},
	{0,			NULL}};

static FormatToken tcpPorts[] = {
	{15,		"Network Statistics"},
	{17,		"Quote of the Day"},
	{21,		"File Transfer Protocol (FTP)"},
	{23,		"Telnet"},
	{25,		"Simple Mail Transport Protocol (SMTP)"},
	{42,		"IEN 116 Nameserver"},
	{43,		"Whois"},
	{79,		"Finger"},
	{119,		"Network News Transfer Protocol (NNTP)"},
	{465,		"Secure Simple Mail Transport Protocol (SSMTP)"},
	{563,		"Secure Network News Transfer Protocol (SNNTP)"},
	{636,		"Secure Lightweight Directory Access Protocol (SLDAP)"},
	{993,		"Secure Internet Message Access Protocol (SIMAP)"},
	{995,		"Secure Post Office Protocol version 3 (SPOP3)"},
	{0,			NULL}};

static FormatToken udpPorts[] = {
	{39,		"Resource Location Protocol (RLP)"},
	{69,		"Trivial File Transfer Protocol (TFTP)"},
	{161,		"Simple Network Management Protocol (SNMP)"},
	{162,		"Simple Network Management Trap Protocol"},
	{0,			NULL}};

FormatRuleTable::FormatRuleTable()
{
	first = 0;
	last = 0;
	defrule = 0;
}

FormatRuleTable::~FormatRuleTable()
{
	FormatRule *current;
	FormatRule *next;
	current = first;
	while(current != 0)
	{
		next = current->next;
		delete current;
		current = next;
	}
	if(defrule)
	{
		delete defrule;
	}
}

void FormatRuleTable::addRule(FormatRule *rule)
{
	rule->next = 0;
	if(last)
	{
		last->next = rule;
	}
	else
	{
		first = rule;
	}
	last = rule;
}

void FormatRuleTable::addDefaultRule(FormatRule *rule)
{
	defrule = rule;
}

int FormatRuleTable::formatPacketShort(NetPacket& packet, FormatOutput& output,
						  			   unsigned long tag)
{
	FormatRule *current = first;
	while(current != 0)
	{
		if(current->tag == tag)
		{
			current->formatPacketShort(packet, output);
			return 1;
		}
		current = current->next;
	}
	if(defrule)
	{
		defrule->formatPacketShort(packet, output);
		return 1;
	}
	return 0;
}

int FormatRuleTable::formatPacketLong(NetPacket& packet, FormatOutput& output,
						 			  unsigned long tag)
{
	FormatRule *current = first;
	while(current != 0)
	{
		if(current->tag == tag)
		{
			current->formatPacketLong(packet, output);
			return 1;
		}
		current = current->next;
	}
	if(defrule)
	{
		defrule->formatPacketLong(packet, output);
		return 1;
	}
	return 0;
}

void FormatRuleTable::clearState(void)
{
	FormatRule *current;
	current = first;
	while(current != 0)
	{
		current->clearState();
		current = current->next;
	}
	if(defrule)
	{
		defrule->clearState();
	}
}

FormatRule::FormatRule(FormatRuleTable *_table, unsigned long _tag)
{
	table = _table;
	tag = _tag;
	table->addRule(this);
}

FormatRule::FormatRule(FormatRuleTable *_table)
{
	table = _table;
	tag = 0;
	table->addDefaultRule(this);
}

FormatRule::~FormatRule()
{
}

void FormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
}

void FormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
}

void FormatRule::clearState(void)
{
}

FormatOutput::FormatOutput()
{
	ipBuf = ipBuf1;
	etherBuf = etherBuf1;
}

FormatOutput::~FormatOutput()
{
}

void FormatOutput::outputShortEscaped(const char *string, int len)
{
	int outlen;
	int ch;
	if(len < 0)
	{
		len = strlen(string);
	}
	outlen = 0;
	while(len > 0)
	{
		int ch = ((*string++) & 0xFF);
		if((ch & 0x80) != 0)
		{
			etherBuf[outlen++] = 'M';
			etherBuf[outlen++] = '-';
			ch &= 0x7F;
		}
		if(ch < 0x20 || ch >= 0x7F)
		{
			etherBuf[outlen++] = '^';
			ch ^= 0x40;
		}
		etherBuf[outlen++] = (char)ch;
		if(outlen >= 50)
		{
			etherBuf[outlen] = '\0';
			outputShortString(etherBuf);
			outlen = 0;
		}
		--len;
	}
	if(outlen > 0)
	{
		etherBuf[outlen] = '\0';
		outputShortString(etherBuf);
	}
}

char *FormatOutput::ipaddrToString(const void *addr)
{
	// Get the buffer to use and then switch buffers.  We need to do
	// this because some callers need to format two IP addresses
	// at the same time.
	char *buf = ipBuf;
	if(buf == ipBuf1)
		ipBuf = ipBuf2;
	else
		ipBuf = ipBuf1;

	// Do we need to resolve the address?
	if(resolveIPAddresses())
	{
		struct in_addr iaddr;
		memcpy(&iaddr, addr, 4);
		const char *name = ResolverLookup(iaddr, !(fullyQualifiedNames()));
		if(name)
		{
			return (char *)name;
		}
		else
		{
			// Record this IP address for later, so that we can
			// "back-patch" the packet when the DNS reply arrives.
			markIPAddress(addr);
		}
	}

	// Format the address.
	struct in_addr inaddr;
	memcpy(&inaddr, addr, sizeof(inaddr));
	strcpy(buf, inet_ntoa(inaddr));
	return buf;
}

char *FormatOutput::ipaddrToStringNoResolve(const void *addr)
{
	// Get the buffer to use and then switch buffers.  We need to do
	// this because some callers need to format two IP addresses
	// at the same time.
	char *buf = ipBuf;
	if(buf == ipBuf1)
		ipBuf = ipBuf2;
	else
		ipBuf = ipBuf1;

	// Format the address.
	struct in_addr inaddr;
	memcpy(&inaddr, addr, sizeof(inaddr));
	strcpy(buf, inet_ntoa(inaddr));
	return buf;
}

char *FormatOutput::ipaddrToDomainName(const void *addr)
{
	// Do we need to resolve the address?
	if(resolveIPAddresses())
	{
		struct in_addr iaddr;
		memcpy(&iaddr, addr, 4);
		const char *name = ResolverLookup(iaddr, !(fullyQualifiedNames()));
		if(name)
		{
			return (char *)name;
		}
		else
		{
			// Record this IP address for later, so that we can
			// "back-patch" the packet when the DNS reply arrives.
			markIPAddress(addr);
		}
	}

	// Address could not be resolved.
	return (char *)0;
}

char *FormatOutput::ipaddr6ToString(const void *_addr, int type)
{
	const unsigned char *addr = (const unsigned char *)_addr;

	// Get the buffer to use and then switch buffers.  We need to do
	// this because some callers need to format two IP addresses
	// at the same time.
	char *buf = ipBuf;
	if(buf == ipBuf1)
		ipBuf = ipBuf2;
	else
		ipBuf = ipBuf1;

	// Check for special cases.
	if(type != IPADDR6_HEX_ONLY)
	{
		if(!memcmp(addr,
		           "\x00\x00\x00\x00\x00\x00\x00\x00"
			       "\x00\x00\x00\x00\x00\x00\x00\x01", 16))
		{
			// IPv6 loopback address.
			return "ipv6-localhost";
		}
		if(!memcmp(addr,
		           "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xFF\xFF", 12) ||
		   (!memcmp(addr,
		            "\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00", 12) &&
			memcmp(addr + 12, "\x00\x00\x00\x00", 4) != 0))
		{
			// IPv4 address encoded into an IPv6 address.
			struct in_addr ip4addr;
			memcpy(&ip4addr, addr + 12, 4);
			if(type == IPADDR6_NAME_ONLY)
			{
				strcpy(buf, inet_ntoa(ip4addr));
			}
			else if(addr[11] == (unsigned char)0xFF)
			{
				strcpy(buf, ":ffff:");
				strcat(buf, inet_ntoa(ip4addr));
			}
			else
			{
				buf[0] = ':';
				strcpy(buf + 1, inet_ntoa(ip4addr));
			}
			return buf;
		}
		if(type == IPADDR6_NAME_ONLY)
		{
			if(addr[0] == (unsigned char)0xFE &&
			   (addr[1] & (unsigned char)0xC0) == (unsigned char)0x80)
			{
				// Link-local address.
				return "(ipv6 link local)";
			}
			if(addr[0] == (unsigned char)0xFE &&
			   (addr[1] & (unsigned char)0xC0) == (unsigned char)0xC0)
			{
				// Site-local address.
				return "(ipv6 site local)";
			}
			if(addr[0] == (unsigned char)0xFF)
			{
				// Multicast address.
				return "(ipv6 multicast)";
			}
			return (char *)0;
		}
	}

	// Format the address into the buffer.
	sprintf(buf, "%x:%x:%x:%x:%x:%x:%x:%x",
	        (((int)(addr[0])) << 8) | ((int)(addr[1])),
	        (((int)(addr[2])) << 8) | ((int)(addr[3])),
	        (((int)(addr[4])) << 8) | ((int)(addr[5])),
	        (((int)(addr[6])) << 8) | ((int)(addr[7])),
	        (((int)(addr[8])) << 8) | ((int)(addr[9])),
	        (((int)(addr[10])) << 8) | ((int)(addr[11])),
	        (((int)(addr[12])) << 8) | ((int)(addr[13])),
	        (((int)(addr[14])) << 8) | ((int)(addr[15])));

	// Return the formatted address to the caller.
	return buf;
}

static char hexcharslower[] = "0123456789abcdef";
static char hexcharsupper[] = "0123456789ABCDEF";

char *FormatOutput::etherToString(const unsigned char *addr)
{
	// Get the buffer to use and then switch buffers.  We need to do
	// this because some callers need to format two Ethernet addresses
	// at the same time.
	char *buf = etherBuf;
	if(buf == etherBuf1)
		etherBuf = etherBuf2;
	else
		etherBuf = etherBuf1;

	// Format the Ethernet address.
	buf[0] = hexcharslower[(addr[0] >> 4) & 0x0F];
	buf[1] = hexcharslower[addr[0] & 0x0F];
	buf[2] = ':';
	buf[3] = hexcharslower[(addr[1] >> 4) & 0x0F];
	buf[4] = hexcharslower[addr[1] & 0x0F];
	buf[5] = ':';
	buf[6] = hexcharslower[(addr[2] >> 4) & 0x0F];
	buf[7] = hexcharslower[addr[2] & 0x0F];
	buf[8] = ':';
	buf[9] = hexcharslower[(addr[3] >> 4) & 0x0F];
	buf[10] = hexcharslower[addr[3] & 0x0F];
	buf[11] = ':';
	buf[12] = hexcharslower[(addr[4] >> 4) & 0x0F];
	buf[13] = hexcharslower[addr[4] & 0x0F];
	buf[14] = ':';
	buf[15] = hexcharslower[(addr[5] >> 4) & 0x0F];
	buf[16] = hexcharslower[addr[5] & 0x0F];
	buf[17] = '\0';

	// Return the result to the caller.
	return buf;
}

const char *FormatOutput::tcpPortToString(unsigned short port, int usedefault)
{
	if(!usedefault)
	{
		const char *token = getToken(tcpPorts, port);
		if(!token)
			token = getToken(commonPorts, port);
		if(token)
			return token;
	}
	struct servent *ent = getservbyport((int)htons(port), "tcp");
	if(ent)
	{
		return ent->s_name;
	}
	if(usedefault)
	{
		sprintf(etherBuf, "%u", (unsigned)port);
		return etherBuf;
	}
	else
	{
		return 0;
	}
}

const char *FormatOutput::udpPortToString(unsigned short port, int usedefault)
{
	if(!usedefault)
	{
		const char *token = getToken(udpPorts, port);
		if(!token)
			token = getToken(commonPorts, port);
		if(token)
			return token;
	}
	struct servent *ent = getservbyport((int)htons(port), "udp");
	if(ent)
	{
		return ent->s_name;
	}
	if(usedefault)
	{
		sprintf(etherBuf, "%u", (unsigned)port);
		return etherBuf;
	}
	else
	{
		return 0;
	}
}

const char *FormatOutput::protoToString(int protocol)
{
	struct protoent *ent;
	if(protocol == IPPROTO_TCP)
	{
		return "Transmission Control Protocol (TCP)";
	}
	else if(protocol == IPPROTO_UDP)
	{
		return "User Datagram Protocol (UDP)";
	}
	else if(protocol == IPPROTO_ICMP)
	{
		return "Internet Control Message Protocol (ICMP)";
	}
	else if(protocol == 58)
	{
		return "Internet Control Message Protocol Version 6 (ICMPv6)";
	}
	else if((ent = getprotobynumber(protocol)) != 0)
	{
		// Use the first alias, if present, because it will normally
		// be the upper-case form of the name, which looks better.
		// (Well, I think it looks better :-) ).
		if(ent->p_aliases && ent->p_aliases[0])
			return ent->p_aliases[0];
		else
			return ent->p_name;
	}
	else
	{
		return 0;
	}
}

void FormatOutput::formatHex(char *output, const unsigned char *block,
				   			 unsigned len, char sep, int useUpperCase)
{
	char *hexchars = (useUpperCase ? hexcharsupper : hexcharslower);
	while(len > 0)
	{
		--len;
		*output++ = hexchars[(*block >> 4) & 0x0F];
		*output++ = hexchars[*block & 0x0F];
		if(len > 0)
		{
			*output++ = sep;
		}
		++block;
	}
	*output = '\0';
}

int FormatOutput::outputToken(FormatToken *tokens, unsigned long num)
{
	while(tokens->token != 0 && tokens->value != num)
	{
		++tokens;
	}
	if(tokens->token)
	{
		outputShortString(tokens->token);
		return 1;
	}
	else
	{
		return 0;
	}
}

const char *FormatOutput::getToken(FormatToken *tokens, unsigned long num,
								   const char *defaultFormat)
{
	while(tokens->token != 0 && tokens->value != num)
	{
		++tokens;
	}
	if(tokens->token)
	{
		return tokens->token;
	}
	else if(defaultFormat)
	{
		sprintf(etherBuf, defaultFormat, (int)num);
		return etherBuf;
	}
	else
	{
		return 0;
	}
}

FormatRuleTable *FormatRulesCreate(void)
{
	FormatRuleTable *linktable;

	// Create the top-level rule table, which contains rules for
	// the link types plus rules for the common payload types.
	linktable = new FormatRuleTable();
	new EthernetLinkFormatRule(linktable, DLT_EN10MB);
	new EthernetLinkFormatRule(linktable, DLT_IEEE802);
	new RawLinkFormatRule(linktable, DLT_RAW);
	new PPPLinkFormatRule(linktable, DLT_PPP);
	new NullLinkFormatRule(linktable, DLT_NULL);
	new IPFormatRule(linktable, PSEUDO_DLT_IP);
	new ARPFormatRule(linktable, PSEUDO_DLT_ARP);
	new ARPFormatRule(linktable, PSEUDO_DLT_RARP);
	new DefaultLinkFormatRule(linktable);

	// Done.
	return linktable;
}
