/*
 * FormatRule.h - Base class for packet formating rules.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__FormatRule_H__
#define	__FormatRule_H__

// Forward declarations.
class FormatRule;
class FormatOutput;
class NetPacket;

// Table of format rules that fall into a particular category.
class FormatRuleTable
{
public:

	FormatRuleTable();
	~FormatRuleTable();

private:

	FormatRule *first;
	FormatRule *last;
	FormatRule *defrule;

public:

	// Add a rule to this table.
	void addRule(FormatRule *rule);

	// Add a default rule to this table.
	void addDefaultRule(FormatRule *rule);

	// Format a packet with a rule that matches a given tag.
	// Returns zero if the tag could not be found.
	int formatPacketShort(NetPacket& packet, FormatOutput& output,
						  unsigned long tag);
	int formatPacketLong(NetPacket& packet, FormatOutput& output,
						 unsigned long tag);

	// Clear all session state from the formatters.
	void clearState(void);

};

// Base class for format rules.
class FormatRule
{
	friend class FormatRuleTable;

public:

	FormatRule(FormatRuleTable *_table, unsigned long _tag);
	FormatRule(FormatRuleTable *_table);
	virtual ~FormatRule();

private:

	FormatRuleTable *table;
	unsigned long    tag;
	FormatRule      *next;

public:

	// Get the rule table associated with this rule.
	FormatRuleTable *getTable() { return table; }

	// Get the tag associated with this rule.
	unsigned long getTag() { return tag; }

	// Format a packet with this rule.
	virtual void formatPacketShort(NetPacket& packet, FormatOutput& output);
	virtual void formatPacketLong(NetPacket& packet, FormatOutput& output);

	// Clear any session state associated with this rule.
	virtual void clearState(void);

};

// Format token structure definition.
typedef struct
{
	unsigned long value;
	const char   *token;

} FormatToken;

// Color values that can be used when formatting short-form packets.
typedef enum
{
	FormatColor_Normal,
	FormatColor_UDP,
	FormatColor_TCPData,
	FormatColor_TCPConnect,
	FormatColor_TCPDisconnect,
	FormatColor_ICMP,
	FormatColor_OtherIP,
	FormatColor_ARP,
	FormatColor_RARP,
	FormatColor_OtherEthernet,

} FormatColor;

#define	IPADDR6_HEX_OR_NAME		0
#define	IPADDR6_HEX_ONLY		1
#define	IPADDR6_NAME_ONLY		2

// Object that formatted output is sent to.
class FormatOutput
{
public:

	FormatOutput();
	virtual ~FormatOutput();

private:

	char	ipBuf1[64];
	char	ipBuf2[64];
	char   *ipBuf;
	char	etherBuf1[64];
	char	etherBuf2[64];
	char   *etherBuf;

public:

	// Output formatted data for short formats.
	virtual void outputShort(const char *format, ...) = 0;
	virtual void outputShortString(const char *string, int len=-1) = 0;
	virtual void outputShortEscaped(const char *string, int len=-1);

	// Output formatted data for long formats.
	virtual void outputSection(const char *name) = 0;
	virtual void outputLong(const char *field, const char *format, ...) = 0;
	virtual void outputHex(const char *field, const void *data,
						   unsigned len) = 0;
	virtual void outputText(const char *field, const void *data,
						    unsigned len, int suppressCRs=0) = 0;

	// Determine if we should include the link level header in short formats.
	virtual int wantLinkHeader(void) = 0;

	// Set the color to display short-form output in.
	virtual void shortFormColor(FormatColor color) = 0;

	// Determine if IP addresses should be resolved.
	virtual int resolveIPAddresses(void) = 0;

	// Determine if fully-qualfied domain names should be shown.
	virtual int fullyQualifiedNames(void) = 0;

	// Mark this packet as needing "back-patching" later to insert
	// hostnames for IP addresses that we attempted to look up
	// asynchronously but don't yet have a reply for.
	virtual void markIPAddress(const void *addr) = 0;

	// Format an IP address for display to the user.
	char *ipaddrToString(const void *addr);

	// Format an IP address for display to the user without resolving it.
	char *ipaddrToStringNoResolve(const void *addr);

	// Look up a domain name for an IP address (NULL if not found).
	char *ipaddrToDomainName(const void *addr);

	// Format an IPv6 address for display to the user.
	char *ipaddr6ToString(const void *addr, int type=IPADDR6_HEX_OR_NAME);

	// Format an Ethernet address for display to the user.
	char *etherToString(const unsigned char *addr);

	// Format a TCP port number in string form.
	const char *tcpPortToString(unsigned short port, int usedefault=1);

	// Format a UDP port number in string form.
	const char *udpPortToString(unsigned short port, int usedefault=1);

	// Format an IP protocol in string form.
	const char *protoToString(int protocol);

	// Format a block of data in hex for display to the user.
	void formatHex(char *output, const unsigned char *block,
				   unsigned len, char sep=' ', int useUpperCase=0);

	// Output a numbered token to the output object.  Returns
	// zero if the token was not found.
	int outputToken(FormatToken *tokens, unsigned long num);

	// Get the string associated with a numbered token.
	const char *getToken(FormatToken *tokens, unsigned long num,
					     const char *defaultFormat=0);

};

// Create the format rule database.  Returns the top-level table
// which is indexed by link type.
FormatRuleTable *FormatRulesCreate(void);

#endif	/* __FormatRule_H__ */
