/*
 * ICMPFormatRule.cc - Formatting rules for ICMP packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "ICMPFormatRule.h"
#include "NetPacket.h"
#include <netinet/ip_icmp.h>

/*
 * Extract ICMP codes and ICMP formatting tables extracted from the
 * source for tcpdump.  Copyright for these follows:
 *
 * Copyright (c) 1988, 1989, 1990, 1991, 1993, 1994, 1995, 1996
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/* rfc1700 */
#ifndef ICMP_UNREACH_NET_UNKNOWN
#define ICMP_UNREACH_NET_UNKNOWN	6	/* destination net unknown */
#endif
#ifndef ICMP_UNREACH_HOST_UNKNOWN
#define ICMP_UNREACH_HOST_UNKNOWN	7	/* destination host unknown */
#endif
#ifndef ICMP_UNREACH_ISOLATED
#define ICMP_UNREACH_ISOLATED		8	/* source host isolated */
#endif
#ifndef ICMP_UNREACH_NET_PROHIB
#define ICMP_UNREACH_NET_PROHIB		9	/* admin prohibited net */
#endif
#ifndef ICMP_UNREACH_HOST_PROHIB
#define ICMP_UNREACH_HOST_PROHIB	10	/* admin prohibited host */
#endif
#ifndef ICMP_UNREACH_TOSNET
#define ICMP_UNREACH_TOSNET		11	/* tos prohibited net */
#endif
#ifndef ICMP_UNREACH_TOSHOST
#define ICMP_UNREACH_TOSHOST		12	/* tos prohibited host */
#endif

/* rfc1716 */
#ifndef ICMP_UNREACH_FILTER_PROHIB
#define ICMP_UNREACH_FILTER_PROHIB	13	/* admin prohibited filter */
#endif
#ifndef ICMP_UNREACH_HOST_PRECEDENCE
#define ICMP_UNREACH_HOST_PRECEDENCE	14	/* host precedence violation */
#endif
#ifndef ICMP_UNREACH_PRECEDENCE_CUTOFF
#define ICMP_UNREACH_PRECEDENCE_CUTOFF	15	/* precedence cutoff */
#endif

/* rfc1256 */
#ifndef ICMP_ROUTERADVERT
#define ICMP_ROUTERADVERT		9	/* router advertisement */
#endif
#ifndef ICMP_ROUTERSOLICIT
#define ICMP_ROUTERSOLICIT		10	/* router solicitation */
#endif

/* Most of the icmp types */
static FormatToken icmp2str[] = {
	{ ICMP_ECHOREPLY,		"echo reply" },
	{ ICMP_SOURCEQUENCH,		"source quench" },
	{ ICMP_REDIRECT,        "redirect" },
	{ ICMP_ECHO,			"echo request" },
	{ ICMP_ROUTERADVERT,    "router advertisement" },
	{ ICMP_ROUTERSOLICIT,		"router solicitation" },
	{ ICMP_TIMXCEED,        "time exceeded" },
	{ ICMP_PARAMPROB,       "parameter problem" },
	{ ICMP_TSTAMP,			"time stamp request" },
	{ ICMP_TSTAMPREPLY,		"time stamp reply" },
	{ ICMP_IREQ,			"information request" },
	{ ICMP_IREQREPLY,		"information reply" },
	{ ICMP_MASKREQ,			"address mask request" },
	{ 0,				NULL }
};

static FormatToken icmp2strlong[] = {
	{ ICMP_ECHOREPLY,		"Echo Reply" },
    { ICMP_UNREACH,         "Destination Unreachable" },
	{ ICMP_SOURCEQUENCH,	"Source Quench" },
	{ ICMP_REDIRECT,        "Redirect" },
	{ ICMP_ECHO,			"Echo Request" },
	{ ICMP_ROUTERADVERT,    "Router Advertisement" },
	{ ICMP_ROUTERSOLICIT,	"Router Solicitation" },
	{ ICMP_TIMXCEED,        "Time Exceeded" },
	{ ICMP_PARAMPROB,       "Parameter Problem" },
	{ ICMP_TSTAMP,			"Time Stamp Request" },
	{ ICMP_TSTAMPREPLY,		"Time Stamp Reply" },
	{ ICMP_IREQ,			"Information Request" },
	{ ICMP_IREQREPLY,		"Information Reply" },
	{ ICMP_MASKREQ,			"Address Mask Request" },
	{ ICMP_MASKREPLY,       "Address Mask Reply" },
	{ 0,				NULL }
};

/* Formats for most of the ICMP_UNREACH codes */
static FormatToken unreach2str[] = {
	{ ICMP_UNREACH_NET,		"net %s unreachable" },
	{ ICMP_UNREACH_HOST,		"host %s unreachable" },
	{ ICMP_UNREACH_SRCFAIL,
	    "%s unreachable - source route failed" },
	{ ICMP_UNREACH_NET_UNKNOWN,	"net %s unreachable - unknown" },
	{ ICMP_UNREACH_HOST_UNKNOWN,	"host %s unreachable - unknown" },
	{ ICMP_UNREACH_ISOLATED,
	    "%s unreachable - source host isolated" },
	{ ICMP_UNREACH_NET_PROHIB,
	    "net %s unreachable - admin prohibited" },
	{ ICMP_UNREACH_HOST_PROHIB,
	    "host %s unreachable - admin prohibited" },
	{ ICMP_UNREACH_TOSNET,
	    "net %s unreachable - tos prohibited" },
	{ ICMP_UNREACH_TOSHOST,
	    "host %s unreachable - tos prohibited" },
	{ ICMP_UNREACH_FILTER_PROHIB,
	    "host %s unreachable - admin prohibited filter" },
	{ ICMP_UNREACH_HOST_PRECEDENCE,
	    "host %s unreachable - host precedence violation" },
	{ ICMP_UNREACH_PRECEDENCE_CUTOFF,
	    "host %s unreachable - precedence cutoff" },
	{ 0,				NULL }
};

static FormatToken unreach2strlong[] = {
	{ ICMP_UNREACH_NET,				"Net Unreachable" },
	{ ICMP_UNREACH_HOST,			"Host Unreachable" },
    { ICMP_UNREACH_PROTOCOL,        "Protocol Unreachable" },
	{ ICMP_UNREACH_PORT,			"Port Unreachable" },
	{ ICMP_UNREACH_NEEDFRAG,        "Need Fragmentation" },
	{ ICMP_UNREACH_SRCFAIL,			"Source Route Failed" },
	{ ICMP_UNREACH_NET_UNKNOWN,		"Net Unknown" },
	{ ICMP_UNREACH_HOST_UNKNOWN,	"Host Unknown" },
	{ ICMP_UNREACH_ISOLATED,		"Source Host Isolated" },
	{ ICMP_UNREACH_NET_PROHIB,		"Net Unreachable - Admin Prohibited" },
	{ ICMP_UNREACH_HOST_PROHIB,		"Host Unreachable - Admin Prohibited" },
	{ ICMP_UNREACH_TOSNET,			"Net Unreachable - TOS Prohibited" },
	{ ICMP_UNREACH_TOSHOST,			"Host Unreachable - TOS Prohibited" },
	{ ICMP_UNREACH_FILTER_PROHIB,	"Host Unreachable - Admin Prohibited Filter" },
	{ ICMP_UNREACH_HOST_PRECEDENCE,	"Host Precedence Violation" },
	{ ICMP_UNREACH_PRECEDENCE_CUTOFF, "Host Precedence Cutoff" },
	{ 0,				NULL }
};

/* Formats for the ICMP_REDIRECT codes */
static FormatToken type2str[] = {
	{ ICMP_REDIRECT_NET,		"redirect %s to net %s" },
	{ ICMP_REDIRECT_HOST,		"redirect %s to host %s" },
	{ ICMP_REDIRECT_TOSNET,		"redirect-tos %s to net %s" },
	{ ICMP_REDIRECT_TOSHOST,	"redirect-tos %s to net %s" },
	{ 0,				NULL }
};

static FormatToken type2strlong[] = {
	{ ICMP_REDIRECT_NET,		"Redirect Network" },
	{ ICMP_REDIRECT_HOST,		"Redirect Host" },
	{ ICMP_REDIRECT_TOSNET,		"Redirect TOS Network" },
	{ ICMP_REDIRECT_TOSHOST,	"Redirect TOS Host" },
	{ 0,				NULL }
};

ICMPFormatRule::ICMPFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
}

ICMPFormatRule::~ICMPFormatRule()
{
}

void ICMPFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	// Set the short-form color for ICMP packets.
	output.shortFormColor(FormatColor_ICMP);

	// Output the source and destination address details from the IP header.
	if(packet.isVersion6())
	{
		output.outputShort("%s > %s: ",
						   output.ipaddr6ToString(packet.ip + 8),
						   output.ipaddr6ToString(packet.ip + 24));
	}
	else
	{
		output.outputShort("%s > %s: ",
						   output.ipaddrToString(&(packet.ipHeader.ip_src)),
						   output.ipaddrToString(&(packet.ipHeader.ip_dst)));
	}

	// Extract the ICMP header details.
	if(packet.subPacketLen < 8)
	{
		output.outputShortString("icmp packet truncated");
		return;
	}
	const u_char *icmpHeader = packet.subPacket;
	unsigned char type = packet.getByte(0);
	unsigned char code = packet.getByte(1);
	packet.adjustSubPacket(8);

	// Output the ICMP message indicator.
	output.outputShortString("icmp: ");

	// Output type-specific information.
	switch(type)
	{
		case ICMP_UNREACH:
		{
			struct ip header;
			if(packet.subPacketLen < sizeof(struct ip))
			{
				output.outputShortString(" - truncated");
				break;
			}
			memcpy(&header, packet.subPacket, sizeof(struct ip));
			if(packet.subPacketLen < (((unsigned)(header.ip_hl)) * 4))
			{
				output.outputShortString(" - truncated");
				break;
			}
			packet.adjustSubPacket(((unsigned)(header.ip_hl)) * 4);
			switch(code)
			{
				case ICMP_UNREACH_PROTOCOL:
				{
					output.outputShort("%s protocol %d unreachable",
									   output.ipaddrToString(&header.ip_dst),
									   (int)(header.ip_p));
				}
				break;

				case ICMP_UNREACH_PORT:
				{
					if(packet.subPacketLen < 4)
					{
						output.outputShortString(" - truncated");
						break;
					}
					unsigned short dport = packet.getShortNetwork(2);
					switch(header.ip_p)
					{
						case IPPROTO_TCP:
						{
							output.outputShort("%s tcp port %s unreachable",
									   output.ipaddrToString(&header.ip_dst),
									   output.tcpPortToString(dport));
						}
						break;

						case IPPROTO_UDP:
						{
							output.outputShort("%s udp port %s unreachable",
									   output.ipaddrToString(&header.ip_dst),
									   output.udpPortToString(dport));
						}
						break;

						default:
						{
							output.outputShort("%s protocol %d port %u "
											   "unreachable",
									   output.ipaddrToString(&header.ip_dst),
									   (int)(header.ip_p), (unsigned)dport);
						}
						break;
					}
				}
				break;

				case ICMP_UNREACH_NEEDFRAG:
				{
					packet.setSubPacket(icmpHeader, 8);
					unsigned short mtu = packet.getShortNetwork(6);
					if(mtu)
					{
						output.outputShort("%s unreachable - need to "
										   "frag (mtu %u)",
								   output.ipaddrToString(&header.ip_dst),
								   (unsigned)mtu);
					}
					else
					{
						output.outputShort("%s unreachable - need to frag",
								   output.ipaddrToString(&header.ip_dst));
					}
				}
				break;

				default:
				{
					const char *format =
						output.getToken(unreach2str, code,
									    "#%d %%s unreachable");
					output.outputShort(format,
									   output.ipaddrToString(&header.ip_dst));
				}
			}
		}
		break;

		case ICMP_REDIRECT:
		{
			struct ip header;
			if(packet.subPacketLen < sizeof(struct ip))
			{
				output.outputShortString(" - truncated");
				break;
			}
			memcpy(&header, packet.subPacket, sizeof(struct ip));
			if(packet.subPacketLen < (((unsigned)(header.ip_hl)) * 4))
			{
				output.outputShortString(" - truncated");
				break;
			}
			output.outputShort(output.getToken(type2str, code,
									    	   "redirect-#%d %%s to net %%s"),
							   output.ipaddrToString(&header.ip_dst),
							   output.ipaddrToString(icmpHeader + 4));
		}
		break;

		case ICMP_MASKREPLY:
		{
			if(packet.subPacketLen < 4)
			{
				output.outputShortString(" - truncated");
				break;
			}
			output.outputShort("address mask is 0x%08x",
							   (long)(packet.getLongNetwork(0)));
		}
		break;

		default:
		{
			output.outputShortString(output.getToken(icmp2str,
													 type, "type-#%d"));
		}
		break;
	}
}

void ICMPFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	// Output the section header for the ICMP information.
	output.outputSection("Internet Control Message Protocol (ICMP)");

	// Make sure that we have enough header information to be useful.
	if(packet.subPacketLen < 8)
	{
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}

	// Parse the ICMP header.
	unsigned char type = packet.getByte(0);
	unsigned char code = packet.getByte(1);
	unsigned short checksum = packet.getShortNetwork(2);
	const u_char *data = packet.subPacket + 4;
	packet.adjustSubPacket(8);

	// Output the common ICMP header fields.
	const char *str = output.getToken(icmp2strlong, type);
	if(str)
	{
		output.outputLong("Type", "%d {{%s}}", (int)type, str);
	}
	else
	{
		output.outputLong("Type", "%d", (int)type);
	}
	switch(type)
	{
		case ICMP_UNREACH:
		{
			str = output.getToken(unreach2strlong, (int)code);
			if(str)
			{
				output.outputLong("Code", "%d {{%s}}", (int)code, str);
			}
			else
			{
				output.outputLong("Code", "%d", (int)code);
			}
		}
		break;

		case ICMP_REDIRECT:
		{
			str = output.getToken(type2strlong, (int)code);
			if(str)
			{
				output.outputLong("Code", "%d {{%s}}", (int)code, str);
			}
			else
			{
				output.outputLong("Code", "%d", (int)code);
			}
		}
		break;

		default:
		{
			output.outputLong("Code", "%d", (int)code);
		}
		break;
	}
	output.outputLong("Checksum", "0x%04x", (int)checksum);

	// Output type-specific fields.
	switch(type)
	{
		case ICMP_ECHOREPLY:
		case ICMP_ECHO:
		{
			output.outputLong("Identifier", "%d",
					(((int)(data[0])) << 8) | ((int)(data[1])));
			if(data[2] != 0 && data[3] == 0)
			{
				// It looks like the sequence number is byte-reversed.
				// Most implementations of ping just copy the sequence
				// number into the ICMP header without converting it
				// into network byte order.
				output.outputLong("Sequence Number", "%d (%d byte reversed)",
						(((int)(data[2])) << 8) | ((int)(data[3])),
						(((int)(data[3])) << 8) | ((int)(data[2])));
			}
			else
			{
				output.outputLong("Sequence Number", "%d",
						(((int)(data[2])) << 8) | ((int)(data[3])));
			}
			if(packet.subPacketLen > 0)
			{
				output.outputSection("Echo Data");
				output.outputHex(0, packet.subPacket, packet.subPacketLen);
			}
		}
		break;

    	case ICMP_UNREACH:
		{
			struct ip header;
			const char *protoname;
			const char *hostname;
			if(packet.subPacketLen < sizeof(struct ip))
			{
				output.outputHex("Truncated Packet",
								 packet.subPacket, packet.subPacketLen);
				break;
			}
			memcpy(&header, packet.subPacket, sizeof(struct ip));
			if(packet.subPacketLen < (((unsigned)(header.ip_hl)) * 4))
			{
				output.outputHex("Truncated Packet",
								 packet.subPacket, packet.subPacketLen);
				break;
			}
			hostname = output.ipaddrToDomainName(&header.ip_dst);
			if(hostname)
			{
				output.outputLong("Destination Address", "%s {{%s}}",
							  output.ipaddrToStringNoResolve(&header.ip_dst),
							  hostname);
			}
			else
			{
				output.outputLong("Destination Address", "%s",
							  output.ipaddrToStringNoResolve(&header.ip_dst));
			}
			switch(code)
			{
				case ICMP_UNREACH_PROTOCOL:
				{
					protoname = output.protoToString((int)(header.ip_p));
					if(protoname)
					{
						output.outputLong("Destination Protocol",
										  "%d {{%s}}",
										  (int)(header.ip_p), protoname);
					}
					else
					{
						output.outputLong("Destination Protocol",
										  "%d", (int)(header.ip_p));
					}
				}
				break;

				case ICMP_UNREACH_PORT:
				{
					unsigned short dport =
						packet.getShortNetwork((((unsigned)(header.ip_hl)) * 4)
												+ 2);
					protoname = output.protoToString((int)(header.ip_p));
					if(protoname)
					{
						output.outputLong("Destination Protocol",
										  "%d {{%s}}",
										  (int)(header.ip_p), protoname);
					}
					else
					{
						output.outputLong("Destination Protocol",
										  "%d", (int)(header.ip_p));
					}
					if(header.ip_p == IPPROTO_TCP)
					{
						protoname = output.tcpPortToString(dport, 0);
					}
					else if(header.ip_p == IPPROTO_UDP)
					{
						protoname = output.udpPortToString(dport, 0);
					}
					else
					{
						protoname = 0;
					}
					if(protoname)
					{
						output.outputLong("Destination Port", "%d {{%s}}",
										  (int)dport, protoname);
					}
					else
					{
						output.outputLong("Destination Port", "%d",
										  (int)dport);
					}
				}
				break;

				case ICMP_UNREACH_NEEDFRAG:
				{
					output.outputLong("Path MTU", "%d",
						(((int)(data[2])) << 8) | ((int)(data[3])));
				}
				break;
			}
			output.outputSection("Original IP Header");
			output.outputHex(0, packet.subPacket, packet.subPacketLen);
		}
		break;

		case ICMP_SOURCEQUENCH:
		{
		}
		break;

		case ICMP_REDIRECT:
		{
		}
		break;

		case ICMP_ROUTERADVERT:
		{
		}
		break;

		case ICMP_ROUTERSOLICIT:
		{
		}
		break;

		case ICMP_TIMXCEED:
		{
		}
		break;

		case ICMP_PARAMPROB:
		{
		}
		break;

		case ICMP_TSTAMP:
		{
		}
		break;

		case ICMP_TSTAMPREPLY:
		{
		}
		break;

		case ICMP_IREQ:
		{
		}
		break;

		case ICMP_IREQREPLY:
		{
		}
		break;

		case ICMP_MASKREQ:
		{
		}
		break;

		case ICMP_MASKREPLY:
		{
		}
		break;

		default:
		{
			if(packet.subPacketLen > 0)
			{
				output.outputSection("ICMP Message Data");
				output.outputHex(0, packet.subPacket, packet.subPacketLen);
			}
		}
		break;
	}
}
