/*
 * IPFormatRule.cc - Formatting rules for IP packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "IPFormatRule.h"
#include "ICMPFormatRule.h"
#include "UDPFormatRule.h"
#include "TCPFormatRule.h"
#include "NetPacket.h"

IPFormatRule::IPFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
	new ICMPFormatRule(&protocolRules, IPPROTO_ICMP);
	new UDPFormatRule(&protocolRules, IPPROTO_UDP);
	new TCPFormatRule(&protocolRules, IPPROTO_TCP);
}

IPFormatRule::~IPFormatRule()
{
}

void IPFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	unsigned short fragment;
	int nextHeader;

	// Make sure that we have enough bytes to provide a useful answer.
	if(packet.subPacketLen < sizeof(struct ip))
	{
		output.outputShortString("truncated IP header");
		return;
	}

	// Extract the header and save it for lower-level formatters.
	memcpy(&packet.ipHeader, packet.subPacket, sizeof(struct ip));

	// Check for IPv6 packets and process them differently if we find them.
	if(packet.isVersion6())
	{
		if(packet.subPacketLen < 40)
		{
			output.outputShortString("truncated IPv6 header");
			return;
		}
		nextHeader = (int)(packet.subPacket[6]);
		packet.adjustSubPacket(40);
		if(!protocolRules.formatPacketShort(packet, output, nextHeader))
		{
			packet.setSubPacket(packet.ip, packet.ipLen);
			output.outputShort("%s > %s: ",
							   output.ipaddr6ToString(packet.ip + 8),
							   output.ipaddr6ToString(packet.ip + 24));
			output.outputShort("ipv6 protocol %d length %u",
							   nextHeader, packet.getShortNetwork(4));
			output.shortFormColor(FormatColor_OtherIP);
		}
		return;
	}

	// If this is a fragment other than the first one, then print
	// generic details about it, but don't do anything else with it.
	fragment = ntohs(packet.ipHeader.ip_off);
	if((fragment & 0x1FFF) != 0)
	{
		output.outputShort("%s > ",
						   output.ipaddrToString(&(packet.ipHeader.ip_src)));
		output.outputShort("%s: ",
						   output.ipaddrToString(&(packet.ipHeader.ip_dst)));
		output.outputShort("fragment id %d offset %d length %d",
						   (int)(ntohs(packet.ipHeader.ip_len)),
						   (int)(fragment & 0x1FFF),
						   (int)(ntohs(packet.ipHeader.ip_len)));
		if((fragment & 0x8000) != 0)
		{
			output.outputShortString(" (RF)");
		}
		if((fragment & 0x4000) != 0)
		{
			output.outputShortString(" (DF)");
		}
		if((fragment & 0x2000) != 0)
		{
			output.outputShortString(" (MF)");
		}
		return;
	}

	// Skip past the IP header.
	packet.adjustSubPacket(((unsigned)(packet.ipHeader.ip_hl)) * 4);

	// Process the packet in a protocol-specific manner.
	if(!protocolRules.formatPacketShort(packet, output, packet.ipHeader.ip_p))
	{
		// We don't have a protocol handler for this protocol, so
		// dump out generic IP header information instead.
		char protobuf[128];
		struct protoent *ent = getprotobynumber((int)(packet.ipHeader.ip_p));
		if(ent)
		{
			sprintf(protobuf, " (%s)", ent->p_name);
		}
		else
		{
			protobuf[0] = '\0';
		}
		output.outputShort("%s > ",
						   output.ipaddrToString(&(packet.ipHeader.ip_src)));
		output.outputShort("%s: ",
						   output.ipaddrToString(&(packet.ipHeader.ip_dst)));
		output.outputShort("protocol %d%s length %d",
						   (int)(packet.ipHeader.ip_p), protobuf,
						   (int)(ntohs(packet.ipHeader.ip_len)));
		output.shortFormColor(FormatColor_OtherIP);
	}

	// If there were fragmentation bits set, then output them now.
	if((fragment & 0x8000) != 0)
	{
		output.outputShortString(" (RF)");
	}
	if((fragment & 0x4000) != 0)
	{
		output.outputShortString(" (DF)");
	}
	if((fragment & 0x2000) != 0)
	{
		output.outputShortString(" (MF)");
	}
}

void IPFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	unsigned short fragment;
	const char *protoname;
	const char *hostname;
	int nextHeader;

	// Make sure that we have enough bytes to provide a useful answer.
	if(packet.subPacketLen < sizeof(struct ip))
	{
		output.outputSection("Internet Protocol (IP)");
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}

	// Extract the header and save it for lower-level formatters.
	memcpy(&packet.ipHeader, packet.subPacket, sizeof(struct ip));

	// Check for IPv6 packets and process them differently if we find them.
	if(packet.isVersion6())
	{
		output.outputSection("Internet Protocol Version 6 (IPv6)");
		if(packet.subPacketLen < 40)
		{
			output.outputHex("Truncated Packet", packet.subPacket,
							 packet.subPacketLen);
			return;
		}
		output.outputLong("Version", "%d",
						  (int)((packet.subPacket[0] >> 4) & 0x0F));
		output.outputLong("Traffic Class", "%d",
						  (int)(packet.subPacket[0] & 0x0F));
		output.outputLong("Flow Label", "%lu",
						  (packet.getLongNetwork(0) & 0x00FFFFFF));
		output.outputLong("Payload Length", "%u",
						  (unsigned)(packet.getShortNetwork(4)));
		nextHeader = (int)(packet.subPacket[6]);
		if((protoname = output.protoToString(nextHeader)) != 0)
		{
			output.outputLong("Next Header", "%d {{%s}}",
							  nextHeader, protoname);
		}
		else
		{
			output.outputLong("Next Header", "%d", nextHeader);
		}
		output.outputLong("Hop Limit", "%d", (int)(packet.subPacket[7]));
		hostname = output.ipaddr6ToString(packet.subPacket + 8,
										  IPADDR6_NAME_ONLY);
		if(hostname)
		{
			output.outputLong("Source Address", "%s {{%s}}",
							  output.ipaddr6ToString(packet.subPacket + 8,
							  						 IPADDR6_HEX_ONLY),
							  hostname);
		}
		else
		{
			output.outputLong("Source Address", "%s",
							  output.ipaddr6ToString(packet.subPacket + 8,
							  						 IPADDR6_HEX_ONLY));
		}
		hostname = output.ipaddr6ToString(packet.subPacket + 24,
										  IPADDR6_NAME_ONLY);
		if(hostname)
		{
			output.outputLong("Destination Address", "%s {{%s}}",
							  output.ipaddr6ToString(packet.subPacket + 24,
							  						 IPADDR6_HEX_ONLY),
							  hostname);
		}
		else
		{
			output.outputLong("Destination Address", "%s",
							  output.ipaddr6ToString(packet.subPacket + 24,
							  						 IPADDR6_HEX_ONLY));
		}
		packet.adjustSubPacket(40);

		// TO DO: process additional IPv4 header blocks.

		// Attempt to process the remaining payload with one of our formatters.
		if(protocolRules.formatPacketLong(packet, output, nextHeader))
		{
			return;
		}

		// Output the IPv6 data payload section in hex.
		output.outputSection("IPv6 Packet Payload Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
		return;
	}

	// Output the IP section header.
	output.outputSection("Internet Protocol (IP)");

	// Format the TOS field as a string.
	char tosbuf[256];
	int needcomma = 0;
	strcpy(tosbuf, "{{");
	if((packet.ipHeader.ip_tos & IPTOS_LOWDELAY) != 0)
	{
		strcat(tosbuf, "Low Delay");
		needcomma = 1;
	}
	if((packet.ipHeader.ip_tos & IPTOS_THROUGHPUT) != 0)
	{
		if(needcomma)
			strcat(tosbuf, ", ");
		strcat(tosbuf, "Throughput");
		needcomma = 1;
	}
	if((packet.ipHeader.ip_tos & IPTOS_RELIABILITY) != 0)
	{
		if(needcomma)
			strcat(tosbuf, ", ");
		strcat(tosbuf, "Reliability");
		needcomma = 1;
	}
	if((packet.ipHeader.ip_tos & IPTOS_LOWCOST) != 0)
	{
		if(needcomma)
			strcat(tosbuf, ", ");
		strcat(tosbuf, "Low Cost");
		needcomma = 1;
	}
	if(needcomma)
		strcat(tosbuf, ", ");
	switch(IPTOS_PREC(packet.ipHeader.ip_tos))
	{
		case IPTOS_PREC_NETCONTROL:
				strcat(tosbuf, "Net Control"); break;
		case IPTOS_PREC_INTERNETCONTROL:
				strcat(tosbuf, "Internet Control"); break;
		case IPTOS_PREC_CRITIC_ECP:
				strcat(tosbuf, "Critic Ecp"); break;
		case IPTOS_PREC_FLASHOVERRIDE:
				strcat(tosbuf, "Flash Override"); break;
		case IPTOS_PREC_FLASH:
				strcat(tosbuf, "Flash"); break;
		case IPTOS_PREC_IMMEDIATE:
				strcat(tosbuf, "Immediate"); break;
		case IPTOS_PREC_PRIORITY:
				strcat(tosbuf, "Priority"); break;
		case IPTOS_PREC_ROUTINE:
				strcat(tosbuf, "Routine"); break;
	}
	strcat(tosbuf, "}}");

	// Output the IP header information.
	output.outputLong("Version", "%d", (int)(packet.ipHeader.ip_v));
	output.outputLong("Header Length", "%d words (%d bytes)",
					  (int)(packet.ipHeader.ip_hl),
					  ((int)(packet.ipHeader.ip_hl)) * 4);
	output.outputLong("Type of Service", "%d %s",
					  (int)(packet.ipHeader.ip_tos), tosbuf);
	output.outputLong("Total Length", "%d",
					  (int)(ntohs(packet.ipHeader.ip_len)));
	output.outputLong("Identifier", "%d",
					  (int)(ntohs(packet.ipHeader.ip_id)));
	fragment = ntohs(packet.ipHeader.ip_off);
	output.outputLong("Fragment Offset", "%d",
					  (int)(fragment & 0x1FFF));
	if((fragment & 0xE000) != 0)
	{
		output.outputLong("Fragmentation Flags", "%s%s%s",
						  (((fragment & 0x8000) != 0)
						  		? "Reserved " : ""),
						  (((fragment & 0x4000) != 0)
						  		? "Dont-Fragment " : ""),
						  (((fragment & 0x2000) != 0)
						  		? "More-Fragments " : ""));
	}
	else
	{
		output.outputLong("Fragmentation Flags", "None");
	}
	output.outputLong("Time to Live", "%d",
					  (int)(packet.ipHeader.ip_ttl));
	if((protoname = output.protoToString((int)(packet.ipHeader.ip_p))) != 0)
	{
		output.outputLong("Protocol", "%d {{%s}}",
						  (int)(packet.ipHeader.ip_p), protoname);
	}
	else
	{
		output.outputLong("Protocol", "%d",
						  (int)(packet.ipHeader.ip_p));
	}
	output.outputLong("Header Checksum", "0x%04x",
					  (int)(ntohs(packet.ipHeader.ip_sum)));
	hostname = output.ipaddrToDomainName(&(packet.ipHeader.ip_src));
	if(hostname)
	{
		output.outputLong("Source Address", "%s {{%s}}",
				  output.ipaddrToStringNoResolve(&(packet.ipHeader.ip_src)),
				  hostname);
	}
	else
	{
		output.outputLong("Source Address", "%s",
				  output.ipaddrToStringNoResolve(&(packet.ipHeader.ip_src)));
	}
	hostname = output.ipaddrToDomainName(&(packet.ipHeader.ip_dst));
	if(hostname)
	{
		output.outputLong("Destination Address", "%s {{%s}}",
				  output.ipaddrToStringNoResolve(&(packet.ipHeader.ip_dst)),
				  hostname);
	}
	else
	{
		output.outputLong("Destination Address", "%s",
				  output.ipaddrToStringNoResolve(&(packet.ipHeader.ip_dst)));
	}

	// Skip past the IP header.
	packet.adjustSubPacket(((unsigned)(packet.ipHeader.ip_hl)) * 4);

	// If this is a fragment other than the first one, then print
	// generic details about it, but don't do anything else with it.
	if((fragment & 0x1FFF) != 0)
	{
		output.outputSection("Fragment Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
	}
	else if(!protocolRules.formatPacketLong(packet, output,
										    packet.ipHeader.ip_p))
	{
		// We don't have a protocol handler for this protocol, so
		// just dump out the payload as-is.
		output.outputSection("IP Packet Payload Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
	}
}

void IPFormatRule::clearState(void)
{
	protocolRules.clearState();
}
