/*
 * LinkFormatRule.cc - Formatting rules for link types.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "LinkFormatRule.h"
#include "NetPacket.h"

LinkFormatRule::LinkFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
}

LinkFormatRule::LinkFormatRule(FormatRuleTable *_table)
	: FormatRule(_table)
{
}

LinkFormatRule::~LinkFormatRule()
{
}

void LinkFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	if(packet.ip)
	{
		packet.setSubPacket(packet.ip, packet.ipLen);
		getTable()->formatPacketShort(packet, output, PSEUDO_DLT_IP);
	}
	else if(packet.arp)
	{
		packet.setSubPacket(packet.arp, packet.arpLen);
		getTable()->formatPacketShort(packet, output, PSEUDO_DLT_ARP);
	}
	else if(packet.rarp)
	{
		packet.setSubPacket(packet.rarp, packet.rarpLen);
		getTable()->formatPacketShort(packet, output, PSEUDO_DLT_RARP);
	}
	else if(packet.payload)
	{
		output.shortFormColor(FormatColor_OtherEthernet);
		output.outputShort("%u bytes of data payload",
						   (unsigned)(packet.payloadLen));
	}
	else
	{
		output.shortFormColor(FormatColor_OtherEthernet);
		packet.resetSubPacket();
		output.outputShort("%u bytes of captured data",
						   (unsigned)(packet.subPacketLen));
	}
}

void LinkFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	if(packet.ip)
	{
		packet.setSubPacket(packet.ip, packet.ipLen);
		getTable()->formatPacketLong(packet, output, PSEUDO_DLT_IP);
	}
	else if(packet.arp)
	{
		packet.setSubPacket(packet.arp, packet.arpLen);
		getTable()->formatPacketLong(packet, output, PSEUDO_DLT_ARP);
	}
	else if(packet.rarp)
	{
		packet.setSubPacket(packet.rarp, packet.rarpLen);
		getTable()->formatPacketLong(packet, output, PSEUDO_DLT_RARP);
	}
	else if(packet.payload)
	{
		output.outputSection("Payload Data");
		output.outputHex(0, packet.payload, packet.payloadLen);
	}
	else
	{
		packet.resetSubPacket();
		output.outputSection("Captured Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
	}
}

EthernetLinkFormatRule::EthernetLinkFormatRule(FormatRuleTable *_table,
											   unsigned long _tag)
	: LinkFormatRule(_table, _tag)
{
}

EthernetLinkFormatRule::~EthernetLinkFormatRule()
{
}

void EthernetLinkFormatRule::formatPacketShort(NetPacket& packet,
											   FormatOutput& output)
{
	if(packet.linkAddrLen == 14 && output.wantLinkHeader())
	{
		unsigned short type;
		const char *linktypename;
		char typebuf[10];
		packet.setSubPacket(packet.linkAddr, packet.linkAddrLen);
		type = packet.getShortNetwork(12);
		if(type == ETHERTYPE_IP)
		{
			linktypename = "ip";
		}
		else if(type == ETHERTYPE_IPV6)
		{
			linktypename = "ipv6";
		}
		else if(type == ETHERTYPE_ARP)
		{
			linktypename = "arp";
		}
		else if(type == ETHERTYPE_REVARP)
		{
			linktypename = "rarp";
		}
		else
		{
			sprintf(typebuf, "0x%04x", (int)type);
			linktypename = typebuf;
		}
		output.outputShort("%s > %s %s: ",
						   output.etherToString(packet.linkAddr + 6),
						   output.etherToString(packet.linkAddr),
						   linktypename);
	}
	LinkFormatRule::formatPacketShort(packet, output);
}

const char *EthernetVendorName(const unsigned char *address);
const char *EthernetTypeName(unsigned short type);

void EthernetLinkFormatRule::formatPacketLong(NetPacket& packet,
											  FormatOutput& output)
{
	if(packet.linkAddrLen == 14)
	{
		const char *vendor;
		output.outputSection("Ethernet Header");
		vendor = EthernetVendorName(packet.linkAddr);
		if(vendor)
		{
			output.outputLong("Destination Address", "%s {{%s}}",
							  output.etherToString(packet.linkAddr),
							  vendor);
		}
		else
		{
			output.outputLong("Destination Address", "%s",
							  output.etherToString(packet.linkAddr));
		}
		vendor = EthernetVendorName(packet.linkAddr + 6);
		if(vendor)
		{
			output.outputLong("Source Address", "%s {{%s}}",
							  output.etherToString(packet.linkAddr + 6),
							  vendor);
		}
		else
		{
			output.outputLong("Source Address", "%s",
							  output.etherToString(packet.linkAddr + 6));
		}
		packet.setSubPacket(packet.linkAddr, packet.linkAddrLen);
		unsigned short type = packet.getShortNetwork(12);
		const char *linktypename = EthernetTypeName(type);
		if(linktypename)
		{
			output.outputLong("Packet Type", "0x%04x {{%s}}", (int)type,
							  linktypename);
		}
		else
		{
			output.outputLong("Packet Type", "0x%04x", (int)type);
		}
	}
	LinkFormatRule::formatPacketLong(packet, output);
}

RawLinkFormatRule::RawLinkFormatRule(FormatRuleTable *_table,
									 unsigned long _tag)
	: LinkFormatRule(_table, _tag)
{
}

RawLinkFormatRule::~RawLinkFormatRule()
{
}

void RawLinkFormatRule::formatPacketShort(NetPacket& packet,
										  FormatOutput& output)
{
	if(output.wantLinkHeader())
	{
		output.outputShortString("ip: ");
	}
	LinkFormatRule::formatPacketShort(packet, output);
}

void RawLinkFormatRule::formatPacketLong(NetPacket& packet,
										 FormatOutput& output)
{
	LinkFormatRule::formatPacketLong(packet, output);
}

PPPLinkFormatRule::PPPLinkFormatRule(FormatRuleTable *_table,
								     unsigned long _tag)
	: LinkFormatRule(_table, _tag)
{
}

PPPLinkFormatRule::~PPPLinkFormatRule()
{
}

void PPPLinkFormatRule::formatPacketShort(NetPacket& packet,
										  FormatOutput& output)
{
	if(packet.linkAddrLen == 4 && output.wantLinkHeader())
	{
		packet.setSubPacket(packet.linkAddr, packet.linkAddrLen);
		output.outputShort("%c %4d %02x %04x: ",
						   ((packet.linkAddr[0] != 0) ? 'O' : 'I'),
						   (int)(packet.header->caplen),
						   (int)(packet.linkAddr[1]),
						   (int)(packet.getShortNetwork(2)));
	}
	LinkFormatRule::formatPacketShort(packet, output);
}

void PPPLinkFormatRule::formatPacketLong(NetPacket& packet,
										 FormatOutput& output)
{
	if(packet.linkAddrLen == 4)
	{
		output.outputSection("Point-to-Point Protocol (PPP) Link Information");
		output.outputLong("Direction", "%s",
						  ((packet.linkAddr[0] != 0)
						  		? "Output" : "Input"));
	}
	LinkFormatRule::formatPacketLong(packet, output);
}

NullLinkFormatRule::NullLinkFormatRule(FormatRuleTable *_table,
									   unsigned long _tag)
	: LinkFormatRule(_table, _tag)
{
}

NullLinkFormatRule::~NullLinkFormatRule()
{
}

void NullLinkFormatRule::formatPacketShort(NetPacket& packet,
										   FormatOutput& output)
{
	if(packet.linkAddrLen == 4 && output.wantLinkHeader())
	{
#ifdef linux
		// The address family is set to something bogus in Linux.
		output.outputShortString("ip: ");
#else
		packet.setSubPacket(packet.linkAddr, packet.linkAddrLen);
		unsigned long family = packet.getLongNetwork(0);
		if(family == AF_INET)
		{
			output.outputShortString("ip: ");
		}
		else
		{
			output.outputShort("AF %d: ", (int)family);
		}
#endif
	}
	LinkFormatRule::formatPacketShort(packet, output);
}

void NullLinkFormatRule::formatPacketLong(NetPacket& packet,
										  FormatOutput& output)
{
#ifndef linux
	if(packet.linkAddrLen == 4)
	{
		packet.setSubPacket(packet.linkAddr, packet.linkAddrLen);
		unsigned long family = packet.getLongNetwork(0);
		if(family != AF_INET)
		{
			output.outputSection("Link Information");
			output.outputLong("Address Family", "%d", (int)family);
		}
	}
#endif
	LinkFormatRule::formatPacketLong(packet, output);
}

DefaultLinkFormatRule::DefaultLinkFormatRule(FormatRuleTable *_table)
	: LinkFormatRule(_table)
{
}

DefaultLinkFormatRule::~DefaultLinkFormatRule()
{
}

void DefaultLinkFormatRule::formatPacketShort(NetPacket& packet,
											  FormatOutput& output)
{
	LinkFormatRule::formatPacketShort(packet, output);
}

void DefaultLinkFormatRule::formatPacketLong(NetPacket& packet,
											 FormatOutput& output)
{
	LinkFormatRule::formatPacketLong(packet, output);
}
