/*
 * LookupDialog.cc - Dialog box code for the host lookup functionality.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gnome.h>
#include "LookupDialog.h"
#include "NetIncludes.h"

LookupDialogBox::LookupDialogBox(GtkWidget *parent)
	: DialogBox(parent, _("Host lookup"),
				DIALOG_BUTTON_OK | DIALOG_BUTTON_CLOSE, 0)
{
	GtkWidget *hbox;
	GtkWidget *label;

	// Create the hostname prompt.
	hbox = gtk_hbox_new(0, 0);
	label = gtk_label_new(_("Hostname or IP address:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 5);
	hostname = addText();
	setFocus(hostname);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), hostname);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Add a horizontal separator to the dialog.
	gtk_box_pack_start_defaults(GTK_BOX(vbox), gtk_hseparator_new());

	// Create the official name output field.
	hbox = gtk_hbox_new(0, 0);
	label = gtk_label_new(_("Official name:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 5);
	officialName = addText(FALSE);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), officialName);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Create the name aliases field.
	hbox = gtk_hbox_new(0, 0);
	label = gtk_label_new(_("Aliases:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 5);
	aliases = addText(FALSE);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), aliases);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Create the IP address field.
	hbox = gtk_hbox_new(0, 0);
	label = gtk_label_new(_("Primary address:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 5);
	ipAddress = addText(FALSE);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), ipAddress);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Create the other IP addresses field.
	hbox = gtk_hbox_new(0, 0);
	label = gtk_label_new(_("Other addresses:"));
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, TRUE, 5);
	otherIPs = addText(FALSE);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), otherIPs);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Trap interesting signals.
	gtk_signal_connect(GTK_OBJECT(hostname), "changed",
					   GTK_SIGNAL_FUNC(cb_hostname_changed), this);

	// Set the initial button states.
	setSensitive(okbutton, FALSE);
	setDefault(closebutton);
}

LookupDialogBox::~LookupDialogBox()
{
}

void LookupDialogBox::cb_hostname_changed(GtkWidget *widget,
										  LookupDialogBox *box)
{
	box->hostnameChanged();
}

void LookupDialogBox::hostnameChanged(void)
{
	if(textEmpty(hostname))
	{
		setSensitive(okbutton, FALSE);
		setDefault(closebutton);
	}
	else
	{
		setSensitive(okbutton, TRUE);
		setDefault(okbutton);
	}
}

void LookupDialogBox::ok(void)
{
	char *name = gtk_entry_get_text(GTK_ENTRY(hostname));
	struct in_addr addr;
	struct hostent *ent;

	if((addr.s_addr = inet_addr(name)) != (unsigned long)(-1L))
	{
		// Perform a reverse lookup based on an IP address.
		ent = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
	}
	else
	{
		// Perform a normal lookup based on a hostname.
		ent = gethostbyname(name);
	}

	if(ent)
	{
		// Fill in the dialog's fields with the result information.
		gtk_entry_set_text(GTK_ENTRY(officialName), ent->h_name);
		if(ent->h_aliases && ent->h_aliases[0] != NULL)
		{
			int alias = 0;
			char *names = 0;
			char *newnames;
			int len = 0;
			while(ent->h_aliases[alias] != NULL)
			{
				if(names)
					len += strlen(ent->h_aliases[alias]) + 3;
				else
					len += strlen(ent->h_aliases[alias]) + 1;
				if((newnames = (char *)realloc(names, len)) == NULL)
					break;
				if(names)
					strcat(newnames, ", ");
				else
					*newnames = '\0';
				strcat(newnames, ent->h_aliases[alias]);
				names = newnames;
				++alias;
			}
			gtk_entry_set_text(GTK_ENTRY(aliases), (names ? names : ""));
			if(names)
			{
				free(names);
			}
		}
		else
		{
			gtk_entry_set_text(GTK_ENTRY(aliases), "");
		}
		memcpy(&addr, ent->h_addr, sizeof(addr));
		gtk_entry_set_text(GTK_ENTRY(ipAddress), inet_ntoa(addr));
		if(ent->h_addr_list[1] != NULL)
		{
			int alias = 1;
			char *names = 0;
			char *newnames;
			char *ipaddr;
			int len = 0;
			while(ent->h_addr_list[alias] != NULL)
			{
				memcpy(&addr, ent->h_addr_list[alias], sizeof(addr));
				ipaddr = inet_ntoa(addr);
				if(names)
					len += strlen(ipaddr) + 3;
				else
					len += strlen(ipaddr) + 1;
				if((newnames = (char *)realloc(names, len)) == NULL)
					break;
				if(names)
					strcat(newnames, ", ");
				else
					*newnames = '\0';
				strcat(newnames, ipaddr);
				names = newnames;
				++alias;
			}
			gtk_entry_set_text(GTK_ENTRY(otherIPs), (names ? names : ""));
			if(names)
			{
				free(names);
			}
		}
		else
		{
			gtk_entry_set_text(GTK_ENTRY(otherIPs), "");
		}
	}
	else
	{
		// An error occurred during the lookup: report it to the user.
		char *message;
		if(h_errno < 0)
			message = g_strerror(errno);
		else if(h_errno == HOST_NOT_FOUND)
			message = _("Host not found (authoritative answer)");
		else if(h_errno == TRY_AGAIN)
			message = _("Host not found (non-authoritative answer)");
		else if(h_errno == NO_RECOVERY)
			message = _("Non recoverable error");
		else if(h_errno == NO_DATA)
			message = _("No data found");
		else
			message = _("Host lookup failed");
		GtkWidget *dlog =
			gnome_error_dialog_parented(message, GTK_WINDOW(dialog));
		gnome_dialog_run(GNOME_DIALOG(dlog));
	}
}
