/*
 * NetMonitor.cc - Network monitoring object.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetMonitor.h"

NetMonitor::NetMonitor()
{
	capture = 0;
	isfile = 0;
	dumper = 0;
	netmask = htonl(0xFFFFFF00L);
	packetNum = 0;
	programCompiled = 0;
	linkType = DLT_RAW;
	snapLen = 2048;
	firstSink = 0;
}

NetMonitor::~NetMonitor()
{
	// Close the dump and capture streams.
	closeDumper();
	closeCapture();

	// Delete all packet sinks that are attached to this monitor.
	NetMonitorSink *current, *next;
	current = firstSink;
	while(current != 0)
	{
		next = current->next;
		delete current;
		current = next;
	}
}

int NetMonitor::openDevice(char *device, char *errbuf, int snaplen)
{
	bpf_u_int32 netnum;

	// Close the existing packet capture stream, if necessary.
	closeCapture();

	// Find the default network device, if necessary.
	if(!device)
	{
		device = pcap_lookupdev(errbuf);
		if(!device)
		{
			return 0;
		}
	}

	// Attempt to open a live connection to a network device.
	if((capture = pcap_open_live(device, snaplen, 1, 1, errbuf)) == 0)
	{
		return 0;
	}
	isfile = 0;
	snapLen = snaplen;

	// Get the netmask for the device, because it is needed when
	// filter programs are compiled for the device.
	if(pcap_lookupnet(device, &netnum, &netmask, errbuf) != 0)
	{
		// Default to 255.255.255.0, which will normally be OK.
		netmask = htonl(0xFFFFFF00L);
	}

	// Get the type of data link in use.
	linkType = pcap_datalink(capture);

	// Reset the packet number.
	packetNum = 0;

	// Ready to go now.
	return 1;
}

int NetMonitor::openFile(const char *filename, char *errbuf)
{
	bpf_u_int32 netnum;
	char *device;

	// Close the existing packet capture stream, if necessary.
	closeCapture();

	// Attempt to open the supplied file.
	if((capture = pcap_open_offline(filename, errbuf)) == 0)
	{
		return 0;
	}
	isfile = 1;
	snapLen = pcap_snapshot(capture);

	// Get the netmask to use for compiling filter programs.
	// Unfortunately, tcpdump and friends don't store the source
	// netmask in the capture file, so we have to guess based
	// on the local network conditions.  This assumes that the
	// captured data was for the local network, which it will be
	// most of the time.  Not sure what else we can do.
	if((device = pcap_lookupdev(errbuf)) == 0 ||
	   pcap_lookupnet(device, &netnum, &netmask, errbuf) != 0)
	{
		// Default to 255.255.255.0, which will normally be OK.
		netmask = htonl(0xFFFFFF00L);
	}

	// Get the type of data link in use.
	linkType = pcap_datalink(capture);

	// Reset the packet number.
	packetNum = 0;

	// Ready to go now.
	return 1;
}

void NetMonitor::closeCapture(void)
{
	if(programCompiled)
	{
		free(program.bf_insns);
		programCompiled = 0;
	}
	if(capture)
	{
		pcap_close(capture);
		capture = 0;
	}
}

int NetMonitor::openDumper(const char *filename)
{
	closeDumper();
	if(!capture)
	{
		return 0;
	}
	dumper = pcap_dump_open(capture, filename);
	return (dumper != 0);
}

void NetMonitor::closeDumper(void)
{
	if(dumper)
	{
		pcap_dump_close(dumper);
		dumper = 0;
	}
}

// A packet filter program that accepts all packets.
static struct bpf_insn nullinstructions[] = {BPF_STMT(BPF_RET, 2048)};
static struct bpf_program nullprogram = {1, nullinstructions};

int NetMonitor::addFilter(char *filter, char *errbuf)
{
	// Abort if packet capturing is not enabled.  This shouldn't
	// happen, but return something meaningful anyway.
	if(!capture)
	{
		strcpy(errbuf, "packet capturing is not enabled");
		return 0;
	}

	// Free the current packet filter program.
	if(programCompiled)
	{
		free(program.bf_insns);
	}

	// Compile the new filter program.
	programCompiled =
		(pcap_compile(capture, &program, filter, 1, netmask) == 0);

	// Set the new filter to be used by the packet capture library.
	// If something goes wrong, then revert to the null program.
	if(programCompiled)
	{
		if(pcap_setfilter(capture, &program) != 0)
		{
			strcpy(errbuf, pcap_geterr(capture));
			free(program.bf_insns);
			programCompiled = 0;
			pcap_setfilter(capture, &nullprogram);
		}
	}
	else
	{
		strcpy(errbuf, pcap_geterr(capture));
		pcap_setfilter(capture, &nullprogram);
	}
	return programCompiled;
}

void NetMonitor::removeFilter(void)
{
	// Abort if packet capturing is not enabled.
	if(!capture)
	{
		return;
	}

	// Delete the current filter, if any.
	if(programCompiled)
	{
		free(program.bf_insns);
		programCompiled = 0;
	}

	// Set a new filter that just accepts everything.
	pcap_setfilter(capture, &nullprogram);
}

void NetMonitor::addMonitorSink(NetMonitorSink *sink)
{
	sink->monitor = this;
	sink->next = firstSink;
	firstSink = sink;
}

int NetMonitor::processSelect(void)
{
	if(capture)
	{
		int result = pcap_dispatch(capture, 0, packetHandler, (u_char *)this);
		if(isfile && result == 0)
		{
			// Reading from a file that is now empty.
			closeDumper();
			closeCapture();
			return 0;
		}
		else if(result < 0)
		{
			// Some kind of error occurred.
			closeDumper();
			closeCapture();
			return 0;
		}
		return 1;
	}
	else
	{
		return 0;
	}
}

void NetMonitor::packetLoop(void)
{
	if(capture)
	{
		pcap_loop(capture, -1, packetHandler, (u_char *)this);
	}
}

void NetMonitor::dumpFilter(void)
{
	if(programCompiled)
	{
		int n;
		printf("Length of filter program: %u\n", program.bf_len);
		for(n = 0; n < (int)(program.bf_len); ++n)
		{
			puts(bpf_image(program.bf_insns + n, n));
		}
	}
}

void NetMonitor::packetHandler(u_char *user, const struct pcap_pkthdr *header,
							   const u_char *data)
{
	NetMonitor *monitor = (NetMonitor *)user;
	NetMonitorSink *sink;

	// Dump the incoming packet to the capture file if necessary.
	if(monitor->dumper)
	{
		pcap_dump((u_char *)(monitor->dumper), header, data);
	}

	// Parse the basic contents of the packet and then pass it
	// to the network monitor sinks attached to the monitor.
	NetPacket packet(monitor->linkType, header, data, ++(monitor->packetNum));
	packet.parseBasic();
	sink = monitor->firstSink;
	while(sink != 0)
	{
		sink->processPacket(packet);
		sink = sink->next;
	}
}

NetMonitorSink::NetMonitorSink()
{
	monitor = 0;
	next = 0;
}

NetMonitorSink::~NetMonitorSink()
{
}
