/*
 * NetMonitor.h - Network monitoring object.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__NetMonitor_H__
#define	__NetMonitor_H__

#include "NetPacket.h"

class NetMonitorSink;

class NetMonitor
{
public:

	NetMonitor();
	~NetMonitor();

private:

	pcap_t *capture;
	int isfile;
	pcap_dumper_t *dumper;
	bpf_u_int32 netmask;
	bpf_u_int32 packetNum;
	struct bpf_program program;
	int programCompiled;
	int linkType;
	int snapLen;
	NetMonitorSink *firstSink;

public:

	// Determine if the network monitor is open.
	int isOpen(void) const { return (capture != 0); }

	// Open a live connection to one of the network devices.
	// If the device name is set to NULL, then use the default.
	int openDevice(char *device, char *errbuf, int snaplen=2048);

	// Open a connection to a data file containing packet data.
	int openFile(const char *filename, char *errbuf);

	// Close the network monitor's capture stream.
	void closeCapture(void);

	// Open a packet dumper for capturing the output to a file
	// while it is being delivered to the higher layers.
	int openDumper(const char *filename);

	// Close the packet dumper, if any.
	void closeDumper(void);

	// Determine if the packet capture is coming from a file.
	int isCaptureFile() { return isfile; }

	// Get the packet capture object that is associated with this monitor.
	pcap_t *getCapture(void) { return capture; }

	// Get the link type for this monitor object.
	int getLinkType(void) { return linkType; }

	// Get the snapshot length used for the last capture or file read.
	int getSnapLen(void) { return snapLen; }

	// Get the file descriptor for the live connection, so that
	// higher levels can select on it for incoming packets.
	int getCaptureFD() { return pcap_fileno(capture); }

	// Add a packet filter to the packet capture input stream.
	// Returns zero on error, with a reason in "errbuf".
	int addFilter(char *filter, char *errbuf);

	// Remove the packet filter from the packet capture input stream.
	// i.e. revert to accepting all packets.
	void removeFilter(void);

	// Add a network packet sink to this monitor.
	void addMonitorSink(NetMonitorSink *sink);

	// Processing incoming packets in response to a "select".
	// Returns zero at EOF of a save file, or on error.
	int processSelect(void);

	// Loop, receiving all packets until the program is terminated.
	// This is intended for debugging use only.
	void packetLoop(void);

	// Dump the current packet filter to stdout, for debugging.
	void dumpFilter(void);

private:

	// Process incoming packets.
	static void packetHandler(u_char *user, const struct pcap_pkthdr *header,
							  const u_char *data);

};

class NetMonitorSink
{
	friend class NetMonitor;

public:

	NetMonitorSink();
	virtual ~NetMonitorSink();

private:

	NetMonitor *monitor;
	NetMonitorSink *next;

public:

	// Get the network monitor associated with this sink.
	NetMonitor *getMonitor(void) { return monitor; }

	// Process an incoming packet from the associated net monitor.
	virtual void processPacket(NetPacket& packet) = 0;

};

#endif	/* __NetMonitor */
