/*
 * NetPacket.cc - Data structure for holding a packet while it is being parsed.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetPacket.h"

NetPacket::NetPacket(int _linkType, const struct pcap_pkthdr *_header,
			  		 const u_char *_packet, bpf_u_int32 _packetNum)
{
	linkType = _linkType;
	packetNum = _packetNum;
	header = _header;
	packet = _packet;
	subPacket = packet;
	subPacketLen = header->caplen;
	linkAddr = 0;
	linkAddrLen = 0;
	payload = 0;
	payloadLen = 0;
	ip = 0;
	ipLen = 0;
	arp = 0;
	arpLen = 0;
	rarp = 0;
	rarpLen = 0;
}

NetPacket::~NetPacket()
{
}

void NetPacket::resetSubPacket(void)
{
	subPacket = packet;
	subPacketLen = header->caplen;
}

int NetPacket::parseBasic(void)
{
	int useful = 1;

	// Reset the sub-packet before we start so that we know where we are.
	resetSubPacket();

	// Determine how to pull apart the packet from the link type.
	switch(linkType)
	{
		case DLT_NULL:
		{
			// The NULL link type prefixes packets with a 32-bit
			// family code.  e.g. "AF_INET".
			getSubBuffer(0, 4, linkAddr, linkAddrLen);
			adjustSubPacket(4);
			payload = subPacket;
			payloadLen = subPacketLen;
			ip = subPacket;
			ipLen = subPacketLen;
		}
		break;

		case DLT_EN10MB:
		case DLT_IEEE802:
		{
			// The Ethernet link type prefixes packets with a 14
			// byte header containing the 6-byte destination address,
			// the 6-byte source address, and a 2-byte payload type.
			unsigned short type = getShortNetwork(12);
			getSubBuffer(0, 14, linkAddr, linkAddrLen);
			adjustSubPacket(14);
			payload = subPacket;
			payloadLen = subPacketLen;
			if(type == ETHERTYPE_IP)
			{
				ip = subPacket;
				ipLen = subPacketLen;
			}
			else if(type == ETHERTYPE_IPV6)
			{
				ip = subPacket;
				ipLen = subPacketLen;
			}
			else if(type == ETHERTYPE_ARP)
			{
				arp = subPacket;
				arpLen = subPacketLen;
			}
			else if(type == ETHERTYPE_REVARP)
			{
				rarp = subPacket;
				rarpLen = subPacketLen;
			}
			else
			{
				useful = 0;
			}
		}
		break;

		case DLT_PPP:
		{
			// The PPP link type prefixes packets with a four byte header.
			// Note: The Linux version of libpcap reports PPP and SLIP
			// packets as DLT_RAW, so this case won't get used on Linux.
			getSubBuffer(0, 4, linkAddr, linkAddrLen);
			adjustSubPacket(4);
			payload = subPacket;
			payloadLen = subPacketLen;
			ip = subPacket;
			ipLen = subPacketLen;
		}
		break;

		case DLT_RAW:
		{
			// The RAW type does not have a link header at all.
			payload = subPacket;
			payloadLen = subPacketLen;
			ip = subPacket;
			ipLen = subPacketLen;
		}
		break;

		default:
		{
			// There is not enough information to provide a useful parse.
			useful = 0;
		}
		break;
	}

	// Reset the sub-packet back to its original state.
	resetSubPacket();
	return useful;
}

NetPacketCached *NetPacket::convertIntoCached(bpf_u_int32 addr1,
											  bpf_u_int32 addr2)
{
	NetPacketCached *cache =
		(NetPacketCached *)malloc(sizeof(NetPacketCached) +
								  header->caplen - 1);
	if(cache)
	{
		cache->linkType = (bpf_u_int32)linkType;
		cache->packetNum = packetNum;
		cache->addr1 = addr1;
		cache->addr2 = addr2;
		cache->header = *header;
		if(header->caplen > 0)
		{
			memcpy(cache->packet, packet, header->caplen);
		}
	}
	return cache;
}
