/*
 * NetPacket.h - Data structure for holding a packet while it is being parsed.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	__NetPacket_H__
#define	__NetPacket_H__

#include "NetIncludes.h"

typedef struct
{
	bpf_u_int32			linkType;
	bpf_u_int32			packetNum;
	bpf_u_int32			addr1;
	bpf_u_int32			addr2;
	struct pcap_pkthdr	header;
	u_char				packet[1];

} NetPacketCached;

class NetPacket
{
public:

	NetPacket(int _linkType, const struct pcap_pkthdr *_header,
			  const u_char *_packet, bpf_u_int32 _packetNum);
	~NetPacket();

public:

	// Data that was acquired from the packet capture layer.
	int          linkType;
	bpf_u_int32  packetNum;
	const struct pcap_pkthdr *header;
	const u_char *packet;

	// Current sub-packet that is being processed.
	const u_char *subPacket;
	unsigned      subPacketLen;

	// Fields parsed while the packet is being processed.
	const u_char *linkAddr;			// Link address information.
	unsigned      linkAddrLen;		// Length of link address information.
	const u_char *payload;			// Payload of the packet.
	unsigned      payloadLen;		// Length of the packet payload.
	const u_char *ip;				// Encapsulated IP packet, if present.
	unsigned      ipLen;			// Length of encapsulated IP packet.
	struct ip	  ipHeader;			// Extracted IP header.
	unsigned short srcPort;			// Source UDP/TCP port.
	unsigned short dstPort;			// Destination UDP/TCP port.
	const u_char *arp;				// Encapsulated ARP packet, if present.
	unsigned	  arpLen;			// Length of encapsulated ARP packet.
	const u_char *rarp;				// Encapsulated RARP packet, if present.
	unsigned	  rarpLen;			// Length of encapsulated RARP packet.

public:

	// Adjust the sub-packet that is being processed.
	void adjustSubPacket(unsigned removeHead, unsigned removeTail=0)
			{ subPacket += removeHead;
			  subPacketLen -= (removeHead + removeTail); }

	// Unadjust the sub-packet that is being processed.
	void unadjustSubPacket(unsigned removeHead, unsigned removeTail=0)
			{ subPacket -= removeHead;
			  subPacketLen += (removeHead + removeTail); }

	// Set the sub-packet to a specific sub-buffer.
	void setSubPacket(const u_char *buf, unsigned len)
			{ subPacket = buf; subPacketLen = len; }

	// Reset the sub-packet to contain the whole captured packet.
	void resetSubPacket(void);

	// Extract quantities of various sizes and endian-ness from the
	// current sub-packet.
	unsigned char getByte(unsigned offset)
		{ return subPacket[offset]; }
	unsigned short getShortNetwork(unsigned offset)
		{ return (((unsigned short)(subPacket[offset])) << 8) |
		          ((unsigned short)(subPacket[offset + 1])); }
	unsigned short getShortLittle(unsigned offset)
		{ return (((unsigned short)(subPacket[offset + 1])) << 8) |
		          ((unsigned short)(subPacket[offset])); }
	unsigned long getLongNetwork(unsigned offset)
		{ return (((unsigned short)(subPacket[offset])) << 24) |
		         (((unsigned short)(subPacket[offset + 1])) << 16) |
		         (((unsigned short)(subPacket[offset + 2])) << 8) |
		          ((unsigned short)(subPacket[offset + 3])); }
	unsigned long getLongLittle(unsigned offset)
		{ return (((unsigned short)(subPacket[offset + 3])) << 24) |
		         (((unsigned short)(subPacket[offset + 2])) << 16) |
		         (((unsigned short)(subPacket[offset + 1])) << 8) |
		          ((unsigned short)(subPacket[offset])); }

	// Extract a sub-buffer from the current sub-packet.
	void getSubBuffer(unsigned offset, unsigned length,
					  const u_char *& buf, unsigned& len)
		{ buf = subPacket + offset; len = length; }

	// Parse the packet enough to extract basic link and payload information.
	// If the payload is something known, then also extract that information.
	// Returns zero if there is not enough information to be useful.
	int parseBasic(void);

	// Convert this packet into the cached form.  NULL if out of memory.
	NetPacketCached *convertIntoCached(bpf_u_int32 addr1=0,
									   bpf_u_int32 addr2=0);

	// Determine if the IP packet is version 6.
	int isVersion6(void) { return (ipHeader.ip_v == 6); }

};

#endif	/* __NetPacket_H__ */
