/*
 * PingDialog.cc - Dialog box code for the ping functionality.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <gnome.h>
#include "PingDialog.h"

PingDialogSupport::PingDialogSupport(PingDialogBox *_dlog)
{
	dlog = _dlog;
	timeoutId = (guint)(-1);
	selectId = -1;
}

PingDialogSupport::~PingDialogSupport()
{
}

gboolean PingDialogSupport::cb_timeout_proc(gpointer data)
{
	((PingDialogSupport *)data)->timeoutFired();
	return TRUE;
}

void PingDialogSupport::cb_select_proc(gpointer data, gint source,
									   GdkInputCondition cond)
{
	((PingDialogSupport *)data)->selectFired();
}

void PingDialogSupport::registerTimeout(int nsecs)
{
	timeoutId = gtk_timeout_add(nsecs * 1000, cb_timeout_proc, this);
}

void PingDialogSupport::unregisterTimeout(void)
{
	gtk_timeout_remove(timeoutId);
}

void PingDialogSupport::registerSelect(void)
{
	selectId = gdk_input_add(fd, GDK_INPUT_READ, cb_select_proc, this);
}

void PingDialogSupport::unregisterSelect(void)
{
	gtk_input_remove(selectId);
}

void PingDialogSupport::pingResponse(struct in_addr& addr, unsigned seqnum,
							  		 unsigned ttl, struct timeval& roundTrip)
{
	GtkWidget *replies = dlog->replies;
	double ms;
	char seqbuf[32];
	char ttlbuf[32];
	char msbuf[64];
	gchar *info[4];
	gint row;

	// Build the column strings to be displayed.
	info[0] = inet_ntoa(addr);
	sprintf(seqbuf, "%u", seqnum);
	info[1] = seqbuf;
	sprintf(ttlbuf, "%u", ttl);
	info[2] = ttlbuf;
	ms = ((double)(roundTrip.tv_sec)) * 1000.0 +
	     ((double)(roundTrip.tv_usec)) / 1000.0;
	sprintf(msbuf, "%.1f ms", ms);
	info[3] = msbuf;

	// Add the new row to the list.
	gtk_clist_freeze(GTK_CLIST(replies));
	row = gtk_clist_append(GTK_CLIST(replies), info);
	gtk_clist_set_selectable(GTK_CLIST(replies), row, FALSE);
	gtk_clist_moveto(GTK_CLIST(replies), row, 0, 1.0, 0.0);
	gtk_clist_thaw(GTK_CLIST(replies));
}

int PingDialogSupport::reportError(void)
{
	GtkWidget *replies = dlog->replies;
	gchar *info[4];
	gint row;

	// Build the column strings to be displayed.
	info[0] = g_strerror(errno);
	info[1] = "";
	info[2] = "";
	info[3] = "";

	// Add the new row to the list.
	gtk_clist_freeze(GTK_CLIST(replies));
	row = gtk_clist_append(GTK_CLIST(replies), info);
	gtk_clist_set_selectable(GTK_CLIST(replies), row, FALSE);
	gtk_clist_moveto(GTK_CLIST(replies), row, 0, 1.0, 0.0);
	gtk_clist_thaw(GTK_CLIST(replies));

	// Abort the ping.
	dlog->setSensitive(dlog->startbutton, TRUE);
	dlog->setSensitive(dlog->stopbutton, FALSE);
	dlog->setDefault(dlog->startbutton);
	dlog->setSensitive(dlog->hostnameLabel, TRUE);
	dlog->setSensitive(dlog->hostname, TRUE);
	return 0;
}

void PingDialogSupport::reportStats(unsigned long sent, unsigned long received,
							 		struct timeval& minTrip,
									struct timeval& maxTrip,
									struct timeval& totalTrip)
{
	char *temp;
	double avg;
	if(received > 0)
	{
		avg = (((double)(totalTrip.tv_sec)) * 1000.0 +
		       ((double)(totalTrip.tv_usec)) / 1000.0) / ((double)received);
		temp = g_strdup_printf
				(_("%lu sent, %lu received, avg round trip = %.1f ms"),
				 sent, received, avg);
	}
	else
	{
		temp = g_strdup_printf(_("%lu sent, %lu received"), sent, received);
	}
	gtk_label_set_text(GTK_LABEL(dlog->status), temp);
	g_free(temp);
	gdk_flush();
}

void PingDialogSupport::stopping(void)
{
	dlog->setSensitive(dlog->startbutton, TRUE);
	dlog->setSensitive(dlog->stopbutton, FALSE);
	dlog->setDefault(dlog->startbutton);
	dlog->setSensitive(dlog->hostnameLabel, TRUE);
	dlog->setSensitive(dlog->hostname, TRUE);
}

PingDialogBox::PingDialogBox(GtkWidget *parent)
	: DialogBox(parent, _("Ping"), DIALOG_BUTTON_CLOSE, 0), pinger(this)
{
	GtkWidget *hbox;
	GtkWidget *temp;
	GtkWidget *scroller;
	GSList *group;
	static char *titles[] = {N_("From"), N_("Seq #"), N_("TTL"), N_("Time")};

	// Create the hostname prompt.
	hbox = gtk_hbox_new(0, 0);
	hostnameLabel = gtk_label_new(_("Host to ping:"));
	gtk_box_pack_start(GTK_BOX(hbox), hostnameLabel, FALSE, TRUE, 5);
	hostname = addText();
	setFocus(hostname);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), hostname);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Create the reply report box.
	scroller = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroller),
								   GTK_POLICY_AUTOMATIC,
								   GTK_POLICY_AUTOMATIC);
	replies = gtk_clist_new_with_titles(4, titles);
	gtk_container_add(GTK_CONTAINER(scroller), replies);
	gtk_clist_column_titles_passive(GTK_CLIST(replies));
	gtk_clist_set_column_justification(GTK_CLIST(replies), 1,
									   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification(GTK_CLIST(replies), 2,
									   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification(GTK_CLIST(replies), 3,
									   GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_width(GTK_CLIST(replies), 0, 160);
	gtk_clist_set_column_width(GTK_CLIST(replies), 1, 8 * 5);
	gtk_clist_set_column_width(GTK_CLIST(replies), 2, 8 * 5);
	gtk_widget_set_usize(scroller, 350, 155);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), scroller);

	// Create the status line.
	hbox = gtk_hbox_new(0, 0);
	temp = gtk_label_new(_("Status:"));
	gtk_box_pack_start(GTK_BOX(hbox), temp, FALSE, TRUE, 5);
	status = gtk_label_new(_("0 sent, 0 received"));
	gtk_box_pack_start(GTK_BOX(hbox), status, FALSE, TRUE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), hbox);

	// Add the start and stop buttons.
	startbutton = addButton(_("Start"), 1);
	stopbutton = addButton(_("Stop"));

	// Trap interesting signals.
	gtk_signal_connect(GTK_OBJECT(hostname), "changed",
					   GTK_SIGNAL_FUNC(cb_hostname_changed), this);

	// Set the initial button states.
	setSensitive(startbutton, FALSE);
	setSensitive(stopbutton, FALSE);
	setDefault(closebutton);

	// No ping start timeout yet.
	startTimeoutId = (guint)(-1);
}

PingDialogBox::~PingDialogBox()
{
	stopPing();
}

void PingDialogBox::cb_hostname_changed(GtkWidget *widget, PingDialogBox *box)
{
	box->hostnameChanged();
}

void PingDialogBox::hostnameChanged(void)
{
	if(isSensitive(hostname))
	{
		if(textEmpty(hostname))
		{
			setSensitive(startbutton, FALSE);
			if(isSensitive(stopbutton))
				setDefault(stopbutton);
			else
				setDefault(closebutton);
		}
		else
		{
			setSensitive(startbutton, TRUE);
			setDefault(startbutton);
		}
	}
}

gboolean PingDialogBox::cb_start_timeout_proc(gpointer data)
{
	PingDialogBox *dlog = (PingDialogBox *)data;
	dlog->startTimeoutId = (guint)(-1);
	dlog->pinger.ping(dlog->pingAddr, 1);
	return FALSE;
}

void PingDialogBox::stopPing(void)
{
	if(startTimeoutId != ((guint)(-1)))
	{
		gtk_timeout_remove(startTimeoutId);
		startTimeoutId = (guint)(-1);
	}
	pinger.stop();
}

void PingDialogBox::close(void)
{
	stopPing();
	closeDialog();
}

void PingDialogBox::wmclose(void)
{
	stopPing();
	closeDialog();
}

void PingDialogBox::button(GtkWidget *buttonWidget)
{
	if(buttonWidget == startbutton)
	{
		// Look up the IP address that is associated with the hostname.
		char *text = gtk_entry_get_text(GTK_ENTRY(hostname));
		if((pingAddr.s_addr = inet_addr(text)) == (unsigned long)(-1L))
		{
			struct hostent *ent = gethostbyname(text);
			if(!ent)
			{
				text = g_strdup_printf(_("%s is not a valid hostname"), text);
				gnome_error_dialog_parented(text, GTK_WINDOW(dialog));
				g_free(text);
				return;
			}
			else
			{
				memcpy(&pingAddr, ent->h_addr, sizeof(pingAddr));
			}
		}

		// Clear the previous display, if any.
		gtk_clist_clear(GTK_CLIST(replies));

		// Make sure that we really can do the ping.
		if(!PingSupport::canPing())
		{
			text = g_strdup_printf(_("Could not open the ping socket: %s"),
								   g_strerror(errno));
			gnome_error_dialog_parented(text, GTK_WINDOW(dialog));
			g_free(text);
			return;
		}

		// Rearrange the sensitivity states of the widgets.
		setSensitive(startbutton, FALSE);
		setSensitive(stopbutton, TRUE);
		setDefault(stopbutton);
		setSensitive(hostnameLabel, FALSE);
		setSensitive(hostname, FALSE);

		// Flush the UI changes in an attempt to reduce the skewing below.
		gdk_flush();

		// Install a timeout to initiate the ping after 200 ms.
		// We need to do this to give GTK/GNOME a little time to
		// settle down after all the widget processing.  Otherwise
		// the first ping will have a badly skewed round-trip time.
		startTimeoutId = gtk_timeout_add(200, cb_start_timeout_proc, this);
	}
	else if(buttonWidget == stopbutton)
	{
		// Stop the pinger.
		stopPing();

		// Reset the sensitivity states of the widgets.
		setSensitive(startbutton, TRUE);
		setSensitive(stopbutton, FALSE);
		setDefault(startbutton);
		setSensitive(hostnameLabel, TRUE);
		setSensitive(hostname, TRUE);
	}
}
