/*
 * PingSupport.cc - Support code for doing network pings.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "PingSupport.h"
#include <sys/uio.h>
#include <netinet/ip_icmp.h>

// Value to advance the global identifier by between ping sessions
// to avoid picking up strays from other sessions or other programs.
#define	ADVANCE_IDENTIFIER		23

unsigned short PingSupport::global_identifier = 0;

PingSupport::PingSupport()
{
	fd = -1;
	maxSent = 100;
	finalCount = 0;

	if(!global_identifier)
	{
		// Initialize the global ping identifier.
		global_identifier = (unsigned short)(getpid());
	}
}

PingSupport::~PingSupport()
{
	stop();
}

/* in_cksum extracted from FreeBSD's version of ping.  BSD Copyright follows */

/*
 * Copyright (c) 1989, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * Mike Muuss.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

u_short in_cksum(u_short *addr, int len)
{
	register int nleft = len;
	register u_short *w = addr;
	register int sum = 0;
	union {
		u_short	us;
		u_char	uc[2];
	} last;
	u_short answer;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if (nleft == 1) {
		last.uc[0] = *(u_char *)w;
		last.uc[1] = 0;
		sum += last.us;
	}

	/* add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = ~sum;				/* truncate to 16 bits */
	return(answer);
}

void PingSupport::sendPing(void)
{
	char packet[64];		// ICMP header plus 56 bytes of data.
	int index;
	struct icmp *hdr = (struct icmp *)packet;

	// Initialize the ICMP header.
	hdr->icmp_type = ICMP_ECHO;
	hdr->icmp_code = 0;
	hdr->icmp_cksum = 0;
	hdr->icmp_seq = htons((unsigned short)numSent);
	hdr->icmp_id = identifier;

	// Put the current time into the first 8 bytes of the data area
	// so that we can measure the round trip time when the reply
	// is sent back to us.
	gettimeofday((struct timeval *)(packet + 8), NULL);

	// Fill the rest of the packet with fixed values.
	for(index = sizeof(struct timeval); index < 56; ++index)
	{
		packet[index + 8] = (char)index;
	}

	// Compute the ICMP checksum.
	hdr->icmp_cksum = in_cksum((u_short *)packet, sizeof(packet));

	// Send the ping packet to the remote host.
again:
	if(sendto(fd, packet, sizeof(packet), 0, (struct sockaddr *)&address,
			  sizeof(address)) < 0)
	{
		if(errno == EINTR)
			goto again;
		if(!reportError())
		{
			stop();
		}
	}
	else
	{
		++numSent;
		reportStats(numSent, numReceived, minRoundTrip,
					maxRoundTrip, totalRoundTrip);
	}
}

void PingSupport::timeoutCallback(void)
{
	if(fd != -1)
	{
		if(numSent >= maxSent)
		{
			// We've reached the ping limit, so don't send any more
			// packets.  Instead, wait up to 5 timeouts for any more
			// outstanding replies and then stop.
			if(++finalCount >= 5)
			{
				finalCount = 5;
				stopping();
				stop();
			}
		}
		else
		{
			// Send another ping to the destination.
			sendPing();
		}
	}
}

void PingSupport::selectCallback(void)
{
	char packet[1024];
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_in from;
	int len, posn;
	struct ip iphdr;
	struct icmp icmphdr;
	struct timeval tv;
	struct timeval now;

	if(fd != -1)
	{
		// Read select occurred: read the incoming ICMP packet.
		memset(&msg, 0, sizeof(msg));
		msg.msg_name = (char *)&from;
		msg.msg_namelen = sizeof(from);
		msg.msg_iov = &iov;
		msg.msg_iovlen = 1;
		iov.iov_base = packet;
		iov.iov_len = sizeof(packet);
		if((len = recvmsg(fd, &msg, 0)) >= 20)
		{
			// Pull apart the IP and ICMP headers on the received packet.
			memcpy(&iphdr, packet, sizeof(struct ip));
			posn = iphdr.ip_hl * 4;
			if(iphdr.ip_p == IPPROTO_ICMP &&
			   len >= (posn + 8 + sizeof(struct timeval)))
			{
				memcpy(&icmphdr, packet + posn, 8);
				if(icmphdr.icmp_type == ICMP_ECHOREPLY &&
				   icmphdr.icmp_id == identifier)
				{
					// This is a ping reply that was directed to us.
					++numReceived;
					memcpy(&tv, packet + posn + 8, sizeof(struct timeval));
					gettimeofday(&now, NULL);
					if(tv.tv_sec < now.tv_sec ||
					   (tv.tv_sec == now.tv_sec &&
					    tv.tv_usec <= now.tv_usec))
					{
						// Determine the round-trip time for the packet.
						now.tv_sec -= tv.tv_sec;
						if(now.tv_usec < tv.tv_usec)
						{
							now.tv_usec = now.tv_usec + 1000000 - tv.tv_usec;
							--(now.tv_sec);
						}
						else
						{
							now.tv_usec -= tv.tv_usec;
						}
					}
					else
					{
						// Huh?  Response arrived before it was sent?
						// If so, treat the ping as zero-length.
						now.tv_sec = 0;
						now.tv_usec = 0;
					}

					// Update the minimum and maximum round trip times.
					if(now.tv_sec < minRoundTrip.tv_sec ||
					   (now.tv_sec == minRoundTrip.tv_sec &&
					    now.tv_usec < minRoundTrip.tv_usec))
					{
						minRoundTrip = now;
					}
					if(now.tv_sec > maxRoundTrip.tv_sec ||
					   (now.tv_sec == maxRoundTrip.tv_sec &&
					    now.tv_usec > maxRoundTrip.tv_usec))
					{
						maxRoundTrip = now;
					}

					// Update the total round trip time.
					totalRoundTrip.tv_sec += now.tv_sec;
					if((totalRoundTrip.tv_usec += now.tv_usec) >= 1000000)
					{
						totalRoundTrip.tv_usec -= 1000000;
						++(totalRoundTrip.tv_sec);
					}

					// Notify the subclass of the packet's arrival.
					pingResponse(from.sin_addr,
								 (unsigned)(ntohs(icmphdr.icmp_seq)),
								 (unsigned)(iphdr.ip_ttl), now);

					// Report the latest statistics to the subclass.
					reportStats(numSent, numReceived,
								minRoundTrip, maxRoundTrip,
								totalRoundTrip);
				}
			}
		}
	}
}

void PingSupport::pingResponse(struct in_addr& addr, unsigned seqnum,
							   unsigned ttl, struct timeval& roundTrip)
{
}

int PingSupport::reportError(void)
{
	return 1;
}

void PingSupport::reportStats(unsigned long sent, unsigned long received,
							  struct timeval& minTrip, struct timeval& maxTrip,
							  struct timeval& totalTrip)
{
}

void PingSupport::stopping(void)
{
}

int PingSupport::canPing(void)
{
	int fd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
	if(fd < 0)
	{
		return 0;
	}
	else
	{
		close(fd);
		return 1;
	}
}

int PingSupport::ping(struct in_addr& addr, int nsecs, int maxToSend)
{
	// If there is already a ping in progress, then stop it first.
	if(fd != -1)
	{
		if(haveTimeout)
			unregisterTimeout();
		unregisterSelect();
		close(fd);
		fd = -1;
		haveTimeout = 0;
	}

	// Open the ping socket.
	if((fd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0)
	{
		return 0;
	}

	// Set the address to be pinged.
	memset(&address, 0, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_addr = addr;

	// Set up the timeout and select callbacks.
	if(nsecs > 0)
	{
		registerTimeout(nsecs);
		haveTimeout = 1;
	}
	else
	{
		haveTimeout = 0;
	}
	registerSelect();

	// Initialize the identifier for this ping session.
	identifier = global_identifier;
	global_identifier += ADVANCE_IDENTIFIER;

	// Initialize the statistics fields.
	maxSent = (unsigned long)maxToSend;
	numSent = 0;
	numReceived = 0;
	finalCount = 0;
	minRoundTrip.tv_sec = 1000000;
	minRoundTrip.tv_usec = 0;
	maxRoundTrip.tv_sec = 0;
	maxRoundTrip.tv_usec = 0;
	totalRoundTrip.tv_sec = 0;
	totalRoundTrip.tv_usec = 0;

	// Send the initial statistics state to the subclass.
	reportStats(numSent, numReceived, minRoundTrip,
				maxRoundTrip, totalRoundTrip);

	// Send the first ping packet and then return to the caller.
	sendPing();
	return 1;
}

void PingSupport::stop(void)
{
	if(fd != -1)
	{
		if(haveTimeout)
			unregisterTimeout();
		unregisterSelect();
		close(fd);
		fd = -1;
		haveTimeout = 0;
	}
}
