/*
 * PrintSupport.cc - Support routines for printing packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "NetPacket.h"
#include "PrintSupport.h"

int32_t gmt2local(time_t t);	// declared in "netpeek.cc".

PrintFormatOutput::PrintFormatOutput(FormatRuleTable *_rules,
									 int _maxPackets,
									 int _allowUnderlines)
{
	rules = _rules;
	sprintf(numFormat, "%d", _maxPackets);
	_maxPackets = strlen(numFormat);
	sprintf(numFormat, "%%%dlu ", _maxPackets);
	allowUnderlines = _allowUnderlines;
	outstream = NULL;
	buflen = 0;
	firstSection = 1;
	thiszone = gmt2local(0);
	today = (time(0) + thiszone) / 86400;
	prevday = 0;
}

PrintFormatOutput::~PrintFormatOutput()
{
	if(outstream != NULL)
	{
		fclose(outstream);
	}
}

void PrintFormatOutput::outputShort(const char *format, ...)
{
	va_list va;
	va_start(va, format);
	vsprintf(buffer + buflen, format, va);
	va_end(va);
	buflen += strlen(buffer + buflen);
}

void PrintFormatOutput::outputShortString(const char *string, int len)
{
	if(len < 0)
	{
		strcpy(buffer + buflen, string);
		buflen += strlen(string);
	}
	else
	{
		memcpy(buffer + buflen, string, len);
		buflen += len;
	}
}

void PrintFormatOutput::outputSection(const char *name)
{
	// Print an extra blank line to separate us from the previous section.
	if(firstSection)
		firstSection = 0;
	else
		putc('\n', outstream);

	// Can we use "_^H" to do underlining on section names?
	if(allowUnderlines)
	{
		while(*name != '\0')
		{
			putc('_', outstream);
			putc(0x08, outstream);
			putc(*name, outstream);
			++name;
		}
		putc('\n', outstream);
	}
	else
	{
		// Output the section name.
		fputs(name, outstream);
		putc('\n', outstream);

		// Output '-' characters on the next line to perform underlining.
		while(*name++ != '\0')
		{
			putc('-', outstream);
		}
		putc('\n', outstream);
	}
}

void PrintFormatOutput::outputLong(const char *field, const char *format, ...)
{
	va_list va;
	char *temp;
	fputs(field, outstream);
	fputs(": ", outstream);
	va_start(va, format);
	vsprintf(buffer, format, va);
	va_end(va);
	if((temp = strstr(buffer, "{{")) == 0)
	{
		fputs(buffer, outstream);
	}
	else
	{
		fwrite(buffer, 1, (int)(temp - buffer), outstream);
		fputs("   ", outstream);
		fwrite(temp + 2, 1, strlen(temp) - 4, outstream);
	}
	putc('\n', outstream);
}

static char hexchars[] = "0123456789abcdef";

void PrintFormatOutput::outputHex(const char *field, const void *data,
						   		  unsigned len)
{
	unsigned char *ptr = (unsigned char *)data;
	unsigned int addr;
	unsigned int offset;
	char *buf;
	char ch;
	if(field)
	{
		fputs(field, outstream);
		fputs(":\n", outstream);
	}
	addr = 0;
	while(addr < len)
	{
		buf = buffer;
		for(offset = 0; offset < 16 && (addr + offset) < len; ++offset)
		{
			*buf++ = hexchars[(ptr[addr + offset] >> 4) & 0x0F];
			*buf++ = hexchars[ptr[addr + offset] & 0x0F];
			*buf++ = ' ';
		}
		while(offset < 16)
		{
			*buf++ = ' ';
			*buf++ = ' ';
			*buf++ = ' ';
			++offset;
		}
		*buf++ = ' ';
		for(offset = 0; offset < 16 && (addr + offset) < len; ++offset)
		{
			ch = ptr[addr + offset];
			if(ch >= 0x20 && ch <= 0x7E)
			{
				*buf++ = ch;
			}
			else
			{
				*buf++ = '.';
			}
		}
		strcpy(buf, "\n");
		fputs(buffer, outstream);
		addr += 16;
	}
}

void PrintFormatOutput::outputText(const char *field, const void *data,
						    	   unsigned len, int suppressCRs)
{
	char *ptr = (char *)data;
	char *buf;
	unsigned int buflen;
	int linelen;
	char ch;
	if(field)
	{
		fputs(field, outstream);
		fputs(":\n", outstream);
	}
	buf = buffer;
	buflen = 0;
	linelen = 0;
	while(len-- > 0)
	{
		ch = *ptr++;
		if(ch >= 0x20 && ch <= 0x7E)
		{
			if(linelen >= 80)
			{
				*buf++ = '\\';
				*buf++ = '\n';
				buflen += 2;
				linelen = 0;
			}
			if(ch == '\\' || ch == '^')
			{
				*buf++ = '\\';
				*buf++ = (char)ch;
				buflen += 2;
				linelen += 2;
			}
			else
			{
				*buf++ = (char)ch;
				++buflen;
				++linelen;
			}
		}
		else if(ch == '\n')
		{
			*buf++ = '\n';
			++buflen;
			linelen = 0;
		}
		else if(ch == '\r')
		{
			if(!suppressCRs)
			{
				*buf++ = '^';
				*buf++ = 'M';
				buflen += 2;
				linelen += 2;
			}
		}
		else if(ch == '\t')
		{
			if(linelen >= 80)
			{
				*buf++ = '\\';
				*buf++ = '\n';
				buflen += 2;
				linelen = 0;
			}
			*buf++ = ' ';
			++buflen;
			++linelen;
			while((linelen % 8) != 0)
			{
				*buf++ = ' ';
				++buflen;
				++linelen;
			}
		}
		else if(ch > 0 && ch < 0x20)
		{
			if(linelen >= 80)
			{
				*buf++ = '\\';
				*buf++ = '\n';
				buflen += 2;
				linelen = 0;
			}
			*buf++ = '^';
			*buf++ = ch + '@';
			buflen += 2;
			linelen += 2;
		}
		else
		{
			if(linelen >= 80)
			{
				*buf++ = '\\';
				*buf++ = '\n';
				buflen += 2;
				linelen = 0;
			}
			*buf++ = '\\';
			*buf++ = '0' + (((int)ch) >> 6) & 0x03;
			*buf++ = '0' + (((int)ch) >> 3) & 0x07;
			*buf++ = '0' + ((int)ch) & 0x07;
			buflen += 4;
			linelen += 4;
		}
		if(buflen >= 2048)
		{
			*buf = '\0';
			fputs(buffer, outstream);
			buf = buffer;
			buflen = 0;
		}
	}
	if(buflen > 0)
	{
		*buf = '\0';
		fputs(buffer, outstream);
	}
	putc('\n', outstream);
}

int PrintFormatOutput::wantLinkHeader(void)
{
	return 0;
}

void PrintFormatOutput::shortFormColor(FormatColor color)
{
	// Nothing to do here.
}

int PrintFormatOutput::resolveIPAddresses(void)
{
	extern int pref_resolve_addresses;
	return pref_resolve_addresses;
}

int PrintFormatOutput::fullyQualifiedNames(void)
{
	extern int pref_fully_qualified;
	return pref_fully_qualified;
}

void PrintFormatOutput::markIPAddress(const void *addr)
{
	// Nothing to do here.
}

int PrintFormatOutput::openCommand(const char *command)
{
	close();
	outstream = popen(command, "w");
	return (outstream != NULL);
}

int PrintFormatOutput::openFile(const char *filename)
{
	close();
	outstream = fopen(filename, "w");
	return (outstream != NULL);
}

void PrintFormatOutput::close(void)
{
	if(outstream != NULL)
	{
		fclose(outstream);
		outstream = NULL;
	}
}

int PrintFormatOutput::writeShort(NetPacket& packet)
{
	int linelen, indent, posn, len, spacelen;

	// Print the packet number.
	fprintf(outstream, numFormat, (unsigned long)(packet.packetNum));
	linelen = ((int)(numFormat[1] - '0')) + 1;

	// Print the timestamp.
	unsigned long t = (packet.header->ts.tv_sec + thiszone);
	unsigned long day = t / 86400;
	t %= 86400;
	if(day != today)
	{
		// The data was captured on a different day, so include the date.
		if(day != prevday)
		{
			// Day changed since the last format call.
			struct tm *tms = localtime(&(packet.header->ts.tv_sec));
			prevday = day;
			memset(datebuf, 0, sizeof(datebuf));
			strftime(datebuf, sizeof(datebuf), "%x", tms);
		}
		fprintf(outstream, "%s %02d:%02d:%02d ", datebuf,
				(int)(t / 3600), (int)((t % 3600) / 60), (int)(t % 60));
		linelen += strlen(datebuf) + 10;
	}
	else
	{
		// The data was captured today, so show the time only.
		fprintf(outstream, "%02d:%02d:%02d.%06u ",
				(int)(t / 3600), (int)((t % 3600) / 60),
				(int)(t % 60), (unsigned)(packet.header->ts.tv_usec));
		linelen += 15;
	}

	// Print the packet size.
	if(packet.linkType == DLT_EN10MB ||
	   packet.linkType == DLT_IEEE802)
	{
		fprintf(outstream, "%4d ", (int)(packet.header->len));
	}
	else
	{
		fprintf(outstream, "%4d ",
		        (int)(packet.header->len - packet.linkAddrLen));
	}
	linelen += 5;

	// Format the data into the short-form buffer.
	buflen = 0;
	rules->formatPacketShort(packet, *this, packet.linkType);

	// Output the short-form data to the print stream.  We also word-wrap
	// the data so that it doesn't go over 78 columns on the page.
	indent = linelen;
	posn = 0;
	while(posn < buflen)
	{
		len = 0;
		while((posn + len) < buflen && buffer[posn + len] == ' ')
		{
			++len;
		}
		spacelen = len;
		while((posn + len) < buflen && buffer[posn + len] != ' ')
		{
			++len;
		}
		if((linelen + len) <= 78 ||
		   (linelen == indent && (len - spacelen) > (78 - indent)))
		{
			// The next word will fit on the current line, or is too
			// big to fit on any line, so just write it as-is.
			if(linelen == indent)
			{
				if(spacelen < len)
				{
					fwrite(buffer + posn + spacelen, 1,
						   len - spacelen, outstream);
				}
			}
			else
			{
				fwrite(buffer + posn, 1, len, outstream);
			}
			linelen += len;
		}
		else
		{
			// Move the word down to the next line.
			posn += spacelen;
			len -= spacelen;
			putc('\n', outstream);
			spacelen = indent;
			while(spacelen-- > 0)
			{
				putc(' ', outstream);
			}
			if(len > 0)
			{
				fwrite(buffer + posn, 1, len, outstream);
			}
			linelen = indent + len;
		}
		posn += len;
	}
	putc('\n', outstream);
	return !(ferror(outstream));
}

int PrintFormatOutput::writeLong(NetPacket& packet)
{
	// Output the long-form data to the print stream.
	firstSection = 1;
	rules->formatPacketLong(packet, *this, packet.linkType);

	// Output a form feed to start the next packet on a new page.
	putc('\f', outstream);

	// Return the stream's error state to the caller.
	return !(ferror(outstream));
}
