/*
 * Resolver.cc - Interface to the IP address to name resolver.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "Resolver.h"
#include <setjmp.h>
#include <signal.h>

static int resolver_fd = -1; // File descriptor to send to the resolver.
static int parent_fd = -1;	 // File descriptor to send to the parent process.
static unsigned long subnet_address = 0xFFFFFFFF; // Address of this subnet.
static unsigned long subnet_mask = 0; // Subnet mask of this subnet.
static int resolve_foreign = 0; // Non-zero to resolve foreign network addrs.

// Structure of the resolver's cache.  The data structure is a hash
// table, based on the least-significant 8 bits of the address.
// We also cache failure, to ensure that we don't try to resolve
// the same bad address more than once.
#define	RESOLVER_STATUS_HAVE_NAME		0
#define	RESOLVER_STATUS_PENDING			1
#define	RESOLVER_STATUS_PREPRIMED		2
#define	RESOLVER_STATUS_FAILED			3
typedef struct _tagResolverEntry
{
	struct _tagResolverEntry *next;
	struct in_addr	address;
	int				status;
	char			name[64];
	char			shortname[64];

} ResolverEntry;
static ResolverEntry *cache[256];

// Structure of a resolver request.
typedef struct
{
	unsigned long	address;

} ResolverRequest;

// Structure of a resolver reply.
typedef struct
{
	unsigned long	address;
	unsigned long	status;
	char			name[64];

} ResolverReply;

static void ResolverRun(void);
static int read_exact(int fd, void *data, int len);
static int write_exact(int fd, const void *data, int len);

int ResolverStart(char *argv0)
{
	int pair[2];
	int pid;

	// Create a pair of sockets for communicating to and from the
	// resolver child process.
	if(socketpair(AF_UNIX, SOCK_STREAM, 0, pair) < 0)
	{
		return -1;
	}
	resolver_fd = pair[0];
	parent_fd = pair[1];

	// Fork off a child process to handle IP address to hostname resolution.
	if((pid = fork()) == 0)
	{
		// We are in the child process: close the resolver fd copy.
		close(resolver_fd);
		resolver_fd = -1;

		// Replace the process name with something more descriptive.
		if(strlen(argv0) >= 7)
		{
			strcpy(argv0, "(npdns)");
		}

		// Run the resolver.  This will never return.
		ResolverRun();

		// Keep the compiler happy.
		return -1;
	}
	else if(pid > 0)
	{
		// We are in the parent process: close the parent fd copy.
		close(parent_fd);
		parent_fd = -1;

		// Set "close on exec" on the resolver fd so that other
		// programs that we exec won't get a copy of it.
		fcntl(resolver_fd, F_SETFD, 1);

		// Return the resolver fd to the caller for selecting.
		return resolver_fd;
	}
	else
	{
		// An error occurred.
		close(resolver_fd);
		close(parent_fd);
		resolver_fd = -1;
		parent_fd = -1;
		return -1;
	}
}

void ResolverStop(void)
{
	if(resolver_fd != -1)
	{
		// Send a quit message to the resolver.
		ResolverRequest req;
		req.address = 0;
		write_exact(resolver_fd, &req, sizeof(req));

		// Close the connection to the resolver.
		close(resolver_fd);
		resolver_fd = -1;
	}
}

void ResolverNetwork(struct in_addr address, struct in_addr netmask,
                     int resolveForeign)
{
	if(address.s_addr != htonl(0x7F000001))
	{
		// Normal network interface.
		subnet_address = address.s_addr & netmask.s_addr;
		subnet_mask = netmask.s_addr;
		resolve_foreign = resolveForeign;
	}
	else
	{
		// Loopback network interface: disable hostname lookups
		// because we probably don't have a live connection to
		// a working DNS anyway.
		subnet_address = 0xFFFFFFFF;
		subnet_mask = 0;
		resolve_foreign = 0;
	}
}

void ResolverChangeForeign(int resolveForeign)
{
	resolve_foreign = resolveForeign;
}

const char *ResolverLookup(struct in_addr address, int suppressDomain)
{
	int isForeign = 0;

	// Check for special cases.
	if(address.s_addr == htonl(0x7F000001))
	{
		// Always translate 127.0.0.1 into "localhost", regardless
		// of how "/etc/hosts" or DNS may be set up.  That way, even
		// if the machine is grossly misconfigured, it won't mess
		// up our handling of the loopback interface.
		return "localhost";
	}
	else if(address.s_addr == 0x00000000 || address.s_addr == 0xFFFFFFFF)
	{
		// Never resolve all-zeros or all-ones addresses.  Such values
		// are used as special indicators in some protocols, and are
		// never likely to have a real hostname associated with them.
		return NULL;
	}
	else if((address.s_addr & subnet_mask) == subnet_address)
	{
		// IP address on the same subnet as this host.  Check for
		// the subnet and broadcast addresses.
		if(address.s_addr == subnet_address ||
		   (address.s_addr & ~subnet_mask) == ~subnet_mask)
		{
			return NULL;
		}
	}
	else if(!resolve_foreign)
	{
		// IP address on a foreign subnet which we don't want to resolve
		// by using a live network request.  If there is something in
		// the cache, then we do want to use that.
		isForeign = 1;
	}

	// Inspect the cache to see if we have resolved this address before,
	// or if a request for the address is already pending.
	ResolverEntry *entry = cache[ntohl(address.s_addr) & 0xFF];
	while(entry != NULL)
	{
		if(entry->address.s_addr == address.s_addr)
		{
			if(entry->status != RESOLVER_STATUS_HAVE_NAME &&
			   entry->status != RESOLVER_STATUS_PREPRIMED)
			{
				// The request is already pending within the resolver,
				// or we've already failed to find the hostname before,
				// so don't send another request.
				return NULL;
			}
			else
			{
				// We've seen this IP address before and have a name for it.
				if(suppressDomain)
				{
					return entry->shortname;
				}
				else
				{
					return entry->name;
				}
			}
		}
		entry = entry->next;
	}

	// If the resolver doesn't exist any more, or if the IP address
	// is foreign, then give up now.
	if(resolver_fd == -1 || isForeign)
	{
		return NULL;
	}

	// Make a resolver entry for the IP address and add it to the cache.
	entry = (ResolverEntry *)malloc(sizeof(ResolverEntry));
	if(!entry)
	{
		return NULL;
	}
	entry->next = cache[ntohl(address.s_addr) & 0xFF];
	entry->address = address;
	entry->status = RESOLVER_STATUS_PENDING;
	cache[ntohl(address.s_addr) & 0xFF] = entry;

	// Send the IP address to the resolver.
	ResolverRequest req;
	req.address = address.s_addr;
	if(!write_exact(resolver_fd, &req, sizeof(req)))
	{
		// Something nasty happened: the resolver isn't running
		// any more (probably because someone killed it manually).
		close(resolver_fd);
		ResolverDisappeared(resolver_fd);
		resolver_fd = -1;
		return NULL;
	}

	// The request is in progress - return back to the caller
	// without a name for now.
	return NULL;
}

const char *ResolverLookupCached(struct in_addr address, int suppressDomain)
{
	// Check for special cases.
	if(address.s_addr = htonl(0x7F000001))
	{
		// "localhost" is always implicitly in the cache.
		return "localhost";
	}

	// Inspect the cache for the address.
	ResolverEntry *entry = cache[ntohl(address.s_addr) & 0xFF];
	while(entry != NULL)
	{
		if(entry->address.s_addr == address.s_addr)
		{
			if(entry->status != RESOLVER_STATUS_HAVE_NAME &&
			   entry->status != RESOLVER_STATUS_PREPRIMED)
			{
				// The request is already pending within the resolver,
				// or we've already failed to find the hostname before,
				// so we don't yet have a cached hostname.
				return NULL;
			}
			else
			{
				// We've seen this IP address before and have a name for it.
				if(suppressDomain)
				{
					return entry->shortname;
				}
				else
				{
					return entry->name;
				}
			}
		}
		entry = entry->next;
	}

	// Cannot find a cached hostname for the address.
	return NULL;
}

void ResolverAddToCache(struct in_addr address, const char *hostname)
{
	char *temp;

	// See if we have a cached entry for the IP address already.
	ResolverEntry *entry = cache[ntohl(address.s_addr) & 0xFF];
	while(entry != NULL)
	{
		if(entry->address.s_addr == address.s_addr)
		{
			if(entry->status == RESOLVER_STATUS_PENDING ||
			   entry->status == RESOLVER_STATUS_FAILED)
			{
				// There is a request already pending with the resolver,
				// or a previous resolver lookup didn't find anything.
				// In the case of a pending request, we use the supplied
				// hostname temporarily until we get a more up to date
				// version from the resolver.  In the case of a failed
				// lookup, we replace the failure with the supplied value
				// because it is likely that we can no longer reach
				// the DNS server that supplied the original hostname,
				// but we do at least know what hostname it supplied.
				strncpy(entry->name, hostname, sizeof(entry->name) - 1);
				entry->name[sizeof(entry->name) - 1] = '\0';
				strcpy(entry->shortname, entry->name);
				temp = strchr(entry->shortname, '.');
				if(temp)
				{
					*temp = '\0';
				}
				if(entry->status == RESOLVER_STATUS_FAILED)
				{
					// We need to notify the surrounding application that
					// we now have a hostname for this address, because it
					// is still waiting for such a notification.
					entry->status = RESOLVER_STATUS_PREPRIMED;
					ResolverCallback(address);
				}
				else
				{
					entry->status = RESOLVER_STATUS_PREPRIMED;
				}
			}
			return;
		}
		entry = entry->next;
	}

	// Make a new permanent entry for the IP address.
	entry = (ResolverEntry *)malloc(sizeof(ResolverEntry));
	if(!entry)
	{
		return;
	}
	entry->next = cache[ntohl(address.s_addr) & 0xFF];
	entry->address = address;
	entry->status = RESOLVER_STATUS_HAVE_NAME;
	strncpy(entry->name, hostname, sizeof(entry->name) - 1);
	entry->name[sizeof(entry->name) - 1] = '\0';
	strcpy(entry->shortname, entry->name);
	temp = strchr(entry->shortname, '.');
	if(temp)
	{
		*temp = '\0';
	}
	cache[ntohl(address.s_addr) & 0xFF] = entry;

	// Notify the surrounding application that we have a hostname for
	// this address now, in case previous packets referred to the IP
	// address but "ResolverLookup" skipped the address because of the
	// subnet settings.
	ResolverCallback(address);
}

void ResolverClear(void)
{
	int index;
	ResolverEntry *entry;
	ResolverEntry *next;
	for(index = 0; index < 256; ++index)
	{
		entry = cache[index];
		while(entry != NULL)
		{
			next = entry->next;
			free(entry);
			entry = next;
		}
		cache[index] = NULL;
	}
}

void ResolverSelect(void)
{
	ResolverReply rep;
	char *temp;

	// Read the next reply from the resolver.
	if(!read_exact(resolver_fd, &rep, sizeof(rep)))
	{
		// The connection with the resolver has died, probably because
		// someone killed the process manually.
		close(resolver_fd);
		ResolverDisappeared(resolver_fd);
		resolver_fd = -1;
		return;
	}

	// Search for the IP address within the local cache.
	ResolverEntry *entry = cache[ntohl(rep.address) & 0xFF];
	while(entry != NULL)
	{
		if(entry->address.s_addr == rep.address &&
		   entry->status == RESOLVER_STATUS_PENDING)
		{
			// We've found the entry, so update it.
			if(entry->status == RESOLVER_STATUS_PENDING ||
			   (entry->status == RESOLVER_STATUS_PREPRIMED &&
			    rep.status == RESOLVER_STATUS_HAVE_NAME))
			{
				// We have a new name to associate with the address.
				entry->status = rep.status;
				if(rep.status == RESOLVER_STATUS_HAVE_NAME)
				{
					// Copy the name and shorten it to the domain-less version.
					memcpy(entry->name, rep.name, sizeof(rep.name));
					memcpy(entry->shortname, rep.name, sizeof(rep.name));
					temp = strchr(entry->shortname, '.');
					if(temp)
					{
						*temp = '\0';
					}

					// Notify the surrounding application of the new address.
					ResolverCallback(entry->address);
				}
			}
			else if(entry->status == RESOLVER_STATUS_PREPRIMED &&
			        rep.status == RESOLVER_STATUS_FAILED)
			{
				// The cache was pre-primed while the DNS request was
				// in progress.  Even though the request has now failed,
				// we do need to notify the application of success so
				// that any packets encountered before the pre-prime
				// will be re-visited and updated.
				ResolverCallback(entry->address);
			}
			break;
		}
		entry = entry->next;
	}
}

// Handle time-outs for "gethostbyaddr" lookup requests.
static jmp_buf resolver_env;
static void ResolverTimeOut(int sig)
{
	longjmp(resolver_env, 1);
}

// Main loop for the resolver child process.  This performs the actual
// work of translating IP addresses into hostnames.
static void ResolverRun(void)
{
	ResolverRequest req;
	ResolverReply rep;
	struct hostent *ent;

	// Read requests from the parent, process them, and send back replies.
	while(read_exact(parent_fd, &req, sizeof(req)))
	{
		if(req.address == 0)
		{
			// The parent process sent us a quit message.
			break;
		}
		if(!setjmp(resolver_env))
		{
			// Set a 20 second timeout around the "gethostbyaddr" call to
			// prevent DNS lookups on bad IP's from retrying for too long.
			signal(SIGALRM, ResolverTimeOut);
			alarm(20);
			ent = gethostbyaddr((char *)&(req.address),
							    sizeof(req.address), AF_INET);
			alarm(0);
		}
		else
		{
			// The request timed out.
			ent = NULL;
		}
		if(ent)
		{
			rep.address = req.address;
			rep.status = RESOLVER_STATUS_HAVE_NAME;
			strncpy(rep.name, ent->h_name, sizeof(rep.name) - 1);
			rep.name[sizeof(rep.name) - 1] = '\0';
		}
		else
		{
			rep.address = req.address;
			rep.status = RESOLVER_STATUS_FAILED;
			rep.name[0] = '\0';
		}
		if(!write_exact(parent_fd, &rep, sizeof(rep)))
		{
			break;
		}
	}

	// If we get here, then the connection with the parent has dropped,
	// or an explicit quit request was received.  In this case, we just
	// exit from the process.
	exit(0);
}

// Read an exact number of bytes, while taking care of annoying errors.
static int read_exact(int fd, void *data, int len)
{
	int templen;
	while(len > 0 && (templen = read(fd, data, len)) < len)
	{
		if(templen < 0)
		{
			if(errno != EINTR && errno != EWOULDBLOCK)
			{
				return 0;
			}
		}
		else
		{
			data = (void *)(((char *)data) + templen);
			len -= templen;
		}
	}
	return 1;
}

// Write an exact number of bytes, while taking care of annoying errors.
static int write_exact(int fd, const void *data, int len)
{
	int templen;
	while(len > 0 && (templen = write(fd, data, len)) < len)
	{
		if(templen < 0)
		{
			if(errno != EINTR && errno != EWOULDBLOCK)
			{
				return 0;
			}
		}
		else
		{
			data = (const void *)(((const char *)data) + templen);
			len -= templen;
		}
	}
	return 1;
}
