/*
 * TCPFormatRule.cc - Formatting rules for TCP packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "TCPFormatRule.h"
#include "TelnetFormatRule.h"
#include "NetPacket.h"
#ifdef linux
	#define	__FAVOR_BSD
#endif
#include <netinet/tcp.h>

/* Additional TCP options extracted from the source for tcpdump */

/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef TCPOPT_WSCALE
#define	TCPOPT_WSCALE		3	/* window scale factor (rfc1072) */
#endif
#ifndef TCPOPT_SACKOK
#define	TCPOPT_SACKOK		4	/* selective ack ok (rfc1072) */
#endif
#ifndef TCPOPT_SACK
#define	TCPOPT_SACK			5	/* selective ack (rfc1072) */
#endif
#ifndef TCPOPT_ECHO
#define	TCPOPT_ECHO			6	/* echo (rfc1072) */
#endif
#ifndef TCPOPT_ECHOREPLY
#define	TCPOPT_ECHOREPLY	7	/* echo (rfc1072) */
#endif
#ifndef TCPOPT_TIMESTAMP
#define TCPOPT_TIMESTAMP	8	/* timestamps (rfc1323) */
#endif
#ifndef TCPOPT_CC
#define TCPOPT_CC			11	/* T/TCP CC options (rfc1644) */
#endif
#ifndef TCPOPT_CCNEW
#define TCPOPT_CCNEW		12	/* T/TCP CC options (rfc1644) */
#endif
#ifndef TCPOPT_CCECHO
#define TCPOPT_CCECHO		13	/* T/TCP CC options (rfc1644) */
#endif

#define	ZEROLENOPT(opt)	((opt) == TCPOPT_EOL || (opt) == TCPOPT_NOP)

/*
 * Table of TCP protocols that should be dumped as a plain text session.
 */
typedef struct
{
	unsigned long	port;
	const char     *sectionName;

} TCPTextProtocol;
static TCPTextProtocol textProtocols[] = {
	{21,	"File Transfer Protocol (FTP)"},
	{25,	"Simple Mail Transport Protocol (SMTP)"},
	{70,	"Gopher"},
	{79,	"Finger"},
	{109,	"Post Office Protocol version 2 (POP2)"},
	{110,	"Post Office Protocol version 3 (POP3)"},
	{119,	"Network News Transport Protocol (NNTP)"},
	{143,	"Internet Message Access Protocol version 2 (IMAP2)"},
	{220,	"Internet Message Access Protocol version 3 (IMAP3)"},
	{0,		0},
};

TCPFormatRule::TCPFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
	sessions = 0;

	// Add telnet to the set of TCP text-based port rules.
	new TelnetFormatRule(&portRules, 23, "Telnet", 1, 0);

	// Add all of the text telnet-like protocols.
	TCPTextProtocol *proto = textProtocols;
	while(proto->port)
	{
		new TelnetFormatRule(&portRules, proto->port, proto->sectionName, 0, 1);
		++proto;
	}
}

TCPFormatRule::~TCPFormatRule()
{
	clearState();
}

void TCPFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	struct tcphdr header;
	unsigned flags;
	const u_char *options;
	unsigned optlen;

	// Extract the TCP header information.
	if(packet.subPacketLen < 20)
	{
		output.shortFormColor(FormatColor_TCPData);
		output.outputShortString("truncated tcp packet");
		return;
	}
	memcpy(&header, packet.subPacket, 20);
	packet.srcPort = ntohs(header.th_sport);
	packet.dstPort = ntohs(header.th_dport);
	if((((unsigned)(header.th_off)) * 4) > packet.subPacketLen)
	{
		output.shortFormColor(FormatColor_TCPData);
		output.outputShortString("truncated tcp packet");
		return;
	}
	options = packet.subPacket + 20;
	optlen = (((unsigned)(header.th_off)) * 4) - 20;
	packet.adjustSubPacket(((unsigned)(header.th_off)) * 4);

	// Output the source and destination address information.
	if(packet.isVersion6())
	{
		output.outputShort("%s.%s > ",
						   output.ipaddr6ToString(packet.ip + 8),
						   output.tcpPortToString(packet.srcPort));
		output.outputShort("%s.%s: ",
						   output.ipaddr6ToString(packet.ip + 24),
						   output.tcpPortToString(packet.dstPort));
	}
	else
	{
		output.outputShort("%s.%s > ",
						   output.ipaddrToString(&(packet.ipHeader.ip_src)),
						   output.tcpPortToString(packet.srcPort));
		output.outputShort("%s.%s: ",
						   output.ipaddrToString(&(packet.ipHeader.ip_dst)),
						   output.tcpPortToString(packet.dstPort));
	}

	// Dump the TCP header information to the output stream.
	flags = header.th_flags;
	if((flags & (TH_SYN | TH_FIN | TH_RST | TH_PUSH)) != 0)
	{
		if(flags & TH_SYN)
			output.outputShortString("S");
		if(flags & TH_FIN)
			output.outputShortString("F");
		if(flags & TH_RST)
			output.outputShortString("R");
		if(flags & TH_PUSH)
			output.outputShortString("P");
	}
	else
	{
		output.outputShortString(".");
	}
	if((flags & TH_SYN) != 0)
	{
		output.shortFormColor(FormatColor_TCPConnect);
	}
	else if((flags & TH_FIN) != 0)
	{
		output.shortFormColor(FormatColor_TCPDisconnect);
	}
	else
	{
		output.shortFormColor(FormatColor_TCPData);
	}
	if(packet.subPacketLen > 0 || (flags & (TH_SYN | TH_FIN | TH_RST)) != 0)
	{
		output.outputShort(" %u:%u(%u)", (unsigned)ntohl(header.th_seq),
						   (unsigned)(ntohl(header.th_seq) +
						   			  packet.subPacketLen),
						   packet.subPacketLen);
	}
	if((flags & TH_ACK) != 0)
	{
		output.outputShort(" ack %u", (unsigned)ntohl(header.th_ack));
	}
	output.outputShort(" win %d", (int)ntohs(header.th_win));
	if((flags & TH_URG) != 0)
	{
		output.outputShort(" urg %d", (int)ntohs(header.th_urp));
	}

	// Dump any options that may be present on the TCP packet.
	if(optlen > 0)
	{
		char optbuf[512];
		char *out = optbuf;
		char sep = '<';
		unsigned len, opt;
		const u_char *nextopt;
		*out++ = ' ';
		while(optlen > 0)
		{
			*out++ = sep;
			sep = ',';
			opt = *options;
			if(ZEROLENOPT(opt))
			{
				len = 1;
			}
			else
			{
				len = options[1];
			}
			nextopt = options + len;
			if(len == 1)
				++options;
			else
				options += 2;
			switch(opt)
			{
				case TCPOPT_MAXSEG:
				{
					sprintf(out, "mss %u",
							(((unsigned)(options[0])) << 8) |
							 ((unsigned)(options[1])));
					out += strlen(out);
				}
				break;

				case TCPOPT_EOL:
				{
					strcpy(out, "eol");
					out += 3;
				}
				break;

				case TCPOPT_NOP:
				{
					strcpy(out, "nop");
					out += 3;
				}
				break;

				case TCPOPT_WSCALE:
				{
					sprintf(out, "wscale %u", (unsigned)(*options));
					out += strlen(out);
				}
				break;

				case TCPOPT_SACKOK:
				{
					strcpy(out, "sackOK");
					out += 6;
				}
				break;

				case TCPOPT_SACK:
				{
					int datalen;
					strcpy(out, "sack");
					out += 4;
					datalen = (int)(len - 2);
					while(datalen > 0)
					{
						sprintf(out, " %u@%u",
							    (((unsigned)(options[2])) << 8) |
							     ((unsigned)(options[3])),
							    (((unsigned)(options[0])) << 8) |
							     ((unsigned)(options[1])));
						out += strlen(out);
						options += 4;
						datalen -= 4;
					}
					if(datalen != 0)
					{
						sprintf(out, "[len %d]", (int)len);
						out += strlen(out);
					}
				}
				break;

				case TCPOPT_ECHO:
				{
					sprintf(out, "echo %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
					out += strlen(out);
				}
				break;

				case TCPOPT_ECHOREPLY:
				{
					sprintf(out, "echoreply %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
					out += strlen(out);
				}
				break;

				case TCPOPT_TIMESTAMP:
				{
					sprintf(out, "timestamp %lu %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])),
							(((unsigned long)(options[4])) << 24) |
							(((unsigned long)(options[5])) << 16) |
							(((unsigned long)(options[6])) << 8) |
							 ((unsigned long)(options[7])));
					out += strlen(out);
				}
				break;

				case TCPOPT_CC:
				{
					sprintf(out, "cc %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
					out += strlen(out);
				}
				break;

				case TCPOPT_CCNEW:
				{
					sprintf(out, "ccnew %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
					out += strlen(out);
				}
				break;

				case TCPOPT_CCECHO:
				{
					sprintf(out, "ccecho %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
					out += strlen(out);
				}
				break;

				default:
				{
					unsigned datalen = len - 2;
					sprintf(out, "opt-%d:", (int)opt);
					out += strlen(out);
					while(datalen > 0)
					{
						sprintf(out, "%02x", (int)(*options));
						out += strlen(out);
						++options;
						--datalen;
					}
				}
				break;
			}
			options = nextopt;
			optlen -= len;
			if(opt == TCPOPT_EOL)
			{
				break;
			}
		}
		*out++ = '>';
		*out = '\0';
		output.outputShortString(optbuf);
	}

	// Try directing the packet using the destination or source port first.
	if(!portRules.formatPacketShort(packet, output, packet.dstPort) &&
	   !portRules.formatPacketShort(packet, output, packet.srcPort))
	{
		// See if we have a session registered for the port.
		TCPSessionInfo *session = sessions;
		while(session != 0)
		{
			if(session->addr == packet.ipHeader.ip_dst.s_addr &&
			   session->port == packet.dstPort)
			{
				// Use the lookup port number to format the packet.
				portRules.formatPacketShort(packet, output,
											session->lookupPort);
				return;
			}
			session = session->next;
		}
	}
}

void TCPFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	struct tcphdr header;
	unsigned flags;
	const u_char *options;
	unsigned optlen;

	// Extract the TCP header information.
	if(packet.subPacketLen < 20)
	{
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}
	memcpy(&header, packet.subPacket, 20);
	packet.srcPort = ntohs(header.th_sport);
	packet.dstPort = ntohs(header.th_dport);
	if((((unsigned)(header.th_off)) * 4) > packet.subPacketLen)
	{
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}
	options = packet.subPacket + 20;
	optlen = (((unsigned)(header.th_off)) * 4) - 20;
	packet.adjustSubPacket(((unsigned)(header.th_off)) * 4);

	// Output the contents of the TCP header.
	output.outputSection("Transmission Control Protocol (TCP)");
	const char *str = output.tcpPortToString(packet.srcPort, 0);
	if(str)
	{
		output.outputLong("Source Port", "%d {{%s}}",
						  (int)(packet.srcPort), str);
	}
	else
	{
		output.outputLong("Source Port", "%d", (int)(packet.srcPort));
	}
	str = output.tcpPortToString(packet.dstPort, 0);
	if(str)
	{
		output.outputLong("Destination Port", "%d {{%s}}",
						  (int)(packet.dstPort), str);
	}
	else
	{
		output.outputLong("Destination Port", "%d", (int)(packet.dstPort));
	}
	output.outputLong("Sequence Number", "%lu",
					  (unsigned long)(ntohl(header.th_seq)));
	output.outputLong("Acknowledgement Number", "%lu",
					  (unsigned long)(ntohl(header.th_ack)));
	output.outputLong("Data Offset", "%u words (%u bytes)",
					  ((unsigned)(header.th_off)),
					  ((unsigned)(header.th_off)) * 4);
	flags = (unsigned)(header.th_flags);
	if(flags)
	{
		char flagbuf[128];
		flagbuf[0] = '\0';
		if((flags & TH_FIN) != 0)
		{
			strcat(flagbuf, "FIN");
		}
		if((flags & TH_SYN) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "SYN");
			else
				strcat(flagbuf, ", SYN");
		}
		if((flags & TH_RST) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "RST");
			else
				strcat(flagbuf, ", RST");
		}
		if((flags & TH_PUSH) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "PUSH");
			else
				strcat(flagbuf, ", PUSH");
		}
		if((flags & TH_ACK) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "ACK");
			else
				strcat(flagbuf, ", ACK");
		}
		if((flags & TH_URG) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "URG");
			else
				strcat(flagbuf, ", URG");
		}
		if((flags & 0x40) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "0x40");
			else
				strcat(flagbuf, ", 0x40");
		}
		if((flags & 0x80) != 0)
		{
			if(flagbuf[0] == '\0')
				strcat(flagbuf, "0x80");
			else
				strcat(flagbuf, ", 0x80");
		}
		output.outputLong("Flags", "0x%02x {{%s}}", (int)flags, flagbuf);
	}
	else
	{
		output.outputLong("Flags", "0x%02x", (int)flags);
	}
	output.outputLong("Window Size", "%u", (unsigned)(ntohs(header.th_win)));
	output.outputLong("Checksum", "0x%04x", (int)(ntohs(header.th_sum)));
	output.outputLong("Urgent Pointer", "%u", (unsigned)(ntohs(header.th_urp)));

	// Dump any options that may be present on the TCP packet.
	if(optlen > 0)
	{
		char optbuf[512];
		unsigned len, opt;
		const u_char *nextopt;
		while(optlen > 0)
		{
			opt = *options;
			if(ZEROLENOPT(opt))
			{
				len = 1;
			}
			else
			{
				len = options[1];
			}
			nextopt = options + len;
			if(len == 1)
				++options;
			else
				options += 2;
			switch(opt)
			{
				case TCPOPT_MAXSEG:
				{
					output.outputLong("Maximum Segment Size Option", "%u",
		  							  (((unsigned)(options[0])) << 8) |
							 		   ((unsigned)(options[1])));
				}
				break;

				case TCPOPT_EOL:	break;
				case TCPOPT_NOP:	break;

				case TCPOPT_WSCALE:
				{
					output.outputLong("Window Scale Factor Option", "%u",
									  (unsigned)(*options));
				}
				break;

				case TCPOPT_SACKOK:
				{
					output.outputLong("Selective ACK OK Option", "Present");
				}
				break;

				case TCPOPT_SACK:
				{
					int datalen;
					char *out = optbuf;
					datalen = (int)(len - 2);
					while(datalen > 0)
					{
						sprintf(out, (out != optbuf ? " %u@%u" : "%u@%u"),
							    (((unsigned)(options[2])) << 8) |
							     ((unsigned)(options[3])),
							    (((unsigned)(options[0])) << 8) |
							     ((unsigned)(options[1])));
						out += strlen(out);
						options += 4;
						datalen -= 4;
					}
					if(datalen != 0)
					{
						sprintf(out, "[len %d]", (int)len);
						out += strlen(out);
					}
					*out = '\0';
					output.outputLong("Selective ACK Option", "%s", optbuf);
				}
				break;

				case TCPOPT_ECHO:
				{
					output.outputLong("Echo Option", "%lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
				}
				break;

				case TCPOPT_ECHOREPLY:
				{
					output.outputLong("Echo Reply Option", "%lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
				}
				break;

				case TCPOPT_TIMESTAMP:
				{
					output.outputLong("Timestamp Option", "%lu %lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])),
							(((unsigned long)(options[4])) << 24) |
							(((unsigned long)(options[5])) << 16) |
							(((unsigned long)(options[6])) << 8) |
							 ((unsigned long)(options[7])));
				}
				break;

				case TCPOPT_CC:
				{
					output.outputLong("CC Option", "%lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
				}
				break;

				case TCPOPT_CCNEW:
				{
					output.outputLong("CC New Option", "%lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
				}
				break;

				case TCPOPT_CCECHO:
				{
					output.outputLong("CC Echo Option", "%lu",
							(((unsigned long)(options[0])) << 24) |
							(((unsigned long)(options[1])) << 16) |
							(((unsigned long)(options[2])) << 8) |
							 ((unsigned long)(options[3])));
				}
				break;

				default:
				{
					char *out = optbuf;
					char *value;
					unsigned datalen = len - 2;
					sprintf(out, "TCP Option %d:", (int)opt);
					out += strlen(out) + 1;
					value = out;
					while(datalen > 0)
					{
						sprintf(out, "%02x", (int)(*options));
						out += strlen(out);
						++options;
						--datalen;
					}
					*out = '\0';
					output.outputLong(optbuf, "%s", value);
				}
				break;
			}
			options = nextopt;
			optlen -= len;
			if(opt == TCPOPT_EOL)
			{
				break;
			}
		}
	}

	// If there is no data payload, then bail out here.
	if(packet.subPacketLen == 0)
	{
		return;
	}

	// Try directing the packet using the destination or source port first.
	if(!portRules.formatPacketLong(packet, output, packet.dstPort) &&
	   !portRules.formatPacketLong(packet, output, packet.srcPort))
	{
		// See if we have a session registered for the port.
		TCPSessionInfo *session = sessions;
		while(session != 0)
		{
			if(session->addr == packet.ipHeader.ip_dst.s_addr &&
			   session->port == packet.dstPort)
			{
				// Use the lookup port number to format the packet.
				portRules.formatPacketLong(packet, output,
										   session->lookupPort);
				return;
			}
			session = session->next;
		}

		// We don't know how to process this packet if we get here.
		output.outputSection("TCP Payload Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
	}
}

void TCPFormatRule::clearState(void)
{
	TCPSessionInfo *current;
	TCPSessionInfo *next;
	current = sessions;
	while(current != 0)
	{
		next = current->next;
		delete current;
		current = next;
	}
	sessions = 0;
	portRules.clearState();
}

TCPPortFormatRule::TCPPortFormatRule(TCPFormatRule *_tcp,
								     FormatRuleTable *_table,
					  				 unsigned long _tag)
	: FormatRule(_table, _tag)
{
	tcp = _tcp;
}

TCPPortFormatRule::~TCPPortFormatRule()
{
}

void TCPPortFormatRule::registerSession(unsigned long addr,
										unsigned short port,
						 				unsigned short lookupPort)
{
	TCPSessionInfo *session;

	// Make sure that the session is not already present.  This prevents
	// multiple session records being created when watching TCP retries.
	session = tcp->sessions;
	while(session != 0)
	{
		if(session->addr == addr &&
		   session->port == port)
		{
			// Record the lookup port in case the previous session
			// was aborted and the destination port is now being
			// reused for a different protocol.
			session->lookupPort = lookupPort;
			return;
		}
		session = session->next;
	}

	// Create a new session record.
	session = new TCPSessionInfo;
	if(session)
	{
		session->next = tcp->sessions;
		session->addr = addr;
		session->port = port;
		session->lookupPort = lookupPort;
		tcp->sessions = session;
	}
}

void TCPPortFormatRule::unregisterSession(unsigned long addr,
									      unsigned short port,
									      unsigned short lookupPort)
{
	TCPSessionInfo *session;
	TCPSessionInfo *prev;

	// Find the session record.
	session = tcp->sessions;
	prev = 0;
	while(session != 0)
	{
		if(session->addr == addr &&
		   session->port == port &&
		   session->lookupPort == lookupPort)
		{
			// Remove the session from the list.
			if(prev)
				prev->next = session->next;
			else
				tcp->sessions = session->next;
			delete session;
			break;
		}
		prev = session;
		session = session->next;
	}
}
