/*
 * TelnetFormatRule.cc - Formatting rules for telnet packets.  This formatter
 *                       can also be used to format simple text-based protocols
 *                       like SMTP, NNTP, POP3, etc.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "TelnetFormatRule.h"
#include "NetPacket.h"
#define	TELCMDS				// Include command code strings
#define	TELOPTS				// Include option code strings
#define	SLC_NAMES			// Include SLC name strings
#include <arpa/telnet.h>

TelnetFormatRule::TelnetFormatRule(FormatRuleTable *_table,
								   unsigned long _tag,
					 			   const char *_sectionName,
								   int _telnetCommands,
								   int _suppressCRs)
	: FormatRule(_table, _tag)
{
	sectionName = _sectionName;
	telnetCommands = _telnetCommands;
	suppressCRs = _suppressCRs;
}

TelnetFormatRule::~TelnetFormatRule()
{
}

void TelnetFormatRule::outputBlock(FormatOutput& output,
								   const void *data, unsigned len)
{
	const u_char *temp;
	unsigned templen;
	unsigned count;
	u_char tempch;

	// Scan the data to count the number of bytes that look
	// like binary.  If there are 10 or more of them, then we
	// dump the data in hex instead of as text, since it
	// could be a binary message transfer in SMTP or some-such.
	temp = (const u_char *)data;
	templen = len;
	count = 0;
	while(templen-- > 0)
	{
		tempch = *temp++;
		if(tempch == 0 || tempch >= 0x80)
		{
			++count;
		}
	}
	if(count >= 10)
	{
		// Dump the packet in hex.
		output.outputHex(0, data, len);
	}
	else
	{
		// Dump the text as-is.  Embedded binary characters will
		// be printed as "^#" character escapes or in octal.
		output.outputText(0, data, len);
	}
}

static char hexchars[] = "0123456789ABCDEF";

unsigned TelnetFormatRule::formatOption(char *buf, int option,
						  			    const unsigned char *data,
										unsigned len)
{
	unsigned posn;

	// Determine the actual length of the option data, by looking for SE.
	posn = 0;
	while(posn < len)
	{
		if(data[posn] == (u_char)0xFF && (posn + 1) < len)
		{
			if(data[posn + 1] == (u_char)SE)
			{
				break;
			}
			++posn;
		}
		++posn;
	}
	len = posn;

	// Determine the option-specific formatting rule to use.
	switch(option)
	{
		case TELOPT_SNDLOC:
		{
			// Format the option as a literal string.
			for(posn = 0; posn < len; ++posn)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
					if(posn < len)
					{
						*buf++ = (char)(data[posn]);
					}
				}
				else
				{
					*buf++ = (char)(data[posn]);
				}
			}
		}
		break;

		case TELOPT_TSPEED:
		case TELOPT_XDISPLOC:
		case TELOPT_TTYPE:
		{
			// Format the option as a literal string with IS/SEND conventions.
			posn = 0;
			if(len > 0)
			{
				if(data[0] == TELQUAL_IS)
				{
					strcpy(buf, "IS ");
					buf += 3;
				}
				else if(data[0] == TELQUAL_SEND)
				{
					strcpy(buf, "SEND ");
					buf += 5;
				}
				else
				{
					sprintf(buf, "SUBCMD-%d ", (int)(data[0]));
					buf += strlen(buf);
				}
				++posn;
			}
			for(; posn < len; ++posn)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
					if(posn < len)
					{
						*buf++ = (char)(data[posn]);
					}
				}
				else
				{
					*buf++ = (char)(data[posn]);
				}
			}
		}
		break;

		case TELOPT_3270REGIME:
		{
			// Format the option as a literal string with IS/ARE conventions.
			posn = 0;
			if(len > 0)
			{
				if(data[0] == TELQUAL_IS)
				{
					strcpy(buf, "IS ");
					buf += 3;
				}
				else if(data[0] == TELQUAL_SEND)
				{
					strcpy(buf, "ARE ");
					buf += 4;
				}
				else
				{
					sprintf(buf, "SUBCMD-%d ", (int)(data[0]));
					buf += strlen(buf);
				}
				++posn;
			}
			for(; posn < len; ++posn)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
					if(posn < len)
					{
						*buf++ = (char)(data[posn]);
					}
				}
				else
				{
					*buf++ = (char)(data[posn]);
				}
			}
		}
		break;

		case TELOPT_NAWS:
		{
			// Window size option.
			int width = 0;
			int height = 0;
			posn = 0;
			if(posn < len)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
				}
				width = (width * 256) + (int)(data[posn]);
				++posn;
			}
			if(posn < len)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
				}
				width = (width * 256) + (int)(data[posn]);
				++posn;
			}
			if(posn < len)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
				}
				height = (height * 256) + (int)(data[posn]);
				++posn;
			}
			if(posn < len)
			{
				if(data[posn] == (u_char)0xFF)
				{
					++posn;
				}
				height = (height * 256) + (int)(data[posn]);
				++posn;
			}
			sprintf(buf, "%d,%d", width, height);
			buf += strlen(buf);
		}
		break;

		case TELOPT_LFLOW:
		{
			// Flow control option.
			if(len > 0 && data[0] < 4)
			{
				static char *flowvalues[] =
					{"OFF", "ON", "RESTART ANY", "RESTART XON"};
				strcpy(buf, flowvalues[data[0]]);
				buf += strlen(buf);
			}
			else
			{
				goto formathex;
			}
		}
		break;

		case TELOPT_LINEMODE:
		{
			// Line mode control option.
			if(len > 0)
			{
				if(data[0] == LM_MODE)
				{
					strcpy(buf, "MODE ");
					buf += 5;
					if(len > 1)
					{
						if((data[1] & MODE_EDIT) != 0)
						{
							strcpy(buf, "EDIT ");
							buf += 5;
						}
						if((data[1] & MODE_TRAPSIG) != 0)
						{
							strcpy(buf, "TRAPSIG ");
							buf += 8;
						}
						if((data[1] & MODE_ACK) != 0)
						{
							strcpy(buf, "ACK ");
							buf += 4;
						}
						if((data[1] & MODE_SOFT_TAB) != 0)
						{
							strcpy(buf, "SOFT_TAB ");
							buf += 9;
						}
						if((data[1] & MODE_LIT_ECHO) != 0)
						{
							strcpy(buf, "LIT_ECHO ");
							buf += 9;
						}
						if((data[1] & 0xE0) != 0)
						{
							sprintf(buf, "EXTRAS-%02X", (int)(data[1] & 0xE0));
							buf += strlen(buf);
						}
					}
				}
				else if(data[0] == LM_FORWARDMASK)
				{
					goto formathex;
				}
				else if(data[0] == LM_SLC)
				{
					int function, modifiers, ch;
					static char *slc_levels[] =
						{"NOSUPPORT ", "CANTCHANGE ", "VARIABLE ", "DEFAULT "};
					static char *slc_new_names[] =
						{"MCL", "MCR", "MCWL", "MCWR", "MCBOL", "MCEOL",
						 "INSRT", "OVER", "ECR", "EWR", "EBOL", "EEOL"};
					posn = 0;
					while(posn < len)
					{
						function = 0;
						modifiers = 0;
						ch = 0;
						if(posn != 0)
						{
							strcpy(buf, ", ");
							buf += 2;
						}
						if(posn < len)
						{
							if(data[posn] == (u_char)0xFF)
							{
								++posn;
							}
							function = (int)(data[posn++]);
						}
						if(posn < len)
						{
							if(data[posn] == (u_char)0xFF)
							{
								++posn;
							}
							modifiers = (int)(data[posn++]);
						}
						if(posn < len)
						{
							if(data[posn] == (u_char)0xFF)
							{
								++posn;
							}
							ch = (int)(data[posn++]);
						}
						if(SLC_NAME_OK(function))
						{
							// Original function codes from RFC 1116.
							sprintf(buf, "%s ", slc_names[function]);
							buf += strlen(buf);
						}
						else if(function >= 19 && function <= 30)
						{
							// New function codes added in RFC 1184.
							sprintf(buf, "%s ", slc_new_names[function - 19]);
							buf += strlen(buf);
						}
						else
						{
							sprintf(buf, "FUNCTION-%d ", function);
							buf += strlen(buf);
						}
						strcpy(buf, slc_levels[modifiers & SLC_LEVELBITS]);
						buf += strlen(buf);
						if((modifiers & SLC_ACK) != 0)
						{
							strcpy(buf, "ACK ");
							buf += 4;
						}
						if((modifiers & SLC_FLUSHIN) != 0)
						{
							strcpy(buf, "FLUSHIN ");
							buf += 8;
						}
						if((modifiers & SLC_FLUSHOUT) != 0)
						{
							strcpy(buf, "FLUSHOUT ");
							buf += 9;
						}
						if((modifiers & 0x1C) != 0)
						{
							sprintf(buf, "EXTRAS-%02X", (modifiers & 0x1C));
							buf += strlen(buf);
						}
						if(ch == 0)
						{
							strcpy(buf, "NUL");
							buf += 3;
						}
						else
						{
							*buf++ = (char)ch;
						}
					}
				}
				else
				{
					goto formathex;
				}
			}
		}
		break;

		case TELOPT_NEW_ENVIRON:
		{
			// New environment option.
			if(len > 0)
			{
				if(data[0] == TELQUAL_IS)
				{
					strcpy(buf, "IS ");
					buf += 3;
				}
				else if(data[0] == TELQUAL_SEND)
				{
					strcpy(buf, "SEND ");
					buf += 5;
				}
				else if(data[0] == TELQUAL_INFO)
				{
					strcpy(buf, "INFO ");
					buf += 5;
				}
				else
				{
					sprintf(buf, "SUBCMD-%d", (int)(data[0]));
					buf += strlen(buf);
				}
				posn = 1;
				int sawch = 0;
				int sawesc = 0;
				while(posn < len)
				{
					if(data[posn] == NEW_ENV_VAR)
					{
						if(sawesc)
						{
							strcpy(buf, "VAR ");
							buf += 4;
							sawesc = 0;
						}
						else
						{
							strcpy(buf, "\nVAR ");
							buf += 5;
						}
						sawch = 0;
					}
					else if(data[posn] == NEW_ENV_VALUE)
					{
						if(sawch)
							*buf++ = ' ';
						strcpy(buf, "VALUE ");
						buf += 6;
						sawch = 0;
						sawesc = 0;
					}
					else if(data[posn] == ENV_ESC)
					{
						if(sawch)
							*buf++ = ' ';
						strcpy(buf, "ESC ");
						buf += 4;
						sawch = 0;
						sawesc = 1;
					}
					else if(data[posn] == ENV_USERVAR)
					{
						if(sawesc)
						{
							strcpy(buf, "USERVAR ");
							buf += 8;
							sawesc = 0;
						}
						else
						{
							strcpy(buf, "\nUSERVAR ");
							buf += 9;
						}
						sawch = 0;
						sawesc = 0;
					}
					else
					{
						*buf++ = (char)(data[posn]);
						sawch = 1;
						sawesc = 0;
					}
					++posn;
				}
			}
		}
		break;

		default:
		{
			// We don't know what this is so format the option in hex.
		formathex:
			for(posn = 0; posn < len; ++posn)
			{
				if(posn != 0)
				{
					*buf++ = ' ';
				}
				*buf++ = hexchars[(data[posn] >> 4) & 0x0F];
				*buf++ = hexchars[data[posn] & 0x0F];
			}
		}
		break;
	}
	*buf = '\0';

	// Return the length of the option data to the caller.
	return len;
}

void TelnetFormatRule::formatPacketShort(NetPacket& packet,
										 FormatOutput& output)
{
	// Nothing to do here: TCPFormatRule has already done all the work.
}

void TelnetFormatRule::formatPacketLong(NetPacket& packet,
									    FormatOutput& output)
{
	const u_char *temp;
	const u_char *temp2;
	unsigned templen;
	unsigned count;
	char buffer[4096];
	char *buf;

	// Output the section header, including the data direction.
	strcpy(buffer, sectionName);
	if(((unsigned long)(packet.dstPort)) == getTag())
	{
		strcat(buffer, " (Client to Server)");
	}
	else
	{
		strcat(buffer, " (Server to Client)");
	}
	output.outputSection(buffer);

	// Check for embedded telnet commands.
	if(telnetCommands &&
	   (temp = (const u_char *)memchr(packet.subPacket, 0xFF,
	   								  packet.subPacketLen)) != 0)
	{
		count = packet.subPacketLen;
		temp2 = packet.subPacket;
		while(count > 0)
		{
			templen = (unsigned)(temp - temp2);
			if(templen > 0)
			{
				// Output a text block prior to the next telnet command.
				outputBlock(output, temp2, templen);
				count -= templen;
			}
			if(count > 0)
			{
				// Output a telnet command.
				++temp;
				--count;
				if(count > 0)
				{
					if(TELCMD_OK(*temp))
					{
						switch(*temp)
						{
							case DONT:
							case DO:
							case WONT:
							case WILL:
							{
								if(count > 1)
								{
									if(TELOPT_OK(temp[1]))
									{
										sprintf(buffer,
												"TELNET-COMMAND: %s %s",
												TELCMD(*temp),
												TELOPT(temp[1]));
									}
									else
									{
										sprintf(buffer,
												"TELNET-COMMAND: %s %d",
												TELCMD(*temp),
												(int)(temp[1]));
									}
									temp += 2;
									count -= 2;
								}
								else
								{
									sprintf(buffer,
											"TELNET-COMMAND: %s (truncated)",
											TELCMD(*temp));
									++temp;
									--count;
								}
							}
							break;

							case SB:
							{
								// Option value subnegotiation.
								if(count > 1)
								{
									++temp;
									--count;
									if(TELOPT(*temp))
									{
										sprintf(buffer,
												"TELNET-COMMAND: SB %s ",
												TELOPT(*temp));
									}
									else
									{
										sprintf(buffer,
												"TELNET-COMMAND: SB %d ",
												(int)(*temp));
									}
									int opt = (int)(*temp);
									++temp;
									--count;
									buf = buffer + strlen(buffer);
									unsigned proclen =
										formatOption(buf, opt, temp, count);
									temp += proclen;
									count -= proclen;
								}
								else
								{
									sprintf(buffer,
											"TELNET-COMMAND: SB (truncated)");
									temp += count;
									count = 0;
								}
							}
							break;

							case SE:
							{
								// Ignore SE, because it is just an end marker.
								++temp;
								--count;
								goto skiptext;
							}
							break;

							default:
							{
								sprintf(buffer, "TELNET-COMMAND: %s",
										TELCMD(*temp));
								++temp;
								--count;
							}
							break;
						}
						output.outputText(0, buffer, strlen(buffer));
					}
					else
					{
						sprintf(buffer, "TELNET-COMMAND: Unknown command %d",
								(int)(*temp));
						output.outputText(0, buffer, strlen(buffer));
						++temp;
						--count;
					}
				}
			skiptext:
				temp2 = temp;
				if(count > 0)
				{
					temp = (const u_char *)memchr(temp2, 0xFF, count);
					if(!temp)
					{
						// Output the final text block fragment.
						outputBlock(output, temp2, count);
						break;
					}
				}
			}
		}
	}
	else
	{
		// Output the packet as either text or hex, depending on
		// the contents of the packet.
		outputBlock(output, packet.subPacket, packet.subPacketLen);
	}
}
