/*
 * UDPFormatRule.cc - Formatting rules for UDP packets.
 *
 * Copyright (c) 1999, Rhys Weatherley.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Rhys Weatherley.
 * 4. The name of Rhys Weatherley may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "UDPFormatRule.h"
#include "DNSFormatRule.h"
#include "NetPacket.h"

UDPFormatRule::UDPFormatRule(FormatRuleTable *_table, unsigned long _tag)
	: FormatRule(_table, _tag)
{
	sessions = 0;
	new DNSFormatRule(&portRules, 53);
}

UDPFormatRule::~UDPFormatRule()
{
	clearState();
}

void UDPFormatRule::formatPacketShort(NetPacket& packet, FormatOutput& output)
{
	// Set the packet color to "UDP".
	output.shortFormColor(FormatColor_UDP);

	// Extract the source and destination port values.
	if(packet.subPacketLen < 8)
	{
		output.outputShortString("truncated udp packet");
		return;
	}
	packet.srcPort = packet.getShortNetwork(0);
	packet.dstPort = packet.getShortNetwork(2);
	if(packet.subPacketLen >= (unsigned)(packet.getShortNetwork(4)))
	{
		packet.adjustSubPacket(8, packet.subPacketLen -
								(unsigned)(packet.getShortNetwork(4)));
	}
	else
	{
		packet.adjustSubPacket(8);
	}

	// Output the source and destination address information.
	if(packet.isVersion6())
	{
		output.outputShort("%s.%s > ",
						   output.ipaddr6ToString(packet.ip + 8),
						   output.udpPortToString(packet.srcPort));
		output.outputShort("%s.%s: ",
						   output.ipaddr6ToString(packet.ip + 24),
						   output.udpPortToString(packet.dstPort));
	}
	else
	{
		output.outputShort("%s.%s > ",
						   output.ipaddrToString(&(packet.ipHeader.ip_src)),
						   output.udpPortToString(packet.srcPort));
		output.outputShort("%s.%s: ",
						   output.ipaddrToString(&(packet.ipHeader.ip_dst)),
						   output.udpPortToString(packet.dstPort));
	}

	// Try directing the packet using the destination or source port first.
	if(!portRules.formatPacketShort(packet, output, packet.dstPort) &&
	   !portRules.formatPacketShort(packet, output, packet.srcPort))
	{
		// See if we have a session registered for the port.
		UDPSessionInfo *session = sessions;
		while(session != 0)
		{
			if(session->addr == packet.ipHeader.ip_dst.s_addr &&
			   session->port == packet.dstPort)
			{
				// Use the lookup port number to format the packet.
				portRules.formatPacketShort(packet, output,
											session->lookupPort);
				return;
			}
			session = session->next;
		}

		// We don't know how to process this packet if we get here.
		output.outputShort("%u bytes of udp data", packet.subPacketLen);
	}
}

void UDPFormatRule::formatPacketLong(NetPacket& packet, FormatOutput& output)
{
	const char *str;

	// Extract the source and destination port values.
	if(packet.subPacketLen < 8)
	{
		output.outputHex("Truncated Packet", packet.subPacket,
						 packet.subPacketLen);
		return;
	}
	packet.srcPort = packet.getShortNetwork(0);
	packet.dstPort = packet.getShortNetwork(2);

	// Output the UDP header information.
	output.outputSection("User Datagram Protocol (UDP)");
	str = output.udpPortToString(packet.srcPort, 0);
	if(str)
	{
		output.outputLong("Source Port", "%d {{%s}}",
						  (int)(packet.srcPort), str);
	}
	else
	{
		output.outputLong("Source Port", "%d", (int)(packet.srcPort));
	}
	str = output.udpPortToString(packet.dstPort, 0);
	if(str)
	{
		output.outputLong("Destination Port", "%d {{%s}}",
						  (int)(packet.dstPort), str);
	}
	else
	{
		output.outputLong("Destination Port", "%d", (int)(packet.dstPort));
	}
	output.outputLong("UDP Length", "%d", (int)(packet.getShortNetwork(4)));
	output.outputLong("UDP Checksum", "0x%04x",
					  (int)(packet.getShortNetwork(6)));
	if(packet.subPacketLen >= (unsigned)(packet.getShortNetwork(4)))
	{
		packet.adjustSubPacket(8, packet.subPacketLen -
								(unsigned)(packet.getShortNetwork(4)));
	}
	else
	{
		packet.adjustSubPacket(8);
	}

	// Try directing the packet using the destination or source port first.
	if(!portRules.formatPacketLong(packet, output, packet.dstPort) &&
	   !portRules.formatPacketLong(packet, output, packet.srcPort))
	{
		// See if we have a session registered for the port.
		UDPSessionInfo *session = sessions;
		while(session != 0)
		{
			if(session->addr == packet.ipHeader.ip_dst.s_addr &&
			   session->port == packet.dstPort)
			{
				// Use the lookup port number to format the packet.
				portRules.formatPacketLong(packet, output,
										   session->lookupPort);
				return;
			}
			session = session->next;
		}

		// We don't know how to process this packet if we get here.
		output.outputSection("UDP Payload Data");
		output.outputHex(0, packet.subPacket, packet.subPacketLen);
	}
}

void UDPFormatRule::clearState(void)
{
	UDPSessionInfo *current;
	UDPSessionInfo *next;
	current = sessions;
	while(current != 0)
	{
		next = current->next;
		delete current;
		current = next;
	}
	sessions = 0;
	portRules.clearState();
}

UDPPortFormatRule::UDPPortFormatRule(UDPFormatRule *_udp,
								     FormatRuleTable *_table,
					  				 unsigned long _tag)
	: FormatRule(_table, _tag)
{
	udp = _udp;
}

UDPPortFormatRule::~UDPPortFormatRule()
{
}

void UDPPortFormatRule::registerSession(unsigned long addr,
										unsigned short port,
						 				unsigned short lookupPort)
{
	UDPSessionInfo *session;

	// Make sure that the session is not already present.  This prevents
	// multiple session records being created when watching UDP retries.
	session = udp->sessions;
	while(session != 0)
	{
		if(session->addr == addr &&
		   session->port == port)
		{
			// Record the lookup port in case the previous session
			// was aborted and the destination port is now being
			// reused for a different protocol.
			session->lookupPort = lookupPort;
			return;
		}
		session = session->next;
	}

	// Create a new session record.
	session = new UDPSessionInfo;
	if(session)
	{
		session->next = udp->sessions;
		session->addr = addr;
		session->port = port;
		session->lookupPort = lookupPort;
		udp->sessions = session;
	}
}

void UDPPortFormatRule::unregisterSession(unsigned long addr,
									      unsigned short port,
									      unsigned short lookupPort)
{
	UDPSessionInfo *session;
	UDPSessionInfo *prev;

	// Find the session record.
	session = udp->sessions;
	prev = 0;
	while(session != 0)
	{
		if(session->addr == addr &&
		   session->port == port &&
		   session->lookupPort == lookupPort)
		{
			// Remove the session from the list.
			if(prev)
				prev->next = session->next;
			else
				udp->sessions = session->next;
			delete session;
			break;
		}
		prev = session;
		session = session->next;
	}
}
